/*
 * resourced
 * Copyright (c) 2017 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License"),
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
*/

/*
 * @file fd-handler.c
 *
 * @desc fd handler using glib
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <glib.h>
#include <gio/gio.h>

#include "trace.h"
#include "fd-handler.h"

struct fd_handler_s {
	int fd;
	GIOChannel *ch;
	guint id;
	fd_changed_cb changed;
	void *data;
	release_cb free_func;
	fd_handler_h *handler_p;
};

int remove_fd_read_handler(fd_handler_h *handler)
{
	struct fd_handler_s *h;

	if (!handler || !(*handler))
		return -EINVAL;

	h = *handler;

	if (h->free_func)
		h->free_func(h->data);

	if (h->id)
		g_source_remove(h->id);

	if (h->ch)
		g_io_channel_unref(h->ch);

	*(h->handler_p) = NULL;
	free(h);

	return 0;
}

static gboolean channel_changed(GIOChannel *source,
		GIOCondition condition, gpointer data)
{
	struct fd_handler_s *h = data;
	bool ret;

	if (!h)
		return TRUE;

	if (h->ch != source)
		return TRUE;

	if (h->fd != g_io_channel_unix_get_fd(source))
		return TRUE;

	if (condition != G_IO_IN)
		return TRUE;

	ret = true;
	if (h->changed)
		ret = h->changed(h->fd, h->data);

	if (ret)
		return TRUE;

	remove_fd_read_handler((fd_handler_h *)&h);

	return FALSE;
}

int add_fd_read_handler(int fd,
		fd_changed_cb callback, void *data,
		release_cb free_func, fd_handler_h *handler)
{
	GIOChannel *ch;
	struct fd_handler_s *h;
	guint id;

	if (fd < 0 || !callback)
		return -EINVAL;

	h = calloc(1, sizeof(struct fd_handler_s));
	if (!h)
		return -ENOMEM;

	ch = g_io_channel_unix_new(fd);
	if (!ch) {
		_E("Failed to create GIOChannel");
		free(h);
		return -ENOMEM;
	}

	id = g_io_add_watch(ch, G_IO_IN | G_IO_ERR,
			channel_changed, h);
	if (id == 0) {
		_E("Failed to add watch for GIOChannel");
		free(h);
		g_io_channel_unref(ch);
		return -ENOMEM;
	}

	h->fd = fd;
	h->changed = callback;
	h->data = data;
	h->free_func = free_func;
	h->ch = ch;
	h->id = id;
	h->handler_p = handler;

	*handler = h;

	return 0;
}
