/*
 * resourced
 *
 * Copyright (c) 2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <errno.h>
#include <stdbool.h>
#include <string.h>
#include <linux/loop.h>

#include "losetup.h"
#include "file-helper.h"
#include "trace.h"
#include "macro.h"
#include "util.h"

#define LOOP_BACKING  "/sys/devices/virtual/block/loop%d/loop/backing_file"

/* List of available compressor for losetup, must be the same as in kernel config */
static const char *avaliable_compressors[] = { "aes" };

/* returns: 1=unused, 0=busy, others=error */
static int check_loop_device(int fd)
{
	struct loop_info64 loopinfo64;
	int r;

	r = ioctl(fd, LOOP_GET_STATUS64, &loopinfo64);
	if (errno == ENXIO)
		return 1;

	if (!r)
		return 0;

	return -errno;
}

int find_loop_device(char *file, char *device)
{
	int r, i, unused = -ENOENT;
	char path[MAX_NAME_LENGTH];

	for (i = 0; i < 8; i++) {
		_cleanup_close_ int fd = -1;

		snprintf(path, MAX_NAME_LENGTH, "/dev/loop%d", i);
		fd = open(path, O_RDONLY);
		if (fd < 0)
			continue;

		r = check_loop_device(fd);

		/*
		 * failed to find loop device retry
		 */
		if (r < 0)
			continue;

		if (r == 0) {
			_cleanup_free_ char *backing_file = NULL;

			backing_file = (char*)calloc(1, MAX_NAME_LENGTH);
			if (!backing_file)
				return -ENOMEM;

			snprintf(path, MAX_NAME_LENGTH, LOOP_BACKING, i);
			r = fread_str(path, &backing_file);
			if (!r && streq(backing_file, file)) {
				_I("same file %s has been already existed", file);
				snprintf(device, MAX_NAME_LENGTH, "/dev/loop%d", i);
				return -EEXIST;
			}
			continue;
		}

		if (unused == -ENOENT)
			unused = i;
	}
	if (unused >= 0) {
		snprintf(device, MAX_NAME_LENGTH, "/dev/loop%d", unused);
		return 0;
	}
	return -ENOENT;
}

int losetup(const char *device,
	    const char *file,
	    unsigned long long offset,
	    unsigned long long sizelimit,
	    const char *crypt_name,
	    const char *passwd,
	    bool readdonly)
{
	_cleanup_close_ int dev_fd = -1, file_fd = -1;
	struct loop_info64 loopinfo64;
	int r, mode;

	mode = readdonly ? O_RDONLY : O_RDWR;

	dev_fd = open(device, mode);
	if (dev_fd < 0) {
		_E("Failed to open device %s: %m", device);
		return -errno;
	}

	file_fd = open(file, mode);
	if (file_fd < 0) {
		_E("Failed to open file %s: %m", file);
		return -errno;
	}

	memset(&loopinfo64, 0, sizeof(loopinfo64));

	strncpy((char *) loopinfo64.lo_file_name, file, LO_NAME_SIZE-1);

	loopinfo64.lo_offset = offset;
	loopinfo64.lo_sizelimit = sizelimit;

	if (!is_empty(crypt_name)) {
		if (is_empty(passwd)) {
			_E("Crypt type was given but password missed");
			return -EINVAL;
		}

		loopinfo64.lo_encrypt_type = LO_CRYPT_CRYPTOAPI;
		strncpy((char *) loopinfo64.lo_crypt_name,
			crypt_name,
			LO_NAME_SIZE-1);

		loopinfo64.lo_encrypt_key_size = LO_KEY_SIZE;
	}

	r = ioctl(dev_fd, LOOP_SET_FD, file_fd);
	if (r < 0) {
		_E("Failed to set dev fd to file fd: %m");
		return -errno;
	}

	r = ioctl(dev_fd, LOOP_SET_STATUS64, &loopinfo64);
	if (r < 0) {
		_E("Failed to set loop status: %m");
		return -errno;
	}

	if (!is_empty(crypt_name))
		_I("'%s' is setup to loop device '%s' with '%s' encryption.",
	   file, device, crypt_name);
	else
		_I("'%s' is setup to loop device '%s'.", file, device);

	return 0;
}

int check_valid_compressor(const char* crypt_type)
{
	int i;

	if (!crypt_type)
		return RESOURCED_ERROR_NO_DATA;

	if (strncmp(crypt_type, "none", strlen(crypt_type)) == 0)
		return RESOURCED_ERROR_NO_DATA;

	for (i = 0; i < ARRAY_SIZE(avaliable_compressors); i++)
		if (strncmp(crypt_type, avaliable_compressors[i], strlen(crypt_type)) == 0)
			return RESOURCED_ERROR_NONE;

	return RESOURCED_ERROR_FAIL;
}

