/*
 * resourced
 *
 * Copyright (c) 2014 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

/**
 * @file memory-common.h
 * @desc header file for handling memory cgroups
 **/

#ifndef __MEMORY_COMMONL_H__
#define __MEMORY_COMMONL_H__

#include <glib.h>
#include <stdbool.h>
#include "const.h"

/*
 * [memory cgroup information]
 * MEMCG_MEMORY : root cgroup for system daemons
 * MEMCG_APPS : cgroup for general apps
 * MEMCG_BGLOCKED : cgroup for background locked and favorite apps
 * MEMCG_LIMIT : cgroup for each app with memory limit set if conguration is enabled
 * MEMCG_SWAP : cgroup for selected victims from background apps
 *
 * [memory cgroup hierarchy]
   (root)
    ├─Apps
    │ └─BgLocked
    ├─MemLimit
    ├─Swap
    └─system.slice (not controlled by resourced)
 */
enum memcg_type {
	MEMCG_ROOT = -1,
	MEMCG_MEMORY,
	MEMCG_APPS,
	MEMCG_BGLOCKED,
	MEMCG_LIMIT,
	MEMCG_SWAP,
	MEMCG_MAX,
};

enum {
	LOWMEM_NORMAL,
	LOWMEM_SWAP,
	LOWMEM_LOW,
	LOWMEM_MEDIUM,
	LOWMEM_MAX_LEVEL,
};

enum lowmem_control_type {
	LOWMEM_MOVE_CGROUP,
	LOWMEM_MANAGE_FOREGROUND,
};

/* number of memory cgroups */
#define MEMCG_DEFAULT_EVENT_LEVEL		"low"
#define MEMCG_DEFAULT_USE_HIERARCHY		0

#define LOWMEM_ROOT_CGROUP			"/sys/fs/cgroup/memory"
#define LOWMEM_APPS_CGROUP			LOWMEM_ROOT_CGROUP"/Apps"
#define LOWMEM_BGLOCKED_CGROUP		LOWMEM_ROOT_CGROUP"/Apps/BgLocked"
#define LOWMEM_MEMLIMIT_CGROUP		LOWMEM_ROOT_CGROUP"/MemLimit"
#define LOWMEM_SWAP_CGROUP			LOWMEM_ROOT_CGROUP"/Swap"

#define MEMCG_OOM_CONTROL_PATH		"memory.oom_control"
#define MEMCG_EVENTFD_CONTROL		"cgroup.event_control"
#define MEMCG_EVENTFD_MEMORY_PRESSURE	"memory.pressure_level"
#define MEMCG_USAGE			"memory.usage_in_bytes"
#define MEMCG_SWAP_USAGE		"memory.memsw.usage_in_bytes"
#define MEMCG_LIMIT_PATH		"memory.limit_in_bytes"
#define MEMCG_SWAP_LIMIT_PATH	"memory.memsw.limit_in_bytes"
#define MEMCG_SWAPNESS_PATH     "memory.swappiness"

#define MEMPS_LOG_PATH			"/var/log"

struct memcg_info {
	/* name of memory cgroup */
	char name[MAX_PATH_LENGTH];
	/* hashname of memory cgroup for restoring memcg info*/
	char hashname[MAX_NAME_LENGTH];
	/* parent id */
	int parent_memcg;
	/* limit ratio, if don't want to set limit, use NO_LIMIT*/
	float limit_ratio;
	unsigned int limit;
	/* leave memory usage */
	unsigned int oomleave;
	/* thresholds, normal, swap, low, medium, and leave */
	unsigned int threshold[LOWMEM_MAX_LEVEL];
	unsigned int threshold_leave;
	int evfd;
};

struct memcg {
	/* parent cgroup */
	struct memcg_info *info;
	/* set when using multiple sub cgroups */
	bool use_hierarchy;
	/* list of child cgroups when using multi groups */
	GSList *cgroups;
};

struct lowmem_control_data {
	enum lowmem_control_type control_type;
	int args[2];
};

void memcg_init(struct memcg *memcg);
void memcg_show(struct memcg *memcg);

/**
 * @desc get anon memory usage of cgroup based on memory.stat
 * @return 0 if the value was correctly read
 */
int memcg_get_anon_usage(char *memcg, unsigned int *anon_usage);

/**
 * @desc get swapped memory usage of cgroup mi based on memory.stat
 * @return 0 if the value was correctly read
 */
int memcg_get_swap_usage(char *memcg, unsigned int *usage);

/**
 * @desc register eventfd to the memory cgroup with desired value
 * @return eventfd if it was registered successfully or -1 on failure
 */
int memcg_set_eventfd(const char *memcg, const char *event, char *value);

/**
 * @desc execute /usr/bin/memps and make log file with pid and process name
 */
void make_memps_log(char *file, pid_t pid, char *victim_name);

#endif /*__MEMORY_COMMONL_H__*/
