/*
 * resourced
 *
 * Copyright (c) 2014 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

/**
 * @file module.c
 * @desc Module helper functions
 **/

#include "macro.h"
#include "module.h"
#include "resourced.h"
#include "trace.h"
#include "dbus-handler.h"

#include <glib.h>

static GSList *modules_list;

void add_module(const struct module_ops *module)
{
	ret_msg_if(!module, "Invalid module handler\n");

	if (module->priority == MODULE_PRIORITY_EARLY)
		modules_list = g_slist_prepend(modules_list, (gpointer)module);
	else
		modules_list = g_slist_append(modules_list, (gpointer)module);
}

void remove_module(const struct module_ops *module)
{
	modules_list = g_slist_remove(modules_list, (gpointer)module);
}

const struct module_ops *find_module(const char *name)
{
	GSList *iter;
	const struct module_ops *module;

	gslist_for_each_item(iter, modules_list) {
		module = (struct module_ops *)iter->data;
		if (!strncmp(module->name, name, strlen(module->name) + 1) &&
		   (module->initalized == MODULE_INITIALIZED))
			return module;
	}
	return NULL;
}

void modules_check_runtime_support(void UNUSED *data)
{
	GSList *iter, *next;
	const struct module_ops *module;
	int ret;

	gslist_for_each_safe(modules_list, iter, next, module) {
		module = (const struct module_ops *)iter->data;
		_D("check runtime support [%s] module\n", module->name);

		if (!module->check_runtime_support)
			continue;

		ret = module->check_runtime_support((void *)module);
		if (ret != RESOURCED_ERROR_NONE) {
			_E("%s module check failed", module->name);
			remove_module(module);
			continue;
		}
	}
}

static void module_initcall_level(void *data, int priority)
{
	GSList *iter;
	struct module_ops *module;
	int ret;

	gslist_for_each_item(iter, modules_list) {
		module = (struct module_ops *)iter->data;

		assert(module);

		if (module->priority < priority)
			continue;

		if (module->initalized != MODULE_NONINITIALIZED)
			continue;

		if (!module->init)
			continue;

		ret = module->init(data);
		/* Module disabled on runtime or just failed to start. */
		if (ret < 0) {
			module->initalized = MODULE_DROPPED;
			_E("Fail to initialize [%s] module, dropped.", module->name);
			continue;
		}

		module->initalized = MODULE_INITIALIZED;
		_D("Initialized [%s] module\n", module->name);
	}
}

void modules_init_right_away(void *data)
{
	module_initcall_level(data, MODULE_PRIORITY_HIGH);
}

void modules_init_early(void *data)
{
	module_initcall_level(data, MODULE_PRIORITY_EARLY);
}

void modules_init(void *data)
{
	module_initcall_level(data, MODULE_PRIORITY_NORMAL);
	module_initcall_level(data, MODULE_PRIORITY_LATE);
}

void modules_exit(void *data)
{
	GSList *iter;
	struct module_ops *module;
	int ret;

	gslist_for_each_item(iter, modules_list) {
		module = (struct module_ops *)iter->data;
		if (module->exit && (module->initalized == MODULE_INITIALIZED)) {
			module->initalized = MODULE_NONINITIALIZED;
			ret = module->exit(data);
			if (ret < 0)
				_E("Fail to deinitialize [%s] module\n", module->name);
			else
				_D("Deinitialize [%s] module\n", module->name);
		}
	}
	g_slist_free(modules_list);
}

void modules_restore(void *data)
{
	struct module_ops *module;
	GSList *iter;
	int ret;

	gslist_for_each_item(iter, modules_list) {
		module = (struct module_ops *)iter->data;

		assert(module);

		if (module->initalized != MODULE_INITIALIZED)
			continue;

		if (!module->restore)
			continue;

		ret = module->restore(data);
		if (ret < 0) {
			_E("Fail to restore [%s] module", module->name);
			continue;
		}

		_I("Restored [%s] module", module->name);
	}
}

void modules_dump(FILE *fp, int mode)
{
	GSList *iter;
	const struct module_ops *module;

	gslist_for_each_item(iter, modules_list) {
		module = (struct module_ops *)iter->data;
		if (module->initalized != MODULE_INITIALIZED)
			continue;

		_D("dump [%s] module\n", module->name);
		if (module->dump)
			module->dump(fp, mode, module->dump_data);
	}
}

static void dbus_list_active_modules_handler(GDBusMethodInvocation *invocation, GVariant *params)
{
	struct module_ops *module;
	GSList *list_iter;
	GVariantBuilder builder, *sub_builder;

	g_variant_builder_init(&builder, G_VARIANT_TYPE_TUPLE);
	sub_builder = g_variant_builder_new(G_VARIANT_TYPE("as"));

	gslist_for_each_item(list_iter, modules_list) {
		module = (struct module_ops *)list_iter->data;
		if (module->initalized != MODULE_INITIALIZED)
			continue;

		g_variant_builder_add(sub_builder, "s", module->name);
	}
	g_variant_builder_add_value(&builder, g_variant_new("as", sub_builder));
	g_variant_builder_unref(sub_builder);

	g_dbus_method_invocation_return_value(invocation,
			g_variant_builder_end(&builder));
}

static const char resourced_module_methods_xml[] =
"<node>"
"	<interface name ='"RESOURCED_DBUS_INTERFACE_NAME"'>"
"		<method name='ListActiveModuels'>"
"			<arg type='as' name='ActiveModules' direction='out'/>"
"		</method>"
"	</interface>"
"</node>";

static struct d_bus_method resourced_module_methods[] = {
	{ "ListActiveModuels", dbus_list_active_modules_handler },
	/* Add methods here */
};

int modules_add_methods(void)
{
	return d_bus_register_methods(RESOURCED_DBUS_OBJECT_PATH, resourced_module_methods_xml,
		resourced_module_methods, ARRAY_SIZE(resourced_module_methods));
}
