/*
 * resourced
 *
 * Copyright (c) 2015 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

/**
 * @file procfs.c
 *
 * @desc communicate with procfs in resourced
 *
 * Copyright (c) 2015 Samsung Electronics Co., Ltd. All rights reserved.
 *
 */

#include <ctype.h>
#include <stdio.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

#include "resourced.h"
#include "trace.h"
#include "macro.h"
#include "util.h"
#include "procfs.h"
#include "proc-common.h"
#include "memory-common.h"
#include "module.h"
#include "file-helper.h"
#include "swap-common.h"
#include "smaps.h"
#include "notifier.h"

#define MEM_SWAP_RATIO			0.5

static struct sys_node_table sys_node_tables[] = {
	{ SYS_VM_SHRINK_MEMORY, "/proc/sys/vm/shrink_memory", 1, 1 },
	{ SYS_VM_COMPACT_MEMORY, "/proc/sys/vm/compact_memory", 1, 1 },
	{ },
};

int proc_get_cmdline(pid_t pid, char *cmdline)
{
	char buf[PROC_BUF_MAX];
	char cmdline_buf[PROC_NAME_MAX];
	char *filename;
	FILE *fp;

	snprintf(buf, sizeof(buf), "/proc/%d/cmdline", pid);
	fp = fopen(buf, "r");
	if (fp == NULL)
		return RESOURCED_ERROR_FAIL;

	if (fgets(cmdline_buf, PROC_NAME_MAX-1, fp) == NULL) {
		fclose(fp);
		return RESOURCED_ERROR_FAIL;
	}
	fclose(fp);

	filename = strrchr(cmdline_buf, '/');
	if (filename == NULL)
		filename = cmdline_buf;
	else
		filename = filename + 1;

	strncpy(cmdline, filename, PROC_NAME_MAX-1);

	return RESOURCED_ERROR_NONE;
}

pid_t find_pid_from_cmdline(char *cmdline)
{
	pid_t pid = -1, foundpid = -1;
	int ret = 0;
	DIR *dp;
	struct dirent *dentry;
	char appname[PROC_NAME_MAX];

	dp = opendir("/proc");
	if (!dp) {
		_E("BACKGRD MANAGE : fail to open /proc");
		return RESOURCED_ERROR_FAIL;
	}
	while ((dentry = readdir(dp))) {
		if (!isdigit(dentry->d_name[0]))
			continue;

		pid = atoi(dentry->d_name);
		if (!pid)
			continue;
		ret = proc_get_cmdline(pid, appname);
		if (ret == RESOURCED_ERROR_NONE) {
			if (!strncmp(cmdline, appname, strlen(appname)+1)) {
				foundpid = pid;
				break;
			}
		}
	}
	closedir(dp);
	return foundpid;
}

int proc_get_oom_score_adj(int pid, int *oom_score_adj)
{
	char buf[sizeof(PROC_OOM_SCORE_ADJ_PATH) + MAX_DEC_SIZE(int)] = {0};
	FILE *fp = NULL;

	if (pid <= 0)
		return RESOURCED_ERROR_FAIL;

	snprintf(buf, sizeof(buf), PROC_OOM_SCORE_ADJ_PATH, pid);
	fp = fopen(buf, "r");

	if (fp == NULL) {
		_E("fopen %s failed", buf);
		return RESOURCED_ERROR_FAIL;
	}
	if (fgets(buf, sizeof(buf), fp) == NULL) {
		fclose(fp);
		return RESOURCED_ERROR_FAIL;
	}
	(*oom_score_adj) = atoi(buf);
	fclose(fp);
	return RESOURCED_ERROR_NONE;
}

int proc_set_oom_score_adj(int pid, int oom_score_adj)
{
	FILE *fp;
	struct lowmem_control_data lowmem_data;
	char buf[sizeof(PROC_OOM_SCORE_ADJ_PATH) + MAX_DEC_SIZE(int)] = {0};

	/* Don't touch OOM-fixed process' score */
	if (proc_priority_is_oom_fixed_process(pid))
		return RESOURCED_ERROR_NONE;

	snprintf(buf, sizeof(buf), PROC_OOM_SCORE_ADJ_PATH, pid);
	fp = fopen(buf, "r+");
	if (fp == NULL)
		return RESOURCED_ERROR_FAIL;
	if (fgets(buf, sizeof(buf), fp) == NULL) {
		fclose(fp);
		return RESOURCED_ERROR_FAIL;
	}
	fprintf(fp, "%d", oom_score_adj);
	fclose(fp);

	if (oom_score_adj >= OOMADJ_SU) {
		lowmem_data.control_type = LOWMEM_MOVE_CGROUP;
		lowmem_data.args[0] = (int)pid;
		lowmem_data.args[1] = (int)oom_score_adj;
		resourced_notify(RESOURCED_NOTIFIER_MEM_CONTROL, &lowmem_data);
	}

	return RESOURCED_ERROR_NONE;
}

int proc_get_label(pid_t pid, char *label)
{
	char buf[PROC_BUF_MAX];
	FILE *fp;

	snprintf(buf, sizeof(buf), "/proc/%d/attr/current", pid);
	fp = fopen(buf, "r");
	if (fp == NULL)
		return RESOURCED_ERROR_FAIL;

	if (fgets(label, PROC_NAME_MAX-1, fp) == NULL) {
		fclose(fp);
		return RESOURCED_ERROR_FAIL;
	}
	fclose(fp);
	return RESOURCED_ERROR_NONE;
}

int proc_get_mem_status(pid_t pid, unsigned int *vmswap, unsigned int *vmrss)
{
	char filename[PROC_BUF_MAX];
	_cleanup_fclose_ FILE *fp = NULL;
	unsigned int swap = 0, rss = 0;

	snprintf(filename, PROC_BUF_MAX, "/proc/%d/status", pid);
	fp = fopen(filename, "r");
	if (!fp)
		return RESOURCED_ERROR_FAIL;

	if (vmrss != NULL) {
		while (fgets(filename, sizeof(filename), fp)) {
			/* Skip the lines, until first match */
			if (!strstart_with(filename, "VmRSS:"))
				continue;

			/* Read RSS value and end this loop. */
			if (sscanf(filename, "VmRSS: %d kB", &rss) == 1)
				break;

			return RESOURCED_ERROR_NO_DATA;
		}
		*vmrss = rss;
	}

	if (vmswap != NULL) {
		/* Interate over rest of Vm* values */
		while (fgets(filename, sizeof(filename), fp)) {
			/* Read VmSwap and return with positive result */
			if (sscanf(filename, "VmSwap: %d kB", &swap) == 1)
				break;

			/* End of file before VmSwap read, return with error */
			if (feof(fp))
				return RESOURCED_ERROR_NO_DATA;
		}
		*vmswap = swap;
	}

	return RESOURCED_ERROR_NONE;
}

int proc_get_pss(pid_t pid, unsigned int *pss)
{
	_cleanup_smaps_free_ struct smaps *smaps = NULL;
	int ret;

	*pss = 0;

	/*
	 * PSS memory usage is number of KB accounted to a process
	 * where shared libraries are included.
	 * All libraries that are shared are accounted by formula
	 * "library size"/"number of processes that are using it"
	 *
	 * PSS provide more real live memory usage of non swapped process.
	 */
	ret = smaps_get(pid, &smaps, SMAPS_MASK_PSS);

	if (ret < 0) {
		_E("Error while reading smaps or totmaps for pid %d", pid);
		return RESOURCED_ERROR_FAIL;
	}

	*pss += smaps->sum[SMAPS_ID_PSS];

	return RESOURCED_ERROR_NONE;
}

int proc_get_uss(pid_t pid, unsigned int *uss)
{
	_cleanup_smaps_free_ struct smaps *smaps = NULL;
	int ret;

	*uss = 0;

	/*
	 * USS memory usage is number of KB accounted to a process
	 * where only private (non-shared) data are accounted.
	 *
	 * USS provides the value how much memory will be freed after
	 * termination of process.
	 */
	ret = smaps_get(pid, &smaps,
			(SMAPS_MASK_PRIVATE_CLEAN | SMAPS_MASK_PRIVATE_DIRTY));

	if (ret < 0) {
		_E("Error while reading smaps or totmaps for pid %d", pid);
		return RESOURCED_ERROR_FAIL;
	}

	*uss += smaps->sum[SMAPS_ID_PRIVATE_CLEAN];
	*uss += smaps->sum[SMAPS_ID_PRIVATE_DIRTY];

	return RESOURCED_ERROR_NONE;
}

int proc_get_zram_usage(pid_t pid, unsigned int *usage)
{
	int ret;
	struct meminfo mi;
	static unsigned int swap_total = 0;
	unsigned int proc_swap_usage, zram_usage;

	/* Read total swap size just once and cache it */
	if (!swap_total) {
		ret = proc_get_meminfo(&mi, MEMINFO_MASK_SWAP_TOTAL);
		if (ret < 0) {
			_E("Failed to get %s: %m",
			   meminfo_id_to_string(MEMINFO_ID_SWAP_TOTAL));
			return RESOURCED_ERROR_FAIL;
		}
		swap_total = mi.value[MEMINFO_ID_SWAP_TOTAL];
	}

	/* Read usage of Swap (VmSwap) of interested process */
	ret = proc_get_mem_status(pid, &proc_swap_usage, NULL);
	if (ret != RESOURCED_ERROR_NONE)
		return ret;

	/* Read current total memory usage of zram device */
	ret = fread_nth_uint(SWAP_ZRAM_SYSFILE"mm_stat", 2, &zram_usage);
	if (ret == -ENOENT) {
		ret = fread_uint(SWAP_ZRAM_SYSFILE"mem_used_total", &zram_usage);
	}

	if (ret < 0)
		return RESOURCED_ERROR_FAIL;

	/*
	 * Calculate aproximated value of zram usage for selected process
	 * by formula: proc_zram_usage = ( VmSwap x ZramMemoryUsage )/SwapTotal
	 */
	*usage = (int)((float)proc_swap_usage * BYTE_TO_KBYTE(zram_usage) / swap_total);

	return RESOURCED_ERROR_NONE;
}

int proc_get_approx_mem_usage(pid_t pid, unsigned int *usage)
{
	int ret;
	unsigned long resident = 0, shared = 0;
	char filename[PROC_BUF_MAX];
	_cleanup_fclose_ FILE *fp = NULL;

	snprintf(filename, PROC_BUF_MAX, "/proc/%d/statm", pid);
	fp = fopen(filename, "r");
	if (!fp)
		return RESOURCED_ERROR_FAIL;

	/*
	 * For quick read, open code by putting numbers directly
	 * expected format is
	 * seq_printf(m, "%lu %lu %lu %lu 0 %lu 0\n",
	 *               size, resident, shared, text, data);
	 */
	ret = fscanf(fp, "%*s %lu %lu %*s %*s %*s %*s\n", &resident, &shared);
	if (ret < 0)
		return RESOURCED_ERROR_FAIL;

	/*
	 * The value resident - shared is mostly similar to Uss.
	 */
	*usage = BYTE_TO_KBYTE((resident - shared) << PAGE_SHIFT);
	return RESOURCED_ERROR_NONE;
}

/**
 * @desc get total memory usage from VmSwap and VmRSS.
 * @return negative value if error or pid doesn't exist
 */
int proc_get_mem_usage(pid_t pid, unsigned int *usage)
{
	int ret;
	unsigned int vmswap = 0, total = 0;

	ret = proc_get_approx_mem_usage(pid, &total);
	if (ret < 0) {
		_E("Failed to get usage : %d", pid);
		return ret;
	}

	if (swap_get_state() == SWAP_ON) {
		ret = proc_get_mem_status(pid, &vmswap, NULL);
		if (ret != RESOURCED_ERROR_NONE)
			goto out;

		total += vmswap;
	}
out:
	*usage = total;
	return RESOURCED_ERROR_NONE;
}

/**
 * @desc get how much ram is used in each application
 * @return negative value if error or pid doesn't exist
 */
int proc_get_ram_usage(pid_t pid, unsigned int *usage)
{
	int ret;
	unsigned int vmswap = 0, total = 0;

	ret = proc_get_approx_mem_usage(pid, &total);
	if (ret < 0) {
		_E("Failed to get usage : %d", pid);
		return ret;
	}

	if (swap_get_state() == SWAP_ON) {
		ret = proc_get_mem_status(pid, &vmswap, NULL);
		if (ret != RESOURCED_ERROR_NONE)
			goto out;

		/*
		 * If it is necessary to know real ram size about each application,
		 * it should consider compression ratio.
		 */
		vmswap *= MEM_SWAP_RATIO;
		total += vmswap;
	}
out:
	*usage = total;
	return RESOURCED_ERROR_NONE;
}

unsigned int proc_get_mem_available(void)
{
	struct meminfo mi;
	int ret;

	ret = proc_get_meminfo(&mi, MEMINFO_MASK_MEM_AVAILABLE);
	if (ret < 0) {
		_E("Failed to get %s: %m",
				meminfo_id_to_string(MEMINFO_ID_MEM_AVAILABLE));
		return 0;
	}

	return KBYTE_TO_MBYTE(mi.value[MEMINFO_ID_MEM_AVAILABLE]);
}

unsigned int proc_get_swap_free(void)
{
	struct meminfo mi;
	int ret;

	ret = proc_get_meminfo(&mi, MEMINFO_MASK_SWAP_FREE);
	if (ret < 0) {
		_E("Failed to get %s: %m",
		   meminfo_id_to_string(MEMINFO_ID_SWAP_FREE));
		return 0;
	}

	return mi.value[MEMINFO_ID_SWAP_FREE];
}

unsigned int proc_get_swap_total(void)
{
	struct meminfo mi;
	int ret;

	ret = proc_get_meminfo(&mi, MEMINFO_MASK_SWAP_TOTAL);
	if (ret < 0) {
		_E("Failed to get %s: %m",
		   meminfo_id_to_string(MEMINFO_ID_SWAP_TOTAL));
		return 0;
	}

	return mi.value[MEMINFO_ID_SWAP_TOTAL];
}

int proc_get_cpu_time(pid_t pid, unsigned long *utime,
		unsigned long *stime, unsigned long *starttime)
{
	char proc_path[sizeof(PROC_STAT_PATH) + MAX_DEC_SIZE(int)];
	_cleanup_fclose_ FILE *fp = NULL;

	assert(utime != NULL);
	assert(stime != NULL);

	snprintf(proc_path, sizeof(proc_path), PROC_STAT_PATH, pid);
	fp = fopen(proc_path, "r");
	if (fp == NULL)
		return RESOURCED_ERROR_FAIL;

	if (fscanf(fp, "%*s %*s %*s %*s %*s %*s %*s %*s %*s %*s %*s %*s %*s") < 0)
		return RESOURCED_ERROR_FAIL;

	if (fscanf(fp, "%lu %lu", utime, stime) < 1)
		return RESOURCED_ERROR_FAIL;

	if (fscanf(fp, "%*s %*s %*s %*s %*s %*s") < 0)
		return RESOURCED_ERROR_FAIL;

	if (fscanf(fp, "%lu", starttime) < 1)
		return RESOURCED_ERROR_FAIL;

	return RESOURCED_ERROR_NONE;
}

unsigned int proc_get_cpu_number(void)
{
	char buf[PATH_MAX];
	FILE *fp;
	int cpu = 0;

	fp = fopen("/proc/cpuinfo", "r");

	if (!fp) {
		_E("/proc/cpuinfo open failed");
		return RESOURCED_ERROR_FAIL;
	}

	while (fgets(buf, PATH_MAX, fp) != NULL) {
		if (!strncmp(buf, "processor", 9))
			cpu++;
	}

	fclose(fp);
	return cpu;
}

int proc_get_exepath(pid_t pid, char *buf, int len)
{
	char path[PROC_BUF_MAX];
	int ret = 0;

	snprintf(path, sizeof(path), "/proc/%d/exe", pid);
	ret = readlink(path, buf, len-1);
	if (ret > 0)
		buf[ret] = '\0';
	else
		buf[0] = '\0';
	return RESOURCED_ERROR_NONE;
}

static int proc_get_data(char *path, char *buf, int len)
{
	_cleanup_close_ int fd = -1;
	int ret;

	fd = open(path, O_RDONLY);
	if (fd < 0)
		return RESOURCED_ERROR_FAIL;

	ret = read(fd, buf, len-1);
	if (ret < 0) {
		buf[0] = '\0';
		return RESOURCED_ERROR_FAIL;
	}
	buf[ret] = '\0';
	return RESOURCED_ERROR_NONE;
}

int proc_get_raw_cmdline(pid_t pid, char *buf, int len)
{
	char path[PROC_BUF_MAX];
	snprintf(path, sizeof(path), "/proc/%d/cmdline", pid);
	return proc_get_data(path, buf, len);
}

int proc_get_stat(pid_t pid, char *buf, int len)
{
	char path[PROC_BUF_MAX];
	snprintf(path, sizeof(path), "/proc/%d/stat", pid);
	return proc_get_data(path, buf, len);
}

int proc_get_status(pid_t pid, char *buf, int len)
{
	char path[PROC_BUF_MAX];
	snprintf(path, sizeof(path), "/proc/%d/status", pid);
	return proc_get_data(path, buf, len);
}

int proc_sys_node_trigger(enum sys_node_id sys_node_id)
{
	FILE *fp = NULL;

	if (sys_node_id >= ARRAY_SIZE(sys_node_tables)) {
		_E("sys_node_id[%d] is out of range.\n", sys_node_id);
		return RESOURCED_ERROR_FAIL;
	}
	if (!sys_node_tables[sys_node_id].valid) {
		_E("sys_node_id[%d] is not valid.\n", sys_node_id);
		return RESOURCED_ERROR_FAIL;
	}

	/* open and check if the path exists, else return fail */
	fp = fopen(sys_node_tables[sys_node_id].path, "w");
	if (fp == NULL) {
		_E("Failed to open %s: %m\n",
			sys_node_tables[sys_node_id].path);
		sys_node_tables[sys_node_id].valid = 0;
		return RESOURCED_ERROR_FAIL;
	}
	fputc(sys_node_tables[sys_node_id].value, fp);
	fclose(fp);
	return RESOURCED_ERROR_NONE;
}

static const char* const meminfo_string_lookup[MEMINFO_ID_MAX] = {
	[MEMINFO_ID_MEM_TOTAL]	   = "MemTotal",
	[MEMINFO_ID_MEM_FREE]	   = "MemFree",
	[MEMINFO_ID_MEM_AVAILABLE] = "MemAvailable",
	[MEMINFO_ID_BUFFERS]	   = "Buffers",
	[MEMINFO_ID_CACHED]	   = "Cached",
	[MEMINFO_ID_SWAP_CACHED]   = "SwapCached",
	[MEMINFO_ID_ACTIVE]	   = "Active",
	[MEMINFO_ID_INACTIVE]	   = "Inactive",
	[MEMINFO_ID_ACTIVE_ANON]   = "Active(anon)",
	[MEMINFO_ID_INACTIVE_ANON] = "Inactive(anon)",
	[MEMINFO_ID_ACTIVE_FILE]   = "Active(file)",
	[MEMINFO_ID_INACTIVE_FILE] = "Inactive(file)",
	[MEMINFO_ID_UNEVICTABLE]   = "Unevictable",
	[MEMINFO_ID_MLOCKED]	   = "Mlocked",
	[MEMINFO_ID_HIGH_TOTAL]	   = "HighTotal",
	[MEMINFO_ID_HIGH_FREE]	   = "HighFree",
	[MEMINFO_ID_LOW_TOTAL]	   = "LowTotal",
	[MEMINFO_ID_LOW_FREE]	   = "LowFree",
	[MEMINFO_ID_SWAP_TOTAL]	   = "SwapTotal",
	[MEMINFO_ID_SWAP_FREE]	   = "SwapFree",
	[MEMINFO_ID_DIRTY]	   = "Dirty",
	[MEMINFO_ID_WRITEBACK]	   = "Writeback",
	[MEMINFO_ID_ANON_PAGES]	   = "AnonPages",
	[MEMINFO_ID_MAPPED]	   = "Mapped",
	[MEMINFO_ID_SHMEM]	   = "Shmem",
	[MEMINFO_ID_SLAB]	   = "Slab",
	[MEMINFO_ID_SRECLAIMABLE]  = "SReclaimable",
	[MEMINFO_ID_SUNRECLAIM]	   = "SUnreclaim",
	[MEMINFO_ID_KERNEL_STACK]  = "KernelStack",
	[MEMINFO_ID_PAGE_TABLES]   = "PageTables",
	[MEMINFO_ID_NFS_UNSTABLE]  = "NFS_Unstable",
	[MEMINFO_ID_BOUNCE]	   = "Bounce",
	[MEMINFO_ID_WRITEBACK_TMP] = "WritebackTmp",
	[MEMINFO_ID_COMMIT_LIMIT]  = "CommitLimit",
	[MEMINFO_ID_COMMITTED_AS]  = "Committed_AS",
	[MEMINFO_ID_VMALLOC_TOTAL] = "VmallocTotal",
	[MEMINFO_ID_VMALLOC_USED]  = "VmallocUsed",
	[MEMINFO_ID_VMALLOC_CHUNK] = "VmallocChunk",
};

const char *meminfo_id_to_string(enum meminfo_id id)
{
	assert(id >= 0 && id < MEMINFO_ID_MAX);

	return meminfo_string_lookup[id];
}

int proc_get_meminfo(struct meminfo *mi, enum meminfo_mask mask)
{
	_cleanup_fclose_ FILE *f = NULL;
	enum meminfo_mask remain_mask = mask;
	char buf[LINE_MAX];

	assert(mi);

	memset(mi, 0x0, sizeof(struct meminfo));

	f = fopen("/proc/meminfo", "r");
	if (!f)
		return -errno;

	if (remain_mask & MEMINFO_MASK_MEM_AVAILABLE)
		remain_mask |= (MEMINFO_MASK_MEM_FREE |
				MEMINFO_MASK_CACHED);

	while (remain_mask) {
		_cleanup_free_ char *k = NULL;
		unsigned int v = 0;
		enum meminfo_id id;
		size_t l;

		if (!fgets(buf, sizeof(buf), f)) {
			if (ferror(f))
				return -errno;
			break;
		}

		l = strcspn(buf, ":");
		if (!l)
			break;

		k = strndup(buf, l);
		if (!k)
			return -errno;

		id = meminfo_string_to_id(k);
		if (id < 0 || id >= MEMINFO_ID_MAX)
			continue;

		if (!(remain_mask & (1ULL << id)))
			continue;

		remain_mask &= ~((1ULL << id));

		if (sscanf(buf + l + 1, "%d", &v) != 1)
			break;

		mi->value[id] = v;
	}

	if (remain_mask & MEMINFO_MASK_MEM_AVAILABLE) {
		mi->value[MEMINFO_ID_MEM_AVAILABLE] =
			mi->value[MEMINFO_ID_MEM_FREE]
			+ mi->value[MEMINFO_ID_CACHED];

		remain_mask &= ~MEMINFO_MASK_MEM_AVAILABLE;
	}

	if (remain_mask) {
		enum meminfo_id i;

		for (i = 0; i < MEMINFO_ID_MAX; i++)
			if (remain_mask & (1ULL << i))
				_E("Failed to get meminfo: '%s'",
				   meminfo_id_to_string(i));
	}

	return RESOURCED_ERROR_NONE;
}

static inline char *proc_skip_blanks(const char *s)
{
	while (s && (isblank(*s) || !isalnum(*s)))
		++s;
	return (char*)s;
}

#define PROC_MAX_REG_MATCH	6

static int proc_find_match(char *s, const struct parse_arg *parse,
				void *data, size_t *parsed)
{
	regex_t regex;
	regmatch_t match[PROC_MAX_REG_MATCH];
	int result;

	if (!s || !(*s != '\0'))
		return -EINVAL;

	if (!parse->callback)
		return 0;

	*parsed = 0;

	result = regcomp(&regex, parse->re_exp, REG_EXTENDED);
	if (!result) {
		result = regexec(&regex, s, ARRAY_SIZE(match), match, 0);
		if (!result) {
			regmatch_t *m = match;

			result =  parse->callback(s, match, parse->tag, data);
			if (result)
				goto leave;
			while ((m - match) < PROC_MAX_REG_MATCH &&
			    m->rm_so >= 0 && m->rm_eo > 0)
				++m;

			*parsed = m > match ? (--m)->rm_eo + 1 : 0;
		}
leave:
		regfree(&regex);
	}
	return result;
}

/**
 * @brief  Rather basic parser that relies on provided arguments,
 *         specifying the actual expression to look for.
 *         Note that it does not make any assumptions as to how
 *         the input to be parsed is actual formated and  what is the
 *         layout of the information held within it.
 *         - mind it as badly specified arguments might result in a failure
 *         to identify potential match or finding a match but not one that
 *         has been expected. It is up to the caller to perform
 *         the actual formatting of data.
 *         Parsing is done per line with possible multiple regexes.
 * @return Returns number of found matches against supplied parse arguments
 */
static int proc_parse_single(const char *buffer,
			     const struct parse_arg *parse_args,
			     void *data)
{
	char *s = (char*)buffer;
	size_t size;
	unsigned int match = 0;

	while (s && *s != '\n' && *s != '\0') {
		const struct parse_arg *p = parse_args;

		size = 0;
		s = proc_skip_blanks(s);

		while (proc_find_match(s, p, data, &size))
			++p;

		if (!size)
			break;
		s += size;
		++match;
	}
	return match;
}

/**
 * @brief Simplified procfs parser
 *
 *  Reads the procfs entry line by line triggering regex match
 *  function for each supplied argument.
 *  Note: The set of arguments should end with an empty one.
 *
 * @return returns 0 upon successfully calling the actual parsing
 *	   procedure at least once, error code otherwise.
 *	   Note: it is up to the caller to properly handle
 *	   callbacks triggered upon each match found and to determine
 *	   if parsing itself actually succeeded.
 */
int proc_parse_entry(const char *path, const struct parse_arg *parse_args,
			void *data)
{
	char buf[LINE_MAX];
	_cleanup_fclose_ FILE *f = NULL;

	f = fopen(path, "r");
	if (!f)
		return -errno;

	while (fgets(buf, sizeof(buf), f))
		proc_parse_single(buf, parse_args, data);
	return ferror(f) ? -errno : 0;
}

int proc_get_uptime(unsigned long *uptime)
{
	_cleanup_fclose_ FILE *fp = NULL;
	double stime;

	fp = fopen("/proc/uptime", "r");
	if (fp == NULL)
		return RESOURCED_ERROR_FAIL;

	if (fscanf(fp, "%lf %*s", &stime) < 0)
		return RESOURCED_ERROR_FAIL;

	*uptime = (unsigned long)stime;
	return RESOURCED_ERROR_NONE;
}

/**
 * @brief print simple memory info with rss and memtotal
 *
 *  Reads the procfs about all process id and print RSS usage.
 *  Note: If fp is valid, the result is written by desired file.
 *            Otherwise, it printed message through dlog.
 */
void proc_print_meninfo(FILE *fp)
{
	_cleanup_closedir_ DIR *dir = NULL;
	struct dirent *de;
	pid_t pid, caller;
	char cmdline[PATH_MAX];
	_cleanup_fclose_ FILE *output_file = NULL;
	int oom_score_adj, ret;
	unsigned int rss, swap;
	struct meminfo mi;
	unsigned int free = 0;
	unsigned int total_mem = 0, available = 0, used;
	unsigned int swap_total = 0, swap_free = 0, zram_used, swap_used;

	dir = opendir("/proc");
	if (dir == NULL) {
		_E("cannot read directory /proc.");
		return;
	}

	LOG_DUMP(fp, "     PID    RSS    SWAP    OOM_SCORE    COMMAND\n");

	caller = getpid();
	while ((de = readdir(dir)) != NULL) {
		pid = atoi(de->d_name);

		/* exclude negative value which means string value and kernel thread */
		if (pid < 1 || pid == caller)
			continue;

		ret = proc_get_cmdline(pid, cmdline);
		if (ret)
			continue;

		ret = proc_get_mem_status(pid, &swap, &rss);
		if (ret)
			continue;

		ret = proc_get_oom_score_adj(pid, &oom_score_adj);
		if (ret)
			continue;

		LOG_DUMP(fp, "%8d %8u  %8u %8d          %s\n",
				pid,
				rss,
				swap,
				oom_score_adj,
				cmdline);

	} /* end of while */

	ret = proc_get_meminfo(&mi,
			(MEMINFO_MASK_MEM_TOTAL |
			 MEMINFO_MASK_MEM_FREE |
			 MEMINFO_MASK_MEM_AVAILABLE |
			 MEMINFO_MASK_CACHED |
			 MEMINFO_MASK_SWAP_TOTAL |
			 MEMINFO_MASK_SWAP_FREE));
	if (ret < 0) {
		_E("Failed to get meminfo");
		return;
	}

	total_mem = mi.value[MEMINFO_ID_MEM_TOTAL];
	free = mi.value[MEMINFO_ID_MEM_FREE];
	available = mi.value[MEMINFO_ID_MEM_AVAILABLE];
	swap_total = mi.value[MEMINFO_ID_SWAP_TOTAL];
	swap_free = mi.value[MEMINFO_ID_SWAP_FREE];

	used = total_mem - available;
	swap_used = swap_total - swap_free;

	ret = fread_nth_uint(SWAP_ZRAM_SYSFILE"mm_stat", 2, &zram_used);
	if (ret == -ENOENT) {
		ret = fread_uint(SWAP_ZRAM_SYSFILE"mem_used_total", &zram_used);
	}

	if (ret != RESOURCED_ERROR_NONE)
		zram_used = 0;

	LOG_DUMP(fp, "====================================================================\n");
	LOG_DUMP(fp, "Total RAM size: \t%15d MB( %6d kB)\n",
			KBYTE_TO_MBYTE(total_mem), total_mem);
	LOG_DUMP(fp, "Used (Mem+Reclaimable): %15d MB( %6d kB)\n",
			KBYTE_TO_MBYTE(total_mem - free), total_mem - free);
	LOG_DUMP(fp, "Used (Mem+Swap): \t%15d MB( %6d kB)\n",
			KBYTE_TO_MBYTE(used), used);
	LOG_DUMP(fp, "Used (Mem):  \t\t%15d MB( %6d kB)\n",
			KBYTE_TO_MBYTE(used), used);
	LOG_DUMP(fp, "Used (Swap): \t\t%15d MB( %6d kB)\n",
			KBYTE_TO_MBYTE(swap_used), swap_used);
	LOG_DUMP(fp, "Used (Zram block device): %13d MB( %6d kB)\n",
			BYTE_TO_MBYTE(zram_used), BYTE_TO_KBYTE(zram_used));
	LOG_DUMP(fp, "Mem Free:\t\t%15d MB( %6d kB)\n",
			KBYTE_TO_MBYTE(free), free);
	LOG_DUMP(fp,  "Available (Free+Reclaimable):%10d MB( %6d kB)\n",
			KBYTE_TO_MBYTE(available), available);
	return;
}

int proc_get_buddyinfo(const char *zone, struct buddyinfo *bi)
{
	_cleanup_fclose_ FILE *f = NULL;
	char buf[LINE_MAX];

	g_assert(zone);
	g_assert(bi);

	f = fopen("/proc/buddyinfo", "re");
	if (!f)
		return -errno;

	for (;;) {
		int n;
		char zonename[32] = { 0, };

		if (!fgets(buf, sizeof(buf), f)) {
			if (ferror(f))
				return -errno;

			break;
		}

		n = sscanf(buf, "Node %d, zone %31s %d %d %d %d %d %d %d %d %d %d %d",
			   &bi->node,
			   zonename,
			   &bi->page[PAGE_4K],
			   &bi->page[PAGE_8K],
			   &bi->page[PAGE_16K],
			   &bi->page[PAGE_32K],
			   &bi->page[PAGE_64K],
			   &bi->page[PAGE_128K],
			   &bi->page[PAGE_256K],
			   &bi->page[PAGE_512K],
			   &bi->page[PAGE_1M],
			   &bi->page[PAGE_2M],
			   &bi->page[PAGE_4M]);
		if (n != 13)
			break;

		if (!streq(zone, zonename))
			continue;

		return 0;
	}

	return -ENODATA;
}

void proc_swap_free(struct proc_swaps *swap)
{
	if (!swap)
		return;

	free(swap->filename);
	free(swap->type);
	free(swap);
}

void proc_swaps_free(struct proc_swaps **swaps)
{
	int i;

	if (!swaps)
		return;

	for (i = 0; swaps[i]; i++)
		proc_swap_free(swaps[i]);

	free(swaps);
}

int proc_get_swaps(struct proc_swaps ***swaps)
{
	_cleanup_proc_swaps_free_ struct proc_swaps **ss = NULL;
	_cleanup_fclose_ FILE *f = NULL;
	char buf[LINE_MAX];
	size_t cnt = 0;

	assert(swaps);
	assert(!*swaps);

	f = fopen("/proc/swaps", "re");
	if (!f)
		return -errno;

	for (;;) {
		_cleanup_proc_swap_free_ struct proc_swaps *swap = NULL;

		if (!fgets(buf, sizeof(buf), f)) {
			if (ferror(f))
				return -errno;

			break;
		}

		swap = calloc(sizeof(struct proc_swaps), 1);
		if (!swap)
			return -ENOMEM;

		if (sscanf(buf, "%ms %ms %u %u %d",
			   &swap->filename,
			   &swap->type,
			   &swap->size,
			   &swap->used,
			   &swap->priority) != 5)
			continue;

		ss = (struct proc_swaps **)realloc(ss, sizeof(struct proc_swaps *) * (cnt + 2));
		if (!ss)
			return -ENOMEM;

		ss[cnt++] = swap;
		ss[cnt] = NULL;
		swap = NULL;
	}

	*swaps = ss;
	ss = NULL;

	return cnt;
}
