/*
 * resourced
 *
 * Copyright (c) 2016 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
*/

#include <pthread.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "macro.h"
#include "util.h"
#include "module.h"
#include "resourced.h"
#include "trace.h"
#include "resourced-helper-worker.h"
#include "fd-handler.h"

/* resourced helper worker */
static struct resourced_worker rhw;
static int pfds[2];
static fd_handler_h phandler;

#define PTHREAD_WORKER_IS_ACTIVE(_pthw)	 g_atomic_int_get(&(_pthw)->active)
#define PTHREAD_WORKER_ACTIVATE(_pthw)	g_atomic_int_set(&(_pthw)->active, 1)
#define PTHREAD_WORKER_DEACTIVATE(_pthw)	g_atomic_int_set(&(_pthw)->active, 0)

#define PTHREAD_WORKER_RUN(_pthw)	g_atomic_int_set(&(_pthw)->running, 1)
#define PTHREAD_WORKER_IDLE(_pthw)	g_atomic_int_set(&(_pthw)->running, 0)

struct pipe_data {
	enum helper_worker_operation op;
	struct resourced_helper_data *rhd;
	int ret;
};

static void resourced_helper_free(void *arg)
{
	if (arg)
		free(arg);
}

static void resourced_helper_drain_queue(struct resourced_worker *rw)
{
	struct resourced_helper_data *rhd;

	g_async_queue_lock(rw->queue);
	while ((rhd = g_async_queue_try_pop_unlocked(rw->queue)))
		resourced_helper_free((void *)rhd);

	g_async_queue_unlock(rw->queue);
}

static int trigger_pipe(struct resourced_helper_data *rhd, int ret)
{
	struct pipe_data pdata = { rhd->op, rhd, ret};
	int n;

	n = write(pfds[1], &pdata, sizeof(struct pipe_data));

	return (n != sizeof(struct pipe_data)) ? -EPERM : 0;
}

static bool receive_pipe_cb(int fd, void *data)
{
	struct pipe_data pdata = {0,};
	int n;
	struct resourced_helper_data *rhd;

	n = read(fd, &pdata, sizeof(pdata));
	if (n != sizeof(pdata) || !pdata.rhd) {
		_E("fail to read struct pipe data");
		goto out;
	}

	rhd = pdata.rhd;
	if (rhd->callback)
		rhd->callback(rhd->data, pdata.ret);

	resourced_helper_free(rhd);
out:
	return true;

}


static int pipe_init(void)
{
	int ret;

	ret = pipe2(pfds, O_CLOEXEC);
	if (ret == -1)
		return -errno;

	add_fd_read_handler(pfds[0], receive_pipe_cb, NULL, NULL, &phandler);
	if (!phandler)
		return -EPERM;

	return 0;
}

static void pipe_exit(void)
{
	if (phandler) {
		remove_fd_read_handler(&phandler);
		phandler = NULL;
	}

	if (pfds[0])
		close(pfds[0]);
	if (pfds[1])
		close(pfds[1]);
}

static void *resourced_helper_worker(void *arg)
{
	struct resourced_worker *rw = (struct resourced_worker *)arg;
	int ret = 0;

	g_async_queue_ref(rw->queue);

	while (1) {
		struct resourced_helper_data *rhd;

		if (!PTHREAD_WORKER_IS_ACTIVE(rw))
			break;

		rhd = (struct resourced_helper_data *)g_async_queue_pop(rw->queue);
		PTHREAD_WORKER_RUN(rw);

		if (rhd->helper_func)
			ret = rhd->helper_func(rhd->data);
		if (rhd->callback)
			trigger_pipe(rhd, ret);
		else
			resourced_helper_free(rhd);

		PTHREAD_WORKER_IDLE(rw);
	}
	g_async_queue_unref(rw->queue);
	pthread_exit(NULL);
}

static void queue_request(struct resourced_worker *worker, void *req)
{
	if (PTHREAD_WORKER_IS_ACTIVE(worker))
		g_async_queue_push(worker->queue, req);
}

int request_helper_worker(enum helper_worker_operation op,
	void *data, int (*function) (void *data), void *callback)
{
	struct resourced_helper_data *rhd;

	rhd = calloc(1, sizeof(struct resourced_helper_data));
	if (!rhd)
		return RESOURCED_ERROR_OUT_OF_MEMORY;

	rhd->op = op;
	rhd->data = data;
	rhd->helper_func = function;
	rhd->callback = callback;

	queue_request(&rhw, rhd);

	return RESOURCED_ERROR_NONE;
}

static int resourced_worker_init(void *data)
{
	int ret = RESOURCED_ERROR_NONE;

	if (PTHREAD_WORKER_IS_ACTIVE(&rhw)) {
		_I("helper worker thread [%d] has already been created\n",
				(unsigned)rhw.worker_thread);
		return ret;
	}
	pipe_init();
	rhw.queue = g_async_queue_new_full(resourced_helper_free);

	if (!rhw.queue) {
		_E("Failed to create request queue\n");
		return RESOURCED_ERROR_FAIL;
	}
	PTHREAD_WORKER_ACTIVATE(&rhw);
	ret = pthread_create(&rhw.worker_thread, NULL,
		(void *)resourced_helper_worker, (void *)&rhw);

	if (ret) {
		PTHREAD_WORKER_DEACTIVATE(&rhw);
		_E("Failed to create resourced herper worker thread: %d\n", ret);
	} else {
		pthread_detach(rhw.worker_thread);
		ret = RESOURCED_ERROR_NONE;
	}
	return ret;
}

static int resourced_worker_finalize(void *data)
{
	if (!PTHREAD_WORKER_IS_ACTIVE(&rhw))
		return RESOURCED_ERROR_NONE;

	pipe_exit();
	PTHREAD_WORKER_DEACTIVATE(&rhw);
	resourced_helper_drain_queue(&rhw);
	g_async_queue_unref(rhw.queue);
	pthread_cancel(rhw.worker_thread);
	return RESOURCED_ERROR_NONE;
}

static struct module_ops resourced_helper_worker_ops = {
	.priority = MODULE_PRIORITY_HIGH,
	.name = "resourced-helper-worker",
	.init = resourced_worker_init,
	.exit = resourced_worker_finalize,
};

MODULE_REGISTER(&resourced_helper_worker_ops)
