/*
 * resourced
 *
 * Copyright (c) 2014 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * @file heart-cpu.c
 *
 * @desc heart cpu module
 *
 * Copyright (c) 2015 Samsung Electronics Co., Ltd. All rights reserved.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <glib.h>
#include <math.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "proc-common.h"
#include "notifier.h"
#include "resourced.h"
#include "dbus-handler.h"
#include "heart.h"
#include "logging.h"
#include "heart-common.h"
#include "trace.h"
#include "module.h"
#include "macro.h"
#include "userinfo-list.h"
#include "proc-main.h"

#define PROC_PATH				"/proc/%d"
#define PROC_STAT_PATH				"/proc/%d/stat"
#define CPU_NAME				"cpu"
#define CPU_DATA_MAX				1024
#define CPU_ARRAY_MAX				24
#define HEART_CPU_SAVE_INTERVAL			3600
#define HEART_CPU_DATA_FILE			HEART_USER_FILE_PATH"/.cpu.dat"

enum {
	SERVICE = 0,
	FOREG = 1,
	BACKG = 2
};

struct heart_cpu_info {
	unsigned long utime;
	unsigned long stime;
	int state;
	pid_t pid;
};

struct heart_cpu_table {
	char appid[MAX_APPID_LENGTH];
	char pkgid[MAX_PKGNAME_LENGTH];
	unsigned long total_utime;
	unsigned long total_stime;
	unsigned long utime;
	unsigned long stime;
	int fg_count;
	unsigned long fg_time;
	unsigned long bg_time;
	GSList *last_pid_info;
	pid_t last_pid;
	time_t last_renew_time;
	GArray *cpu_info;
};

struct heart_cpu_dat_cache {
	char *path;
	GHashTable *list;
	time_t last_file_commit_time;
};

static GHashTable *heart_cpu_app_list;
static pthread_mutex_t heart_cpu_mutex = PTHREAD_MUTEX_INITIALIZER;

static void heart_cpu_remove_last_pid_info_exited(struct heart_cpu_table *table)
{
	char proc_path[sizeof(PROC_PATH) + MAX_DEC_SIZE(int)];
	GSList *iter, *next;
	struct heart_cpu_info *ci = NULL;

	if (!table || !table->last_pid_info)
		return;

	gslist_for_each_safe(table->last_pid_info, iter, next, ci) {
		snprintf(proc_path, sizeof(proc_path), PROC_PATH, ci->pid);
		if (!access(proc_path, F_OK))
			continue;
		table->last_pid_info = g_slist_remove(table->last_pid_info, ci);
		free(ci);
	}
}

static struct heart_cpu_info *find_pid_info(struct heart_cpu_table *table, pid_t pid)
{
	GSList *iter = NULL;
	struct heart_cpu_info *ci = NULL;

	if (!table || !table->last_pid_info)
		return NULL;

	gslist_for_each_item(iter, table->last_pid_info) {
		ci = (struct heart_cpu_info *)iter->data;
		if (ci && ci->pid == pid)
			return ci;
	}
	return NULL;
}

static int heart_cpu_get_cpu_time(pid_t pid, unsigned long *utime,
		unsigned long *stime)
{
	char proc_path[sizeof(PROC_STAT_PATH) + MAX_DEC_SIZE(int)];
	FILE *fp;

	assert(utime != NULL);
	assert(stime != NULL);

	snprintf(proc_path, sizeof(proc_path), PROC_STAT_PATH, pid);
	fp = fopen(proc_path, "r");
	if (fp == NULL)
		return RESOURCED_ERROR_FAIL;

	if (fscanf(fp, "%*s %*s %*s %*s %*s %*s %*s %*s %*s %*s %*s %*s %*s") < 0) {
		fclose(fp);
		return RESOURCED_ERROR_FAIL;
	}

	if (fscanf(fp, "%lu %lu", utime, stime) < 1) {
		fclose(fp);
		return RESOURCED_ERROR_FAIL;
	}

	fclose(fp);

	return RESOURCED_ERROR_NONE;
}

static GHashTable* heart_cpu_get_user_app_list(int uid)
{
	struct heart_cpu_dat_cache *cache;

	cache = (struct heart_cpu_dat_cache*)g_hash_table_lookup(heart_cpu_app_list, (gconstpointer)&uid);
	if (!cache) {
		_E("There is no table of uid %d", uid);
		return NULL;
	}

	return cache->list;
}

static int heart_cpu_write_data(struct proc_status *ps, pid_t pid, int type)
{
	int ret;
	unsigned long utime, stime;
	char info[CPU_DATA_MAX];
	char *appid, *pkgid;

	ret = heart_cpu_get_cpu_time(pid, &utime, &stime);
	if (ret != RESOURCED_ERROR_NONE)
		return ret;

	snprintf(info, sizeof(info), "%lu %lu %d %d ", utime, stime, pid, type);

	ret = proc_get_id_info(ps, &appid, &pkgid);
	if (ret != RESOURCED_ERROR_NONE) {
		_E("Failed to proc_get_id_info");
		return ret;
	}
	ret = logging_write(pid, CPU_NAME, appid, pkgid, time(NULL), info);
	_D("heart_cpu_write_data : pid = %d, appname = %s, pkgname = %s, type=%d",
			pid, appid, pkgid, type);
	return ret;
}

static int heart_cpu_service_launch(void *data)
{
	int ret;
	struct proc_status *ps = (struct proc_status *)data;

	ret = heart_cpu_write_data(ps, ps->pid, SERVICE);
	if (ret != RESOURCED_ERROR_NONE) {
		_E("Failed to write cpu info %d", ps->pid);
		return ret;
	}
	resourced_notify(RESOURCED_NOTIFIER_LOGGING_START, NULL);
	return RESOURCED_ERROR_NONE;
}

static int heart_cpu_foreground_state(void *data)
{
	int ret;
	struct proc_status *ps = (struct proc_status *)data;

	ret = heart_cpu_write_data(ps, ps->pid, FOREG);
	if (ret != RESOURCED_ERROR_NONE) {
		_E("Failed to write cpu info %d", ps->pid);
		return ret;
	}
	resourced_notify(RESOURCED_NOTIFIER_LOGGING_START, NULL);
	return RESOURCED_ERROR_NONE;
}

static int heart_cpu_background_state(void *data)
{
	int ret;
	GSList *giter = NULL;
	struct proc_status *ps = (struct proc_status *)data;

	ret = heart_cpu_write_data(ps, ps->pid, BACKG);
	if (ret != RESOURCED_ERROR_NONE) {
		_E("Failed to write cpu info %d", ps->pid);
		return ret;
	}
	if (!ps->pai->childs)
		return RESOURCED_ERROR_NONE;
	gslist_for_each_item(giter, ps->pai->childs) {
		pid_t child = GPOINTER_TO_PID(giter->data);
		ret = heart_cpu_write_data(ps, child, BACKG);
		if (ret != RESOURCED_ERROR_NONE) {
			_E("Failed to write child cpu info %d", child);
			return ret;
		}
	}
	resourced_notify(RESOURCED_NOTIFIER_LOGGING_START, NULL);
	return RESOURCED_ERROR_NONE;
}

static int heart_cpu_update_state(void *data)
{
	int ret, state;
	GSList *giter = NULL;
	struct proc_status *ps = (struct proc_status *)data;

	if (!ps->pai) {
		_E("Invalid parameter");
		return RESOURCED_ERROR_INVALID_PARAMETER;
	}
	if (ps->pai->lru_state == PROC_FOREGROUND)
		state = FOREG;
	else
		state = BACKG;

	ret = heart_cpu_write_data(ps, ps->pid, state);
	if (ret != RESOURCED_ERROR_NONE) {
		_E("Failed to write cpu info %d", ps->pid);
		return ret;
	}
	_D("heart_cpu_update_state : pid = %d, state = %d",
			ps->pid, state);
	if (!ps->pai->childs)
		return RESOURCED_ERROR_NONE;
	gslist_for_each_item(giter, ps->pai->childs) {
		pid_t child = GPOINTER_TO_PID(giter->data);
		ret = heart_cpu_write_data(ps, child, state);
		if (ret != RESOURCED_ERROR_NONE) {
			_E("Failed to write cpu info %d", child);
			return ret;
		}
	}
	resourced_notify(RESOURCED_NOTIFIER_LOGGING_START, NULL);
	return RESOURCED_ERROR_NONE;
}

static int heart_cpu_update_app_list(void *data)
{
	_cleanup_app_list_close_ GSList *proc_app_list = PAL_INIT_VALUE;
	GSList *giter = NULL;
	struct proc_app_info *pai = NULL;

	proc_app_list = proc_app_list_open();
	gslist_for_each_item(giter, proc_app_list) {
		struct proc_status ps;
		pai = (struct proc_app_info *)giter->data;
		if (!pai->ai)
			continue;
		ps.pid = pai->main_pid;
		ps.pai = pai;
		heart_cpu_update_state(&ps);
	}
	return RESOURCED_ERROR_NONE;
}

static void heart_cpu_free_uid(gpointer data)
{
	if (data)
		free(data);
}

static void heart_cpu_free_dat_cache(gpointer data)
{
	struct heart_cpu_dat_cache *cache = (struct heart_cpu_dat_cache*)data;
	if (cache) {
		free(cache->path);
		g_hash_table_destroy(cache->list);
		free(cache);
	}
}

static void heart_free_value(gpointer value)
{
	int i;
	struct heart_cpu_info *info;
	struct heart_cpu_table *table = (struct heart_cpu_table *)value;

	if (!table)
		return;

	if (table->last_pid_info)
		g_slist_free_full(table->last_pid_info, free);

	if (table->cpu_info) {
		for (i = 0; i < table->cpu_info->len; i++) {
			info = g_array_index(table->cpu_info,
					struct heart_cpu_info *, i);
			free(info);
		}
		g_array_free(table->cpu_info, TRUE);
	}

	free(table);
}

static int heart_cpu_read_length(char *buf, int count)
{
	int i, find = 0;
	int len = strlen(buf);

	for (i = 0; i < len; i++) {
		if (buf[i] == ' ')
			find++;
		if (find == count)
			return i + 1;
	}
	return RESOURCED_ERROR_FAIL;
}

static int heart_cpu_read_from_file(struct heart_cpu_dat_cache *cache)
{
	int i, len, ret, fg_count, state;
	unsigned long total_utime, total_stime;
	unsigned long utime, stime;
	unsigned long fg_time, bg_time;
	pid_t pid;
	FILE *fp;
	struct heart_cpu_table *table;
	char appid[MAX_APPID_LENGTH] = {0, };
	char pkgid[MAX_PKGNAME_LENGTH] = {0, };
	char buf[CPU_DATA_MAX] = {0, };

	if (!cache) {
		_E("Cache doesn't exist");
		return RESOURCED_ERROR_INVALID_PARAMETER;
	}

	snprintf(buf, CPU_DATA_MAX, HEART_CPU_DATA_FILE, cache->path);
	fp = fopen(buf, "r");
	if (!fp) {
		if (errno != ENOENT) {
			_E("Fail to open %s (%d)", buf, errno);
			return RESOURCED_ERROR_FAIL;
		}

		_D("%s doesn't exist. Make new one.", buf);
		snprintf(buf, CPU_DATA_MAX, HEART_USER_FILE_PATH, cache->path);
		ret = mkdir(buf, S_IRWXU | S_IRWXG | S_IROTH);
		if (ret != 0 && errno != EEXIST) {
			_E("Fail to create %s (%d)", buf, errno);
			return RESOURCED_ERROR_FAIL;
		}

		snprintf(buf, CPU_DATA_MAX, HEART_CPU_DATA_FILE, cache->path);
		fp = fopen(buf, "w+");
		if (!fp) {
			_E("Fail to create %s (%d)", buf, errno);
			return RESOURCED_ERROR_FAIL;
		}
		fclose(fp);
		return RESOURCED_ERROR_NONE;
	}

	while (fgets(buf, CPU_DATA_MAX, fp)) {
		table = malloc(sizeof(struct heart_cpu_table));

		if (!table) {
			_E("malloc failed");
			fclose(fp);
			return RESOURCED_ERROR_OUT_OF_MEMORY;
		}

		/* make return values */
		ret = sscanf(buf, STR_FS(MAX_APPID_LENGTH_M1)" "STR_FS(MAX_PKGNAME_LENGTH_M1)" %lu %lu %lu %lu %lu %lu %d ",
				appid, pkgid,
				&total_utime, &total_stime,
				&utime, &stime,
				&fg_time, &bg_time, &fg_count);

		if (ret <= 0) {
			_E("sscanf failed");
			free(table);
			fclose(fp);
			return RESOURCED_ERROR_FAIL;
		}

		if (snprintf(table->appid, MAX_APPID_LENGTH, "%s", appid) < 0) {
			_E("sprintf failed");
			free(table);
			fclose(fp);
			return RESOURCED_ERROR_OUT_OF_MEMORY;
		}

		if (snprintf(table->pkgid, MAX_PKGNAME_LENGTH, "%s", pkgid) < 0) {
			_E("snprintf failed");
			free(table);
			fclose(fp);
			return RESOURCED_ERROR_OUT_OF_MEMORY;
		}
		len = heart_cpu_read_length(buf, 9);
		if (len <= 0) {
			_E("sscanf failed");
			free(table);
			fclose(fp);
			return RESOURCED_ERROR_FAIL;
		}
		table->total_utime = total_utime;
		table->total_stime = total_stime;
		table->utime = utime;
		table->stime = stime;
		table->last_pid_info = NULL;
		table->last_pid = 0;
		table->fg_time = fg_time;
		table->bg_time = bg_time;
		table->fg_count = fg_count;
		table->last_renew_time = 0;
		table->cpu_info =
			g_array_new(FALSE, FALSE, sizeof(struct heart_cpu_info *));

		for (i = 0; i < CPU_ARRAY_MAX; i++) {
			struct heart_cpu_info *ci;

			ret = sscanf(buf + len, "%lu %lu %d %d ", &utime, &stime, &pid, &state);
			if (ret <= 0) {
				_E("file read fail %s", buf + len);
				free(table);
				fclose(fp);
				return RESOURCED_ERROR_FAIL;
			}
			ci = malloc(sizeof(struct heart_cpu_info));
			if (!ci) {
				free(table);
				fclose(fp);
				return RESOURCED_ERROR_OUT_OF_MEMORY;
			}
			ci->utime = utime;
			ci->stime = stime;
			ci->pid = pid;
			ci->state = state;
			len += heart_cpu_read_length(buf + len, 4);
			g_array_append_val(table->cpu_info, ci);
		}

		ret = pthread_mutex_lock(&heart_cpu_mutex);
		if (ret) {
			_E("pthread_mutex_lock() failed, %d", ret);
			g_array_free(table->cpu_info, TRUE);
			free(table);
			fclose(fp);
			return RESOURCED_ERROR_FAIL;
		}
		g_hash_table_insert(cache->list, (gpointer)table->appid, (gpointer)table);
		ret = pthread_mutex_unlock(&heart_cpu_mutex);
		if (ret) {
			_E("pthread_mutex_unlock() failed, %d", ret);
			fclose(fp);
			return RESOURCED_ERROR_FAIL;
		}
	}

	fclose(fp);

	return RESOURCED_ERROR_NONE;
}

static int heart_cpu_save_to_file(struct heart_cpu_dat_cache *cache)
{
	int i, len, ret, array_len;
	gpointer value;
	gpointer key;
	GHashTableIter iter;
	struct heart_cpu_table *table;
	FILE *fp;
	char buf[CPU_DATA_MAX] = {0, };

	if (!cache) {
		_D("Cache is empty. Do nothing.");
		return RESOURCED_ERROR_NONE;
	}

	snprintf(buf, CPU_DATA_MAX, HEART_CPU_DATA_FILE, cache->path);
	fp = fopen(buf, "w");
	if (!fp) {
		_E("%s fopen failed %d", buf, errno);
		return RESOURCED_ERROR_FAIL;
	}

	if (!(cache->list)) {
		_E("empty app list");
		fclose(fp);
		return RESOURCED_ERROR_FAIL;
	}

	if (!g_hash_table_size(cache->list)) {
		_E("hash table is empty");
		fclose(fp);
		return RESOURCED_ERROR_FAIL;
	}

	ret = pthread_mutex_lock(&heart_cpu_mutex);
	if (ret) {
		_E("pthread_mutex_lock() failed, %d", ret);
		fclose(fp);
		return RESOURCED_ERROR_FAIL;
	}

	g_hash_table_iter_init(&iter, cache->list);

	while (g_hash_table_iter_next(&iter, &key, &value)) {
		table = (struct heart_cpu_table *)value;
		array_len = table->cpu_info->len;
		len = snprintf(buf, CPU_DATA_MAX, "%s %s %lu %lu %lu %lu %lu %lu %d ",
				table->appid, table->pkgid,
				table->total_utime,
				table->total_stime,
				table->utime,
				table->stime,
				table->fg_time,
				table->bg_time,
				table->fg_count);

		for (i = 0; i < CPU_ARRAY_MAX; i++) {
			struct heart_cpu_info *ci;
			if (array_len <= i) {
				len += snprintf(buf + len, CPU_DATA_MAX - len, "0 0 0 0 ");
			} else {
				ci = g_array_index(table->cpu_info, struct heart_cpu_info *, i);
				if (!ci)
					break;
				len += snprintf(buf + len, CPU_DATA_MAX - len, "%lu %lu %d %d ",
						ci->utime,
						ci->stime,
						ci->pid,
						ci->state);
			}
		}
		snprintf(buf + len, CPU_DATA_MAX - len, "\n");
		fputs(buf, fp);
	}
	ret = pthread_mutex_unlock(&heart_cpu_mutex);
	if (ret) {
		_E("pthread_mutex_unlock() failed, %d", ret);
		fclose(fp);
		return RESOURCED_ERROR_FAIL;
	}

	fclose(fp);

	return RESOURCED_ERROR_NONE;
}

static int heart_cpu_hashtable_renew(GHashTable *hashtable, time_t now)
{
	int ret;
	gpointer value;
	gpointer key;
	GHashTableIter iter;
	struct heart_cpu_table *table;

	if (!hashtable) {
		_E("empty app list");
		return RESOURCED_ERROR_FAIL;
	}

	if (!g_hash_table_size(hashtable)) {
		_E("hash table is empty");
		return RESOURCED_ERROR_FAIL;
	}
	ret = pthread_mutex_lock(&heart_cpu_mutex);
	if (ret) {
		_E("pthread_mutex_lock() failed, %d", ret);
		return RESOURCED_ERROR_FAIL;
	}
	g_hash_table_iter_init(&iter, hashtable);

	while (g_hash_table_iter_next(&iter, &key, &value)) {
		table = (struct heart_cpu_table *)value;
		table->total_utime = 0;
		table->total_stime = 0;
		table->last_renew_time = now;
		table->fg_count = 0;
		table->fg_time = 0;
		table->bg_time = 0;
	}
	ret = pthread_mutex_unlock(&heart_cpu_mutex);
	if (ret) {
		_E("pthread_mutex_unlock() failed, %d", ret);
		return RESOURCED_ERROR_FAIL;
	}
	return RESOURCED_ERROR_NONE;
}

void heart_cpu_update(struct logging_table_form *data, void *user_data)
{
	const char *path;
	int ret;
	pid_t pid;
	uid_t *uid;
	int state;
	unsigned long utime, stime;
	unsigned long utime_diff = 0, stime_diff = 0;
	time_t curr_time = logging_get_time(CLOCK_BOOTTIME);
	struct heart_cpu_table *table;
	struct heart_cpu_info *ci = NULL;
	struct heart_cpu_dat_cache *cache;
	GHashTable *cpu_usage_list = NULL;

	cache = (struct heart_cpu_dat_cache*)g_hash_table_lookup(
			heart_cpu_app_list, (gpointer)(&(data->uid)));
	if (!cache) {
		path = userinfo_find_home_dir((uid_t)data->uid);
		if (!path) {
			_E("uid %d doesn't existed", data->uid);
			return;
		}

		cache = (struct heart_cpu_dat_cache*)malloc(sizeof(struct heart_cpu_dat_cache));
		if (!cache) {
			_E("malloc failed");
			return;
		}

		cache->path = strndup(path, strlen(path));
		if (!cache->path) {
			_E("strndup failed");
			free(cache);
			return;
		}

		cache->list = g_hash_table_new_full(
				g_str_hash,
				g_str_equal,
				NULL,
				heart_free_value);
		if (!cache->list) {
			_E("g_hash_table_new_full failed");
			free(cache->path);
			free(cache);
			return;
		}

		uid = malloc(sizeof(uid_t));
		if (!uid) {
			_E("malloc failed");
			g_hash_table_unref(cache->list);
			free(cache->path);
			free(cache);
			return;
		}
		*uid = (uid_t)data->uid;

		g_hash_table_insert(heart_cpu_app_list, (gpointer)uid, (gpointer)cache);
	}
	cpu_usage_list = cache->list;

	_D("%d %s %s %ld %s", data->uid, data->appid, data->pkgid, data->time, data->data);
	if (sscanf(data->data, "%lu %lu %d %d ", &utime, &stime, &pid, &state) < 0) {
		_E("sscanf failed");
		return;
	}

	ret = pthread_mutex_lock(&heart_cpu_mutex);
	if (ret) {
		_E("pthread_mutex_lock() failed, %d", ret);
		return;
	}
	table =
		g_hash_table_lookup(cpu_usage_list, data->appid);
	/* update */
	if (table) {

		if (table->last_renew_time > data->time)
			goto unlock_exit;

		ci = find_pid_info(table, pid);
		if (table->last_pid_info && ci) {
			utime_diff = utime - ci->utime;
			table->utime += utime_diff;
			table->total_utime += utime_diff;
			stime_diff = stime - ci->stime;
			table->stime += stime_diff;
			table->total_stime += stime_diff;
			ci->utime = utime;
			ci->stime = stime;
			if (ci->state == BACKG || ci->state == SERVICE) {
				table->bg_time += utime_diff + stime_diff;
				if (state == FOREG)
					table->fg_count++;
			} else
				table->fg_time += utime_diff + stime_diff;
			ci->state = state;
		} else {
			table->utime += utime;
			table->total_utime += utime;
			table->stime += stime;
			table->total_stime += stime;
			if (table->last_pid_info)
				heart_cpu_remove_last_pid_info_exited(table);
			ci = malloc(sizeof(struct heart_cpu_info));
			if (!ci) {
				_E("malloc failed");
				goto unlock_exit;
			}
			ci->pid = pid;
			ci->utime = utime;
			ci->stime = stime;
			ci->state = state;
			if (ci->state == FOREG)
				table->fg_count++;
			table->last_pid_info = g_slist_prepend(table->last_pid_info, ci);
			table->last_pid = pid;
		}
	} else {
		table = calloc(1, sizeof(struct heart_cpu_table));

		if (!table) {
			_E("malloc failed");
			goto unlock_exit;
		}

		if (snprintf(table->appid, MAX_APPID_LENGTH,  "%s", data->appid) < 0) {
			free(table);
			_E("snprintf failed");
			goto unlock_exit;
		}

		if (snprintf(table->pkgid, MAX_PKGNAME_LENGTH, "%s", data->pkgid) < 0) {
			free(table);
			_E("snprintf failed");
			goto unlock_exit;
		}
		table->total_utime = utime;
		table->total_stime = stime;
		table->utime = utime;
		table->stime = stime;
		table->fg_count = 0;
		table->fg_time = 0;
		table->bg_time = 0;
		if (state == FOREG)
			table->fg_count = 1;

		table->cpu_info =
			g_array_new(FALSE, FALSE, sizeof(struct heart_cpu_info *));
		if (!table->cpu_info) {
			free(table);
			_E("g_array_new failed");
			goto unlock_exit;
		}

		ci = malloc(sizeof(struct heart_cpu_info));
		if (!ci) {
			_E("malloc failed");
			free(table);
			goto unlock_exit;
		}
		ci->pid = pid;
		ci->utime = utime;
		ci->stime = stime;
		ci->state = state;
		table->last_pid_info = g_slist_prepend(NULL, ci);
		table->last_pid = pid;

		g_hash_table_insert(cpu_usage_list, (gpointer)table->appid, (gpointer)table);
	}
	ret = pthread_mutex_unlock(&heart_cpu_mutex);
	if (ret) {
		_E("pthread_mutex_unlock() failed, %d", ret);
		return;
	}

	if (cache->last_file_commit_time + HEART_CPU_SAVE_INTERVAL < curr_time) {
		/* all hash table update and make new array */
		gpointer value;
		gpointer key;
		GHashTableIter iter;
		struct heart_cpu_table *search;
		struct heart_cpu_info *ci;

		ret = pthread_mutex_lock(&heart_cpu_mutex);
		if (ret) {
			_E("pthread_mutex_lock() failed, %d", ret);
			return;
		}

		g_hash_table_iter_init(&iter, cpu_usage_list);

		while (g_hash_table_iter_next(&iter, &key, &value)) {
			search = (struct heart_cpu_table *)value;

			ci = malloc(sizeof(struct heart_cpu_info));

			if (!ci) {
				_E("malloc failed");
				goto unlock_exit;
			}
			/* make new array node */
			ci->pid = search->last_pid;
			ci->utime = search->utime;
			ci->stime = search->stime;
			search->utime = 0;
			search->stime = 0;
			/* hashtable sliding : remove last node and make new one */
			if (search->cpu_info->len == CPU_ARRAY_MAX)
				g_array_remove_index(search->cpu_info, CPU_ARRAY_MAX - 1);
			g_array_prepend_val(search->cpu_info, ci);
		}
		ret = pthread_mutex_unlock(&heart_cpu_mutex);
		if (ret) {
			_E("pthread_mutex_unlock() failed, %d", ret);
			return;
		}
		/* rewrite hashtable list file */
		ret = heart_cpu_save_to_file(cache);
		if (ret) {
			_E("save to file failed");
			goto unlock_exit;
		}

		cache->last_file_commit_time = curr_time;
	}

	return;

unlock_exit:
	ret = pthread_mutex_unlock(&heart_cpu_mutex);
	if (ret) {
		_E("pthread_mutex_unlock() failed, %d", ret);
		return;
	}
}

struct heart_cpu_data *heart_cpu_get_data(int uid, char *appid, enum heart_data_period period, struct heart_cpu_data *data)
{
	int index, i, ret;
	struct heart_cpu_table *table;
	GHashTable *hashtable;

	if (!appid) {
		_E("Wrong arguments!");
		return NULL;
	}

	hashtable = heart_cpu_get_user_app_list(uid);
	if (!hashtable) {
		_E("Fail to get app list");
		return NULL;
	}

	if (!g_hash_table_size(hashtable)) {
		_D("hash table is empty");
		return NULL;
	}

	switch (period) {
	case DATA_LATEST:
		index = 0;
		break;
	case DATA_3HOUR:
		index = 3;
		break;
	case DATA_6HOUR:
		index = 6;
		break;
	case DATA_12HOUR:
		index = 12;
		break;
	case DATA_1DAY:
		index = 24;
		break;
	default:
		_E("Wrong message arguments! %d", period);
		return NULL;
	}

	ret = pthread_mutex_lock(&heart_cpu_mutex);
	if (ret) {
		_E("pthread_mutex_lock() failed, %d", ret);
		return NULL;
	}

	table = g_hash_table_lookup(hashtable, (gconstpointer)appid);
	if (!table) {
		data = NULL;
		goto unlock_exit;
	}
	if (snprintf(data->appid, MAX_APPID_LENGTH, "%s", table->appid) < 0) {
		_E("snprintf failed");
		data = NULL;
		goto unlock_exit;
	}
	if (snprintf(data->pkgid, MAX_PKGNAME_LENGTH, "%s", table->pkgid) < 0) {
		_E("snprintf failed");
		data = NULL;
		goto unlock_exit;
	}
	if (period == DATA_LATEST) {
		data->utime = table->total_utime;
		data->stime = table->total_stime;
	} else {
		data->utime = table->utime;
		data->stime = table->stime;
		i = table->cpu_info->len;
		if (i == 0) {
			data = NULL;
			goto unlock_exit;
		}
		if (i < index)
			index = i;
		for (i = 0; i < index; i++) {
			struct heart_cpu_info *cpu_info;
			cpu_info =
				g_array_index(table->cpu_info, struct heart_cpu_info *, i);
			if (!cpu_info)
				break;
			data->utime += cpu_info->utime;
			data->stime += cpu_info->stime;
		}
	}
unlock_exit:
	ret = pthread_mutex_unlock(&heart_cpu_mutex);
	if (ret) {
		_E("pthread_mutex_unlock() failed, %d", ret);
		return NULL;
	}
	return data;
}

static int compare_usage(const struct heart_app_usage *lau_a,
	    const struct heart_app_usage *lau_b)
{
	if (lau_a->point != lau_b->point)
		return (lau_b->point - lau_a->point);

	return 0;
}

/*
 * Calculate application usage using frequency and time
 */
static double heart_cpu_get_point(int freq, int time)
{
	double weightForFrequence = 3;
	double point = 0;
	point = sqrt(time + (freq*weightForFrequence));
	return point;
}

int heart_cpu_get_table(int uid, GArray *arrays, enum heart_data_period period)
{
	int index, i, ret;
	gpointer value;
	gpointer key;
	GHashTableIter h_iter;
	struct heart_cpu_table *table;
	struct heart_cpu_data *cdata;
	GHashTable *hashtable;

	hashtable = heart_cpu_get_user_app_list(uid);
	if (!hashtable) {
		_E("Fail to get app list");
		return RESOURCED_ERROR_FAIL;
	}

	if (!g_hash_table_size(hashtable)) {
		_D("hash table is empty");
		return RESOURCED_ERROR_NO_DATA;
	}

	switch (period) {
	case DATA_LATEST:
		index = 0;
		break;
	case DATA_3HOUR:
		index = 3;
		break;
	case DATA_6HOUR:
		index = 6;
		break;
	case DATA_12HOUR:
		index = 12;
		break;
	case DATA_1DAY:
		index = 24;
		break;
	default:
		_E("Wrong message arguments! %d", period);
		return RESOURCED_ERROR_FAIL;
	}

	ret = pthread_mutex_lock(&heart_cpu_mutex);
	if (ret) {
		_E("pthread_mutex_lock() failed, %d", ret);
		return RESOURCED_ERROR_FAIL;
	}

	g_hash_table_iter_init(&h_iter, hashtable);

	while (g_hash_table_iter_next(&h_iter, &key, &value)) {

		table = (struct heart_cpu_table *)value;
		cdata = malloc(sizeof(struct heart_cpu_data));
		if (!cdata) {
			_E("malloc failed");
			goto unlock_out_of_memory_exit;
		}
		if (snprintf(cdata->appid, MAX_APPID_LENGTH, "%s", table->appid) < 0) {
			_E("snprintf failed");
			free(cdata);
			goto unlock_out_of_memory_exit;
		}
		if (snprintf(cdata->pkgid, MAX_PKGNAME_LENGTH, "%s", table->pkgid) < 0) {
			_E("snprintf failed");
			free(cdata);
			goto unlock_out_of_memory_exit;
		}
		if (period == DATA_LATEST) {
			cdata->utime = table->total_utime;
			cdata->stime = table->total_stime;
		} else {
			cdata->utime = table->utime;
			cdata->stime = table->stime;
			i = table->cpu_info->len;
			if (i == 0) {
				free(cdata);
				break;
			}
			if (i < index)
				index = i;
			for (i = 0; i < index; i++) {
				struct heart_cpu_info *cpu_info;
				cpu_info =
					g_array_index(table->cpu_info, struct heart_cpu_info *, i);
				if (!cpu_info)
					break;
				cdata->utime += cpu_info->utime;
				cdata->stime += cpu_info->stime;
			}
		}
		g_array_append_val(arrays, cdata);
	}
	ret = pthread_mutex_unlock(&heart_cpu_mutex);
	if (ret) {
		_E("pthread_mutex_unlock() failed, %d", ret);
		return RESOURCED_ERROR_FAIL;
	}
	return RESOURCED_ERROR_NONE;
unlock_out_of_memory_exit:
	ret = pthread_mutex_unlock(&heart_cpu_mutex);
	if (ret) {
		_E("pthread_mutex_unlock() failed, %d", ret);
		return RESOURCED_ERROR_FAIL;
	}
	return RESOURCED_ERROR_OUT_OF_MEMORY;
}

int heart_cpu_get_appusage_list(int uid, GHashTable *lists, int top)
{
	int index = top, i, ret;
	gpointer value;
	gpointer key;
	GHashTableIter h_iter;
	struct heart_cpu_table *table;
	struct heart_app_usage lau;
	GArray *app_lists = NULL;
	GHashTable *hashtable;

	hashtable = heart_cpu_get_user_app_list(uid);
	if (!hashtable) {
		_E("Fail to get app list");
		return RESOURCED_ERROR_FAIL;
	}

	if (!g_hash_table_size(hashtable)) {
		_E("hash table is empty");
		return RESOURCED_ERROR_FAIL;
	}

	app_lists = g_array_new(false, false, sizeof(struct heart_app_usage));
	ret = pthread_mutex_lock(&heart_cpu_mutex);
	if (ret) {
		_E("pthread_mutex_lock() failed, %d", ret);
		g_array_free(app_lists, true);
		return RESOURCED_ERROR_FAIL;
	}

	g_hash_table_iter_init(&h_iter, hashtable);

	while (g_hash_table_iter_next(&h_iter, &key, &value)) {

		table = (struct heart_cpu_table *)value;
		if (!table->fg_count)
			continue;

		lau.appid = table->appid;
		lau.pkgid = table->pkgid;
		lau.fg_count = table->fg_count;
		lau.used_time = table->fg_time;
		lau.point = (int)heart_cpu_get_point(lau.fg_count, lau.used_time);
		/*
		 * make all application lists with weighted point value excepting service application
		 */
		g_array_append_val(app_lists, lau);
	}
	ret = pthread_mutex_unlock(&heart_cpu_mutex);
	if (ret) {
		_E("pthread_mutex_unlock() failed, %d", ret);
		g_array_free(app_lists, true);
		return RESOURCED_ERROR_FAIL;
	}
	if (app_lists->len < top) {
		_I("too small data for making app usage lists");
		g_array_free(app_lists, true);
		return RESOURCED_ERROR_NO_DATA;
	}

	g_array_sort(app_lists, (GCompareFunc)compare_usage);

	if (!top)
		index = app_lists->len;

	/*
	 * replace application usage lists with sorted usage arrays
	 */
	g_hash_table_remove_all(lists);
	for (i = 0; i < index; i++) {
		struct heart_app_usage *usage = &g_array_index(app_lists, struct heart_app_usage, i);
		_D("appid : %s, point : %d", usage->appid, usage->point);
		g_hash_table_insert(lists, g_strndup(usage->appid, strlen(usage->appid)), GINT_TO_POINTER(1));
	}
	g_array_free(app_lists, true);
	return RESOURCED_ERROR_NONE;
}

static int heart_cpu_update_appid(void *data)
{
	int pid;
	char old_appid[MAX_APPID_LENGTH], new_appid[MAX_APPID_LENGTH];
	char *info;

	if (!data) {
		_E("Invalid parameter");
		return RESOURCED_ERROR_INVALID_PARAMETER;
	}

	info = (char*)data;
	if (sscanf(info, "%d "STR_FS(MAX_APPID_LENGTH_M1)" "STR_FS(MAX_APPID_LENGTH_M1),
			&pid, old_appid, new_appid) < 0) {
		_E("sscanf failed, %m");
		return RESOURCED_ERROR_FAIL;
	}

	return logging_modify_appid(CPU_NAME, old_appid, new_appid, pid);
}

static void dbus_heart_get_cpu_data(GDBusMethodInvocation *invocation, GVariant *params)
{
	int period = -1;
	char *appid = NULL;
	struct heart_cpu_data *data = NULL;

	int uid = -1;

	do_expr_unless_g_variant_get_typechecked(goto failure, params, "(i&si)", &uid, &appid, &period);
	if (uid < 0 || !appid || period < 0) {
		_E("Wrong message arguments!");
		goto failure;
	}

	data = malloc(sizeof(struct heart_cpu_data));
	if (!data) {
		_E("malloc failed");
		goto failure;
	}

	if (heart_cpu_get_data(uid, appid, period, data) == NULL)
		goto failure;

	g_dbus_method_invocation_return_value(invocation, g_variant_new("(ii)",
				data->utime, data->stime));
	free(data);
	return;

failure:
	if (data)
		free(data);
	D_BUS_REPLY_NULL(invocation);
}

static void dbus_heart_get_cpu_data_list(GDBusMethodInvocation *invocation, GVariant *params)
{
	int period = -1, index, i, ret;
	gpointer value;
	gpointer key;
	GHashTableIter h_iter;
	struct heart_cpu_table *table;

	char *appid;
	unsigned long utime, stime, ftime, total;

	int uid = -1;
	GHashTable *hashtable;

	GVariantBuilder builder, *sub_builder;

	utime = stime = ftime = total = 0;

	do_expr_unless_g_variant_get_typechecked(goto failure, params, "(ii)", &uid, &period);
	if (uid < 0 || period < 0) {
		_E("Wrong message arguments!");
		goto failure;
	}

	heart_cpu_update_app_list(NULL);

	logging_save_to_storage(true);
	/* update data list from db */
	logging_update(true);

	switch (period) {
	case DATA_LATEST:
		index = 0;
		break;
	case DATA_3HOUR:
		index = 3;
		break;
	case DATA_6HOUR:
		index = 6;
		break;
	case DATA_12HOUR:
		index = 12;
		break;
	case DATA_1DAY:
		index = 24;
		break;
	default:
		_E("Wrong message arguments! %d", period);
		goto failure;
	}

	hashtable = heart_cpu_get_user_app_list(uid);
	if (!hashtable) {
		_E("Fail to get app list");
		goto failure;
	}

	if (!g_hash_table_size(hashtable)) {
		_E("hash table is empty");
		goto failure;
	}

	ret = pthread_mutex_lock(&heart_cpu_mutex);
	if (ret) {
		_E("pthread_mutex_lock() failed, %d", ret);
		goto failure;
	}

	g_variant_builder_init(&builder, G_VARIANT_TYPE_TUPLE);
	sub_builder = g_variant_builder_new(G_VARIANT_TYPE("a(sii)"));

	g_hash_table_iter_init(&h_iter, hashtable);
	while (g_hash_table_iter_next(&h_iter, &key, &value)) {
		table = (struct heart_cpu_table *)value;
		if (!table)
			break;
		if (period == DATA_LATEST) {
			utime = table->total_utime;
			stime = table->total_stime;
		} else {
			utime = table->utime;
			stime = table->stime;
			i =  table->cpu_info->len;
			if (i < index)
				index = i;
			for (i = 0; i < index; i++) {
				struct heart_cpu_info *ci;
				ci = g_array_index(table->cpu_info, struct heart_cpu_info *, i);
				if (!ci)
					break;
				utime += ci->utime;
				stime += ci->stime;
			}
		}
		ftime = table->fg_time;
		total = utime + stime;
		if (total == 0)
			continue;
		appid = table->appid;

		g_variant_builder_add(sub_builder, "(sii)", appid, total, ftime);
	}

	ret = pthread_mutex_unlock(&heart_cpu_mutex);
	if (ret) {
		_E("pthread_mutex_unlock() failed, %d", ret);
		g_variant_builder_unref(sub_builder);
		goto failure;
	}

	g_variant_builder_add_value(&builder, g_variant_new("a(sii)", sub_builder));
	g_variant_builder_unref(sub_builder);

	g_dbus_method_invocation_return_value(invocation, g_variant_builder_end(&builder));
	return;

failure:
	D_BUS_REPLY_NULL(invocation);
}

static void dbus_heart_reset_cpu_data(GDBusMethodInvocation *invocation, GVariant *params)
{
	int ret = -1;

	GHashTableIter iter_table;
	gpointer key, value;

	g_hash_table_iter_init(&iter_table, heart_cpu_app_list);
	while (g_hash_table_iter_next(&iter_table, &key, &value))
		ret = heart_cpu_hashtable_renew(((struct heart_cpu_dat_cache*)value)->list, time(NULL));


	g_dbus_method_invocation_return_value(invocation, g_variant_new("(i)", ret));
}

static void dbus_heart_update_cpu_data(GDBusMethodInvocation *invocation, GVariant *params)
{
	heart_cpu_update_app_list(NULL);

	g_dbus_method_invocation_return_value(invocation, g_variant_new("(i)", 0));
}

static void dbus_heart_sync_cpu_data(GDBusMethodInvocation *invocation, GVariant *params)
{
	int ret = 0;

	heart_cpu_update_app_list(NULL);

	ret = logging_sync(invocation);
	if (ret)
		g_dbus_method_invocation_return_value(invocation, g_variant_new("(i)", ret));
}

static void dbus_heart_save_to_file(GDBusMethodInvocation *invocation, GVariant *params)
{
	int ret;

	GHashTableIter iter_table;
	gpointer key, value;
	struct heart_cpu_dat_cache *cache;

	g_hash_table_iter_init(&iter_table, heart_cpu_app_list);
	while (g_hash_table_iter_next(&iter_table, &key, &value)) {
		cache = (struct heart_cpu_dat_cache*)value;
		ret = heart_cpu_save_to_file(cache);
		if (ret) {
			_E("save to file failed");
			D_BUS_REPLY_NULL(invocation);
			return;
		}
		cache->last_file_commit_time = logging_get_time(CLOCK_BOOTTIME);
	}

	g_dbus_method_invocation_return_value(invocation, g_variant_new("(i)", ret));
}

static const char dbus_methods_xml[] =
"		<method name='GetCpuData'>"
"			<arg type='i' name='Uid' direction='in'/>"
"			<arg type='s' name='Appid' direction='in'/>"
"			<arg type='i' name='Period' direction='in'/>"
"			<arg type='i' name='Utime' direction='out'/>"
"			<arg type='i' name='Stime' direction='out'/>"
"		</method>"
"		<method name='GetCpuDataList'>"
"			<arg type='i' name='Uid' direction='in'/>"
"			<arg type='i' name='Period' direction='in'/>"
"			<arg type='a(sii)' name='CpuTimePerApp' direction='out'/>"
"		</method>"
"		<method name='ResetCpuData'>"
"			<arg type='i' name='ZeroOnSuccess' direction='out'/>"
"		</method>"
"		<method name='UpdateCpuData'>"
"			<arg type='i' name='Zero' direction='out'/>"
"		</method>"
"		<method name='SyncCpuData'>"
"			<arg type='i' name='ZeroOnSuccess' direction='out'/>"
"		</method>"
"		<method name='SaveCpuData'>"
"			<arg type='i' name='ZeroOnSuccess' direction='out'/>"
"		</method>";

static struct d_bus_method dbus_methods[] = {
	{ "GetCpuData",     dbus_heart_get_cpu_data },
	{ "GetCpuDataList", dbus_heart_get_cpu_data_list },
	{ "ResetCpuData",   dbus_heart_reset_cpu_data },
	{ "UpdateCpuData",  dbus_heart_update_cpu_data },
	{ "SyncCpuData",    dbus_heart_sync_cpu_data },
	{ "SaveCpuData",    dbus_heart_save_to_file },
};

static int heart_cpu_reset(void *data)
{
	GHashTableIter iter_table;
	gpointer key, value;
	int ret = RESOURCED_ERROR_FAIL;

	g_hash_table_iter_init(&iter_table, heart_cpu_app_list);
	while (g_hash_table_iter_next(&iter_table, &key, &value))
		ret = heart_cpu_hashtable_renew(((struct heart_cpu_dat_cache*)value)->list, time(NULL));

	return ret;
}

static int heart_cpu_init(void *data)
{
	const GArray *user_list = NULL;
	int ret;
	uid_t *uid;
	struct heart_cpu_dat_cache *cache;

	ret = logging_module_init(CPU_NAME, ONE_DAY, TEN_MINUTE, heart_cpu_update, TEN_MINUTE, USER_DEFAULT);
	if (ret != RESOURCED_ERROR_NONE) {
		_E("logging module init failed");
		return RESOURCED_ERROR_FAIL;
	}

	user_list = userinfo_get_list();
	if (!user_list) {
		_E("Fail to get user table");
		return RESOURCED_ERROR_FAIL;
	}

	if (!heart_cpu_app_list) {
		heart_cpu_app_list = g_hash_table_new_full(
				g_int_hash,
				g_int_equal,
				heart_cpu_free_uid,
				heart_cpu_free_dat_cache);

		userinfo_for_each(elem, user_list) {
			cache = (struct heart_cpu_dat_cache*)malloc(sizeof(struct heart_cpu_dat_cache));
			assert(cache);

			cache->path = strndup(elem->home_dir, strlen(elem->home_dir));
			cache->list = g_hash_table_new_full(
					g_str_hash,
					g_str_equal,
					NULL,
					heart_free_value);

			ret = heart_cpu_read_from_file(cache);
			assert(ret == RESOURCED_ERROR_NONE);

			cache->last_file_commit_time = logging_get_time(CLOCK_BOOTTIME);

			uid = malloc(sizeof(uid_t));
			assert(uid);

			*uid = elem->uid;
			g_hash_table_insert(heart_cpu_app_list, (gpointer)uid, (gpointer)cache);
		}
	}

	heart_dbus_declare_methods(dbus_methods_xml, dbus_methods, ARRAY_SIZE(dbus_methods));

	register_notifier(RESOURCED_NOTIFIER_SERVICE_LAUNCH, heart_cpu_service_launch);
	register_notifier(RESOURCED_NOTIFIER_APP_FOREGRD, heart_cpu_foreground_state);
	register_notifier(RESOURCED_NOTIFIER_APP_BACKGRD, heart_cpu_background_state);
	register_notifier(RESOURCED_NOTIFIER_APP_GROUP, heart_cpu_update_appid);
	register_notifier(RESOURCED_NOTIFIER_DATA_UPDATE, heart_cpu_update_app_list);
	register_notifier(RESOURCED_NOTIFIER_DATA_RESET, heart_cpu_reset);

	_D("heart cpu init finished");
	return RESOURCED_ERROR_NONE;
}

static int heart_cpu_exit(void *data)
{
	GHashTableIter iter;
	gpointer key, value;
	struct heart_cpu_dat_cache *cache;

	unregister_notifier(RESOURCED_NOTIFIER_SERVICE_LAUNCH, heart_cpu_service_launch);
	unregister_notifier(RESOURCED_NOTIFIER_APP_FOREGRD, heart_cpu_foreground_state);
	unregister_notifier(RESOURCED_NOTIFIER_APP_BACKGRD, heart_cpu_background_state);
	unregister_notifier(RESOURCED_NOTIFIER_APP_GROUP, heart_cpu_update_appid);
	unregister_notifier(RESOURCED_NOTIFIER_DATA_UPDATE, heart_cpu_update_app_list);
	unregister_notifier(RESOURCED_NOTIFIER_DATA_RESET, heart_cpu_reset);

	if (heart_cpu_app_list) {
		g_hash_table_iter_init(&iter, heart_cpu_app_list);
		while (g_hash_table_iter_next(&iter, &key, &value)) {
			cache = (struct heart_cpu_dat_cache*)value;
			heart_cpu_save_to_file(cache);
		}
		g_hash_table_destroy(heart_cpu_app_list);
	}

	logging_module_exit();

	_D("heart cpu exit");
	return RESOURCED_ERROR_NONE;
}

static int heart_cpu_dump(FILE *fp, int mode, void *data)
{
	time_t starttime;
	char timestr[80];
	struct tm loc_tm;
	static GHashTable *cpu_usage_list;
	gpointer value;
	gpointer key;
	GHashTableIter h_iter;
	struct heart_cpu_table *table;

	starttime = time(NULL);

	starttime -= mode;
	localtime_r(&starttime, &loc_tm);
	/* print timestamp */
	strftime(timestr, sizeof(timestr),
			"%Y-%m-%d %H:%M:%S%z", &loc_tm);

	cpu_usage_list = g_hash_table_new_full(
			g_str_hash,
			g_str_equal,
			NULL,
			free);


	logging_read_foreach(CPU_NAME, NULL, NULL, starttime, 0,
			heart_cpu_update, cpu_usage_list);

	if (!g_hash_table_size(cpu_usage_list)) {
		_E("hash table is empty");
		return 0;
	}

	LOG_DUMP(fp, "[CPU USAGE LISTS] since %s\n", timestr);
	LOG_DUMP(fp, "appid pkgid total fg_count fg_time\n");
	g_hash_table_iter_init(&h_iter, cpu_usage_list);

	while (g_hash_table_iter_next(&h_iter, &key, &value)) {

		table = (struct heart_cpu_table *)value;
		if (!table)
			break;
		LOG_DUMP(fp, "%s %s %ld %d %ld\n", table->appid, table->pkgid,
				table->total_utime + table->total_stime,
				table->fg_count, table->fg_time);
	}
	fflush(fp);
	g_hash_table_destroy(cpu_usage_list);

	return RESOURCED_ERROR_NONE;
}

static const struct heart_module_ops heart_cpu_ops = {
	.name           = "CPU",
	.init           = heart_cpu_init,
	.dump		    = heart_cpu_dump,
	.exit           = heart_cpu_exit,
};
HEART_MODULE_REGISTER(&heart_cpu_ops)
