/*
 * resourced
 *
 * Copyright (c) 2012 - 2014 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * @file logging.c
 *
 * @desc start logging system for resourced
 *
 * Copyright (c) 2014 Samsung Electronics Co., Ltd. All rights reserved.
 *
 */

#include <leveldb/c.h>
#include <unistd.h>
#include <pthread.h>
#include <glib.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <stdarg.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <assert.h>

#include "trace.h"
#include "heart.h"
#include "logging.h"
#include "resourced.h"
#include "macro.h"
#include "module.h"
#include "config-parser.h"
#include "notifier.h"
#include "userinfo-list.h"

#define LOGGING_BUF_MAX			1024
#define LOGGING_PTIORITY		20

#define DB_SIZE_THRESHOLD		(50<<20)
#define LOGGING_FILE_PATH		HEART_FILE_PATH
#define CREATE_QUERY			"CREATE TABLE IF NOT EXISTS %s (appid TEXT, pkgid TEXT, time INT, data TEXT, idx INT, PRIMARY KEY(time, idx));"
#define DELETE_QUERY_WITH_TIME	"DELETE from %s where time < %d;"
#define INSERT_QUERY			"INSERT INTO %s values (?, ?, ?, ?, ?);"
#define SELECT_QUERY			"SELECT * FROM %s WHERE time > %d AND time < %d;"
#define UPDATE_APPID_QUERY		"UPDATE %s SET appid = \'%s\' WHERE appid = \'%s\';"

#define SELECT_BEGIN_QUERY		"SELECT * FROM %s "
#define SELECT_WHERE_QUERY		"WHERE"
#define SELECT_AND_QUERY		" AND"
#define SELECT_APPID_QUERY		" appid = \'%s\'"
#define SELECT_PKGID_QUERY		" pkgid = \'%s\'"
#define SELECT_START_QUERY		" time > %d"
#define SELECT_END_QUERY		" time < %d"
#define SELECT_FINISH_QUERY		";"

#define SQLITE_FLUSH_MAX		(1024*1024)

#define FINALIZE_AND_RETURN_IF(cond, func, format, arg...) do \
	{ if (CONDITION(cond)) { \
		func(format, ##arg); \
		pthread_mutex_unlock(&(module->cache_mutex)); \
		sqlite3_finalize(db_elem->insert_stmt); \
		db_elem->insert_stmt = NULL; \
		return; \
	} } while (0)

enum { read_until_null = -1 };

struct logging_db {
	char *path;
	sqlite3 *file;
	sqlite3_stmt *insert_stmt;
};

struct logging_module {
	char *name;
	GHashTable *db_list; /* key=uid, value=logging_db */
	enum logging_db_type db_type;
	enum logging_period max_period;
	pthread_mutex_t cache_mutex;
	logging_info_cb func;
	time_t latest_update_time;
	int saved_interval;
	int updated_interval;
	enum logging_interval save_interval;
	enum logging_interval update_interval;
	GQueue *cache;
	GSList *listener;
};

struct logging_search {
	char *appid;
	char *pkgid;
	time_t start;
	time_t end;
	void *user_data;
	logging_info_cb func;
};

struct logging_listerner {
	void (*func)(char *data);
};

static pthread_t logging_sync_thread = 0;

static pthread_t logging_update_thread = 0;
static pthread_mutex_t logging_update_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t logging_update_cond = PTHREAD_COND_INITIALIZER;
static GSource *logging_update_timer = NULL;

static GArray *logging_modules;
static GHashTable *system_default_db = NULL;
static GHashTable *user_default_db = NULL;
static leveldb_t *logging_leveldb;
static leveldb_options_t *options;
static leveldb_readoptions_t *roptions;
static leveldb_writeoptions_t *woptions;

static struct logging_object *logging_instance = NULL;

static GDBusMethodInvocation *sync_reply_invocation;

time_t logging_get_time(int clk_id)
{
	struct timespec ts;

	if (clock_gettime(clk_id, &ts) == -1) {
		_E("clock_gettime failed");
		clock_gettime(CLOCK_MONOTONIC, &ts);
	}

	return ts.tv_sec;
}

long logging_get_time_ms(void)
{
	struct timeval tv;

	gettimeofday(&tv, NULL);
	return (tv.tv_sec * 1000 + tv.tv_usec / 1000);
}

static int logging_convert_pid_to_uid(int pid)
{
	char buf[LOGGING_BUF_MAX];
	char key[20];
	FILE *fp = NULL;
	int ret = -1;

	if (pid == PID_FOR_ROOT)
		return UID_FOR_ROOT;

	snprintf(buf, LOGGING_BUF_MAX, "/proc/%d/status", pid);
	fp = fopen(buf, "r");
	if (fp == NULL) {
		_E("Fail to open status file");
		return -1;
	}

	while (fgets(buf, LOGGING_BUF_MAX, fp) != NULL) {
		int uid;
		if (sscanf(buf, "%19s %d", key, &uid) < 0) {
			_E("sscanf failed: %m");
			break;
		}
		if (!strncmp(key, "Uid:", sizeof("Uid:"))) {
			ret = uid;
			break;
		}
	}

	fclose(fp);
	return ret;
}

static struct logging_module *logging_find_module(char *name)
{
	int i;
	int namelen = strlen(name) + 1;

	if (!logging_modules)
		return NULL;

	for (i = 0; i < logging_modules->len; i++) {
		struct logging_module *module = g_array_index(logging_modules,
				struct logging_module *, i);
		if (!strncmp(name, module->name, namelen))
			return module;
	}

	return NULL;
}

static int logging_db_busy(void * UNUSED user, int attempts)
{
	struct timespec req, rem;

	_E("DB locked by another process, attempts number %d", attempts);

	req.tv_sec = 0;
	req.tv_nsec = 500000000;
	nanosleep(&req, &rem);
	return 1;
}

static void logging_free_db_key(gpointer data)
{
	if (data)
		free(data);
}

static void logging_free_db_value(gpointer data)
{
	struct logging_db *db_elem = (struct logging_db*)data;
	if (db_elem->insert_stmt) {
		sqlite3_finalize(db_elem->insert_stmt);
		db_elem->insert_stmt = NULL;
	}
	if (db_elem->file)
		sqlite3_close(db_elem->file);
	if (db_elem->path)
		free(db_elem->path);
	free(db_elem);
}

static void logging_free_queue(gpointer data)
{
	struct logging_table_form *table = (struct logging_table_form*)data;
	if (table->data)
		free(table->data);
	free(table);
}

/* If db already exists, just open that db */
static int logging_create_default_db(char *name, sqlite3 **db)
{
	if (sqlite3_open(name, db) != SQLITE_OK) {
		_E("%s DB open failed (%s)", name, sqlite3_errmsg(*db));
		return RESOURCED_ERROR_DB_FAILED;
	}

	/* For backward compatibility */
	if (strstr(name, "storage")) {
		if (sqlite3_exec(*db, "PRAGMA journal_mode = WAL", 0, 0, 0) != SQLITE_OK) {
			_E("Can't set journal mode %s", sqlite3_errmsg(*db));
			return RESOURCED_ERROR_DB_FAILED;
		}
	} else {
		if (sqlite3_exec(*db, "PRAGMA locking_mode = NORMAL", 0, 0, 0) != SQLITE_OK) {
			_E("Can't set locking mode %s", sqlite3_errmsg(*db));
			return RESOURCED_ERROR_DB_FAILED;
		}
	}

	/* Set how many times we'll repeat our attempts for sqlite_step */
	if (sqlite3_busy_handler(*db, logging_db_busy, NULL) != SQLITE_OK)
		_E("Couldn't set busy handler!");

	return RESOURCED_ERROR_NONE;
}

int logging_get_default_db(char *name, sqlite3 **db)
{
	struct logging_db *db_info;
	struct logging_module *module = logging_find_module(name);
	if (!module) {
		_E("HEART-%s is disabled", name);
		return RESOURCED_ERROR_FAIL;
	}

	db_info = (struct logging_db*)g_hash_table_lookup(module->db_list, GINT_TO_POINTER(UID_FOR_ROOT));
	if (!db_info) {
		_E("HEART-%s doesn't use default DB", name);
		return RESOURCED_ERROR_FAIL;
	}

	*db = db_info->file;
	return RESOURCED_ERROR_NONE;
}

static int logging_update_user_db_list(char *name, enum logging_db_type type, GHashTable **db_list)
{
	const GArray *user_list = NULL;
	uid_t *uid;
	char path[LOGGING_BUF_MAX];
	sqlite3 *file = NULL;
	struct logging_db *db_elem;

	user_list = userinfo_get_list();
	if (!user_list) {
		_E("Fail to get user table");
		return RESOURCED_ERROR_FAIL;
	}

	if (*db_list)
		g_hash_table_unref(*db_list);
	*db_list = g_hash_table_new_full(g_int_hash, g_int_equal,
			logging_free_db_key, logging_free_db_value);
	if (!(*db_list)) {
		_E("Fail to create user default DB table");
		return RESOURCED_ERROR_FAIL;
	}

	userinfo_for_each(elem, user_list) {
		switch (type) {
		case USER_DEFAULT:
			snprintf(path, LOGGING_BUF_MAX, USER_DEFAULT_DB_NAME, elem->home_dir);
			break;
		case USER_OWN:
			if (!name) {
				_E("You must write name to create your own DB");
				return RESOURCED_ERROR_INVALID_PARAMETER;
			}
			snprintf(path, LOGGING_BUF_MAX, USER_OWN_DB_NAME, elem->home_dir, name);
			break;
		default:
			_E("HEART-%s doesn't use per-user DB", name);
			return RESOURCED_ERROR_FAIL;
			break;
		}

		if (logging_create_default_db(path, &file) != RESOURCED_ERROR_NONE) {
			_E("Fail to create user default DB");
			return RESOURCED_ERROR_DB_FAILED;
		}

		db_elem = NULL;
		uid = malloc(sizeof(uid_t));
		if (!uid)
			goto oom_failure;

		*uid = elem->uid;

		db_elem = (struct logging_db*)malloc(sizeof(struct logging_db));
		if (!db_elem)
			goto oom_failure;

		db_elem->path = strndup(path, strlen(path));
		if (!db_elem->path)
			goto oom_failure;

		db_elem->file = file;
		db_elem->insert_stmt = NULL;
		g_hash_table_insert(*db_list, (gpointer)uid, (gpointer)db_elem);
	}

	return RESOURCED_ERROR_NONE;

oom_failure:
	free(uid);
	if (db_elem)
		free(db_elem->path);
	free(db_elem);

	_E("out of memory!");
	return RESOURCED_ERROR_OUT_OF_MEMORY;
}

int logging_module_init(char *name, enum logging_period max_period,
		enum logging_interval save_interval, logging_info_cb func,
		enum logging_interval update_interval, enum logging_db_type db_type)
{
	int ret;
	GHashTableIter iter;
	gpointer key, value;
	char db_path[LOGGING_BUF_MAX];
	sqlite3 *db_file;
	sqlite3_stmt *stmt = NULL;
	struct logging_db *db_elem;
	char buf[LOGGING_BUF_MAX] = {0, };
	struct logging_module *module;

	if (!logging_instance) {
		logging_instance = (struct logging_object *)malloc(sizeof(struct logging_object));
		if (!logging_instance) {
			_E("out of memory");
			return RESOURCED_ERROR_OUT_OF_MEMORY;
		}

		logging_instance->ref = 0;
		ret = logging_init(NULL);
		if (ret != RESOURCED_ERROR_NONE) {
			free(logging_instance);
			logging_instance = NULL;
			_E("failed to init logging");
			return ret;
		}
	}

	logging_instance->ref++;

	if (logging_find_module(name)) {
		_E("%s is already exist", name);
		return RESOURCED_ERROR_INVALID_PARAMETER;
	}

	module = calloc(1, sizeof(struct logging_module));
	if (!module) {
		_E("out of memory");
		return RESOURCED_ERROR_OUT_OF_MEMORY;
	}

	/* DB create */
	switch (db_type) {
	case SYSTEM_DEFAULT:
		module->db_list = system_default_db;
		break;
	case SYSTEM_OWN:
		module->db_list = g_hash_table_new_full(g_direct_hash, g_direct_equal,
				NULL, logging_free_db_value);
		if (!(module->db_list)) {
			_E("Fail to create HEART-%s DB table", name);
			ret = RESOURCED_ERROR_FAIL;
			goto error;
		}
		snprintf(db_path, LOGGING_BUF_MAX, SYSTEM_OWN_DB_NAME, name);
		if (logging_create_default_db(db_path, &db_file)
				!= RESOURCED_ERROR_NONE) {
			_E("Fail to create %s", db_path);
			ret = RESOURCED_ERROR_DB_FAILED;
			goto error;
		}
		db_elem = (struct logging_db*)malloc(sizeof(struct logging_db));
		assert(db_elem);

		db_elem->path = strndup(db_path, sizeof(db_path));
		assert(db_elem->path);

		db_elem->file = db_file;
		db_elem->insert_stmt = NULL;
		g_hash_table_insert(module->db_list, GINT_TO_POINTER(UID_FOR_ROOT), (gpointer)db_elem);
		/* NOTICE : You MUST define schema at the sub-module */
		break;
	case USER_DEFAULT:
		module->db_list = user_default_db;
		break;
	case USER_OWN:
		if (logging_update_user_db_list(name, USER_OWN, &(module->db_list))) {
			_E("Fail to create HEART-%s DB", name);
			ret = RESOURCED_ERROR_FAIL;
			goto error;
		}
		break;
	default:
		_E("Unknown DB type");
		ret = RESOURCED_ERROR_INVALID_PARAMETER;
		goto error;
	}

	snprintf(buf, LOGGING_BUF_MAX, CREATE_QUERY, name);
	if (module->db_list == NULL) {
		_E("db_list is NULL, name(%s), db_type(%d)", name, db_type);
		ret = RESOURCED_ERROR_FAIL;
		goto error;
	}
	g_hash_table_iter_init(&iter, module->db_list);
	while (g_hash_table_iter_next(&iter, &key, &value)) {
		/* create table using module name and field_forms */
		db_file = ((struct logging_db*)value)->file;
		ret = sqlite3_prepare_v2(db_file, buf, read_until_null, &stmt, NULL);
		if (ret != SQLITE_OK) {
			_E("create %s table failed %s", name, sqlite3_errmsg(db_file));
			sqlite3_finalize(stmt);
			ret = RESOURCED_ERROR_DB_FAILED;
			goto error;
		}

		if (sqlite3_step(stmt) != SQLITE_DONE) {
			_E("create %s table failed %s", name, sqlite3_errmsg(db_file));
			sqlite3_finalize(stmt);
			ret = RESOURCED_ERROR_DB_FAILED;
			goto error;
		}

		sqlite3_finalize(stmt);
	}

	/* make logging_module_inform and set module_inform */
	module->db_type = db_type;
	module->func = func;
	module->latest_update_time = time(NULL);
	module->save_interval = save_interval;
	module->saved_interval = save_interval;
	module->update_interval = update_interval;
	module->updated_interval = update_interval;
	module->listener = NULL;

	if (asprintf(&(module->name), "%s", name) < 0) {
		_E("asprintf failed");
		ret = RESOURCED_ERROR_OUT_OF_MEMORY;
		goto error;
	}

	module->max_period = max_period;

	if (pthread_mutex_init(&module->cache_mutex, NULL) < 0) {
		_E("%s module mutex_init failed %d", name, errno);
		ret = RESOURCED_ERROR_FAIL;
		goto error;
	}

	module->cache = g_queue_new();

	if (!module->cache) {
		_E("g_queue_new failed");
		ret = RESOURCED_ERROR_OUT_OF_MEMORY;
		goto error;
	}

	g_queue_init(module->cache);

	g_array_append_val(logging_modules, module);

	return RESOURCED_ERROR_NONE;

error:
	if (module->name)
		free(module->name);
	if (module->db_list && (db_type == SYSTEM_OWN || db_type == USER_OWN))
		g_hash_table_unref(module->db_list);
	if (module)
		free(module);

	return ret;
}

int logging_module_exit(void)
{
	if (!logging_instance)
		return RESOURCED_ERROR_NONE;

	logging_instance->ref--;

	if (logging_instance->ref == 0) {
		free(logging_instance);
		logging_instance = NULL;
		logging_exit(NULL);
	}
	return RESOURCED_ERROR_NONE;
}

int logging_register_listener(char *name, logging_listener_cb listener_cb)
{
	GSList *registerd;
	struct logging_listerner *listener;
	struct logging_module *module;

	if (!listener_cb) {
		_E("invalid listern func");
		return RESOURCED_ERROR_INVALID_PARAMETER;
	}

	module = logging_find_module(name);

	if (!module) {
		_E("There is no %s module", name);
		return RESOURCED_ERROR_INVALID_PARAMETER;
	}

	registerd = g_slist_find(module->listener, listener_cb);

	if (registerd) {
		_E("already registerd listener %p", listener_cb);
		return RESOURCED_ERROR_INVALID_PARAMETER;
	}

	listener = malloc(sizeof(struct logging_listerner));
	if (!listener) {
		_E("Fail to malloc for notifier!");
		return RESOURCED_ERROR_OUT_OF_MEMORY;
	}

	listener->func = listener_cb;

	module->listener = g_slist_append(module->listener, listener);

	return RESOURCED_ERROR_NONE;
}

int logging_unregister_listener(char *name, logging_listener_cb listener_cb)
{
	GSList *registerd;
	struct logging_module *module;

	if (!listener_cb) {
		_E("invalid listern func");
		return RESOURCED_ERROR_INVALID_PARAMETER;
	}

	module = logging_find_module(name);

	if (!module) {
		_E("There is no %s module", name);
		return RESOURCED_ERROR_INVALID_PARAMETER;
	}

	registerd = g_slist_find(module->listener, listener_cb);

	if (!registerd) {
		_E("It is not registered listerner %p", listener_cb);
		return RESOURCED_ERROR_INVALID_PARAMETER;
	}

	module->listener = g_slist_remove(module->listener, registerd->data);

	free(registerd->data);

	return RESOURCED_ERROR_NONE;
}

int logging_operate(int uid, char *name, char *appid, char *pkgid, time_t time, char *data, int operation)
{
	/* Save data to cache  */
	struct logging_module *module;
	struct logging_table_form *table;

	module = logging_find_module(name);

	if (!module) {
		_E("There is no %s module", name);
		return RESOURCED_ERROR_INVALID_PARAMETER;
	}

	table = malloc(sizeof(struct logging_table_form));

	if (!table) {
		_E("malloc failed");
		return RESOURCED_ERROR_OUT_OF_MEMORY;
	}

	table->uid = uid;

	if (operation == INSERT) {
		if (snprintf(table->appid, MAX_APPID_LENGTH, "%s", appid) < 0) {
			_E("snprintf failed");
			free(table);
			return RESOURCED_ERROR_OUT_OF_MEMORY;
		}

		if (snprintf(table->pkgid, MAX_PKGNAME_LENGTH, "%s", pkgid) < 0) {
			_E("snprintf failed");
			free(table);
			return RESOURCED_ERROR_OUT_OF_MEMORY;
		}
	}

	table->time = time;

	if (asprintf(&(table->data), "%s", data) < 0) {
		_E("asprintf failed");
		free(table);
		return RESOURCED_ERROR_OUT_OF_MEMORY;
	}
	table->operation = operation;

	pthread_mutex_lock(&(module->cache_mutex));
	g_queue_push_tail(module->cache, (gpointer)table);
	pthread_mutex_unlock(&(module->cache_mutex));

	/* call listeners */
	if (module->listener) {
		GSList *iter;
		struct logging_listerner *listener;
		char buf[LOGGING_BUF_MAX];

		gslist_for_each_item(iter, module->listener) {
			listener = (struct logging_listerner *)iter->data;
			snprintf(buf, LOGGING_BUF_MAX, "%d %s %s %d %s", uid, appid, pkgid, (int)time, data);

			listener->func(buf);
		}
	}

	return RESOURCED_ERROR_NONE;
}

int logging_write(int pid, char *name, char *appid, char *pkgid, time_t time, char *data)
{
	int uid = logging_convert_pid_to_uid(pid);

	if (uid < 0) {
		_E("Fail to get uid");
		return RESOURCED_ERROR_FAIL;
	}
	return logging_operate(uid, name, appid, pkgid, time, data, INSERT);
}

int logging_leveldb_put(char *key, unsigned int key_len, char *value, unsigned int value_len)
{
	char *err =  NULL;

	if (!key || !key_len || !value || !value_len)
		return RESOURCED_ERROR_INVALID_PARAMETER;

	if (!logging_leveldb) {
		_E("leveldb is not initialized");
		return RESOURCED_ERROR_DB_FAILED;
	}

	leveldb_put(logging_leveldb, woptions, key, key_len, value, value_len, &err);
	if (err != NULL) {
		_E("Failed to put to leveldb");
		free(err);
		return RESOURCED_ERROR_DB_FAILED;
	}
	_D("%s:%s", key, value);
	return RESOURCED_ERROR_NONE;
}

int logging_leveldb_putv(char *key, unsigned int key_len, const char *fmt, ...)
{
	char *err =  NULL;
	va_list ap;
	char value[LOGGING_BUF_MAX];
	unsigned int value_len;

	if (!key || !key_len || !fmt)
		return RESOURCED_ERROR_INVALID_PARAMETER;

	if (!logging_leveldb) {
		_E("leveldb is not initialized");
		return RESOURCED_ERROR_DB_FAILED;
	}

	va_start(ap, fmt);
	vsnprintf(value, LOGGING_BUF_MAX, fmt, ap);
	va_end(ap);

	value_len = strlen(value);
	if (!value_len) {
		_E("Failed to get length of string");
		return RESOURCED_ERROR_DB_FAILED;
	}

	leveldb_put(logging_leveldb, woptions, key, key_len, value, value_len, &err);
	if (err != NULL) {
		_E("Failed to put to leveldb");
		free(err);
		return RESOURCED_ERROR_DB_FAILED;
	}
	_D("%s:%s", key, value);
	return RESOURCED_ERROR_NONE;
}

int logging_leveldb_read(char *key, unsigned int key_len, char *value, unsigned int value_len)
{
	size_t read_len;
	char *err =  NULL;
	char *result = NULL;

	if (!key || !key_len || !value || !value_len)
		return RESOURCED_ERROR_INVALID_PARAMETER;

	if (!logging_leveldb) {
		_E("leveldb is not initialized");
		return RESOURCED_ERROR_DB_FAILED;
	}

	result = leveldb_get(logging_leveldb, roptions, key, key_len, &read_len, &err);
	if (err != NULL) {
		_E("Failed to get from leveldb");
		free(err);
		return RESOURCED_ERROR_DB_FAILED;
	}
	if (read_len < value_len) {
		memcpy(value, result, read_len);
		value[read_len] = '\0';
	} else {
		memcpy(value, result, value_len - 1);
		value[value_len - 1] = '\0';
	}

	free(result);

	_D("%s:%s", key, value);
	return RESOURCED_ERROR_NONE;
}

int logging_leveldb_delete(char *key, unsigned int key_len)
{
	char *err =  NULL;

	if (!key || !key_len)
		return RESOURCED_ERROR_INVALID_PARAMETER;

	if (!logging_leveldb) {
		_E("leveldb is not initialized");
		return RESOURCED_ERROR_DB_FAILED;
	}

	leveldb_delete(logging_leveldb, woptions, key, key_len, &err);
	if (err != NULL) {
		_E("Failed to delete from leveldb");
		free(err);
		return RESOURCED_ERROR_DB_FAILED;
	}
	return RESOURCED_ERROR_NONE;
}

int logging_get_latest_in_cache(char *name, char *appid, char **data)
{
	int i, len;
	struct logging_module *module;
	struct logging_table_form *table;

	module = logging_find_module(name);

	if (!module) {
		_E("There is no %s module", name);
		return RESOURCED_ERROR_INVALID_PARAMETER;
	}

	if (!appid) {
		_E("appid parameter should be valid");
		return RESOURCED_ERROR_INVALID_PARAMETER;
	}

	pthread_mutex_lock(&(module->cache_mutex));
	len = g_queue_get_length(module->cache);
	if (!len) {
		_I("%s cache is empty", module->name);
		pthread_mutex_unlock(&(module->cache_mutex));
		return RESOURCED_ERROR_NO_DATA;
	}

	*data = NULL;

	for (i = 0; i < len; i++) {
		table = g_queue_peek_nth(module->cache, i);

		if (table &&
			!strncmp(appid, table->appid, strlen(table->appid)+1))
			*data = table->data;
	}
	pthread_mutex_unlock(&(module->cache_mutex));

	if (!*data) {
		_E("NOT found in cache %s", appid);
		return RESOURCED_ERROR_NO_DATA;
	}

	return RESOURCED_ERROR_NONE;
}

static void logging_cache_search(struct logging_table_form *data, struct logging_search *search)
{
	/* search in cache */
	/* true condition, call function */
	if (search->appid) {
		if (strncmp(search->appid, data->appid, strlen(data->appid)+1))
			return;

		if (search->start && search->start > data->time)
			return;

		if (search->end && search->end < data->time)
			return;
	} else if (search->pkgid) {
		if (strncmp(search->pkgid, data->pkgid, strlen(data->pkgid)+1))
			return;

		if (search->start && search->start > data->time)
			return;

		if (search->end && search->end < data->time)
			return;
	} else if (search->start) {
		if (search->start > data->time)
			return;

		if (search->end && search->end < data->time)
			return;
	} else if (search->end) {
		if (search->end < data->time)
			return;
	}

	search->func(data, search->user_data);
}

int logging_read_foreach(char *name, char *appid, char *pkgid,
		time_t start_time, time_t end_time, logging_info_cb callback, void *user_data)
{
	/* Read from storage (cache & db) */
	int result;
	int len;
	time_t cur_time;
	sqlite3_stmt *stmt = NULL;
	struct logging_table_form table;
	struct logging_module *module;
	struct logging_search search;
	char buf[LOGGING_BUF_MAX] = {0, };
	GHashTableIter iter;
	gpointer key, value;
	sqlite3 *db;

	module = logging_find_module(name);

	if (!module) {
		_E("There is no %s module", name);
		return RESOURCED_ERROR_INVALID_PARAMETER;
	}
	cur_time = time(NULL);
	search.appid = NULL;
	search.pkgid = NULL;
	search.start = 0;
	search.end = cur_time;
	search.func = callback;
	search.user_data = user_data;

	len = snprintf(buf, LOGGING_BUF_MAX, SELECT_BEGIN_QUERY, name);

	if (appid) {
		len += snprintf(buf+len, LOGGING_BUF_MAX - len, SELECT_WHERE_QUERY);

		search.appid = appid;
		len += snprintf(buf+len, LOGGING_BUF_MAX - len, SELECT_APPID_QUERY, appid);

		if (start_time) {
			search.start = start_time;
			len += snprintf(buf+len, LOGGING_BUF_MAX - len, SELECT_AND_QUERY);
			len += snprintf(buf+len, LOGGING_BUF_MAX - len, SELECT_START_QUERY, (int)start_time);
		}
	} else if (pkgid) {
		len += snprintf(buf+len, LOGGING_BUF_MAX - len, SELECT_WHERE_QUERY);

		search.pkgid = pkgid;
		len += snprintf(buf+len, LOGGING_BUF_MAX - len, SELECT_PKGID_QUERY, pkgid);

		if (start_time) {
			search.start = start_time;
			len += snprintf(buf+len, LOGGING_BUF_MAX - len, SELECT_AND_QUERY);
			len += snprintf(buf+len, LOGGING_BUF_MAX - len, SELECT_START_QUERY, (int)start_time);
		}
	} else if (start_time) {
		len += snprintf(buf+len, LOGGING_BUF_MAX - len, SELECT_WHERE_QUERY);

		search.start = start_time;
		len += snprintf(buf+len, LOGGING_BUF_MAX - len, SELECT_START_QUERY, (int)start_time);
	}
	if (end_time && cur_time > end_time)
		search.end = end_time;
	len += snprintf(buf+len, LOGGING_BUF_MAX - len, SELECT_AND_QUERY);
	len += snprintf(buf+len, LOGGING_BUF_MAX - len, SELECT_END_QUERY, (int)search.end);

	len += snprintf(buf+len, LOGGING_BUF_MAX - len, SELECT_FINISH_QUERY);

	/* search from db */
	g_hash_table_iter_init(&iter, module->db_list);
	while (g_hash_table_iter_next(&iter, &key, &value)) {
		db = ((struct logging_db*)value)->file;

		if (sqlite3_prepare_v2(db, buf, read_until_null, &stmt, NULL) != SQLITE_OK) {
			_E("select failed");
			sqlite3_finalize(stmt);
			return RESOURCED_ERROR_DB_FAILED;
		}

		do {
			result = sqlite3_step(stmt);
			switch (result) {
			case SQLITE_ROW:
				snprintf(table.appid, MAX_APPID_LENGTH, "%s", (char *)sqlite3_column_text(stmt, 0));
				snprintf(table.pkgid, MAX_PKGNAME_LENGTH, "%s", (char *)sqlite3_column_text(stmt, 1));
				table.time = sqlite3_column_int(stmt, 2);
				if (module->latest_update_time < table.time)
					module->latest_update_time = table.time;
				table.data = (char *)sqlite3_column_text(stmt, 3);

				callback(&table, user_data);
				break;
			case SQLITE_DONE:
				break;
			case SQLITE_ERROR:
				_E("select %s table failed %s", name, sqlite3_errmsg(db));
				/* FALLTHROUGH */
			default:
				sqlite3_finalize(stmt);
				return RESOURCED_ERROR_DB_FAILED;
			}
		} while (result == SQLITE_ROW);

		sqlite3_finalize(stmt);
	}

	/* search from cache */
	if (!g_queue_is_empty(module->cache)) {
		pthread_mutex_lock(&(module->cache_mutex));
		g_queue_foreach(module->cache, (GFunc)logging_cache_search, (gpointer)&search);
		pthread_mutex_unlock(&(module->cache_mutex));
	}

	return RESOURCED_ERROR_NONE;
}

static int logging_reset(char *name)
{
	/* Table cut using max period */
	time_t curr_time, del_time;
	sqlite3_stmt *stmt = NULL;
	struct logging_module *module;
	char buf[LOGGING_BUF_MAX] = {0, };
	GHashTableIter iter;
	gpointer key, value;
	sqlite3 *db;

	module = logging_find_module(name);

	if (!module) {
		_E("There is no %s module", name);
		return RESOURCED_ERROR_INVALID_PARAMETER;
	}

	time(&curr_time);

	switch (module->max_period) {
	case ONE_HOUR:
		del_time = curr_time - HOUR_TO_SEC(1);
		break;
	case THREE_HOUR:
		del_time = curr_time - HOUR_TO_SEC(3);
		break;
	case SIX_HOUR:
		del_time = curr_time - HOUR_TO_SEC(6);
		break;
	case TWELVE_HOUR:
		del_time = curr_time - HOUR_TO_SEC(12);
		break;
	case ONE_DAY:
		del_time = curr_time - HOUR_TO_SEC(24);
		break;
	case ONE_WEEK:
		del_time = curr_time - DAY_TO_SEC(7);
		break;
	case ONE_MONTH:
		del_time = curr_time - MONTH_TO_SEC(1);
		break;
	case FOUR_MONTH:
		del_time = curr_time - MONTH_TO_SEC(4);
		break;
	default:
		_E("%s invalid max period", module->name);
		return RESOURCED_ERROR_INVALID_PARAMETER;
	}

	snprintf(buf, LOGGING_BUF_MAX, DELETE_QUERY_WITH_TIME, name, (int)del_time);

	g_hash_table_iter_init(&iter, module->db_list);
	while (g_hash_table_iter_next(&iter, &key, &value)) {
		db = ((struct logging_db*)value)->file;

		if (sqlite3_prepare_v2(db, buf, read_until_null, &stmt, NULL) != SQLITE_OK) {
			_E("delete %s table failed %s", name, sqlite3_errmsg(db));
			sqlite3_finalize(stmt);
			return RESOURCED_ERROR_DB_FAILED;
		}

		if (sqlite3_step(stmt) != SQLITE_DONE) {
			_E("delete %s table failed %s", name, sqlite3_errmsg(db));
			sqlite3_finalize(stmt);
			return RESOURCED_ERROR_DB_FAILED;
		}

		sqlite3_finalize(stmt);
	}

	return RESOURCED_ERROR_NONE;
}

static int logging_check_storage_size(const char *db_path)
{
	int ret;
	struct stat db_stat = {0};

	if (!db_path)
		return RESOURCED_ERROR_INVALID_PARAMETER;

	ret = stat(db_path, &db_stat);

	if (ret) {
		_E("Failed to get statistics for %s errno %d",
				db_path, errno);
		return RESOURCED_ERROR_DB_FAILED;
	}

	if (db_stat.st_size >= DB_SIZE_THRESHOLD)
		return RESOURCED_ERROR_FAIL;

	return RESOURCED_ERROR_NONE;
}

int logging_modify_appid(char *module_name, char *old_appid, char *new_appid, int pid)
{
	struct logging_module *module = NULL;
	int uid = -1;
	struct logging_db *db = NULL;
	char query[LOGGING_BUF_MAX];
	int i;
	struct logging_table_form *entry;

	if (!module_name || !old_appid || !new_appid || pid < 0) {
		_E("Invalid parameter");
		return RESOURCED_ERROR_INVALID_PARAMETER;
	}

	module = logging_find_module(module_name);
	if (!module) {
		_E("There is no %s module", module_name);
		return RESOURCED_ERROR_NO_DATA;
	}

	uid = logging_convert_pid_to_uid(pid);
	if (uid < 0) {
		_E("Fail to get uid");
		return RESOURCED_ERROR_FAIL;
	}

	if (pthread_mutex_lock(&logging_update_mutex)) {
		_E("Failed to lock mutex");
		return RESOURCED_ERROR_FAIL;
	}

	/* Modify cache */
	if (!g_queue_is_empty(module->cache)) {
		pthread_mutex_lock(&(module->cache_mutex));
		for (i = 0; i < g_queue_get_length(module->cache); i++) {
			entry = (struct logging_table_form *)g_queue_peek_nth(
					module->cache, i);
			if (entry->uid == uid && !strncmp(entry->appid, old_appid, strlen(old_appid)))
				snprintf(entry->appid, MAX_APPID_LENGTH, "%s", new_appid);
		}
		pthread_mutex_unlock(&(module->cache_mutex));
	}

	/* Modify DB */
	switch (module->db_type) {
	case SYSTEM_DEFAULT:
	case SYSTEM_OWN:
		db = g_hash_table_lookup(module->db_list, GINT_TO_POINTER(UID_FOR_ROOT));
		break;
	case USER_DEFAULT:
	case USER_OWN:
		db = g_hash_table_lookup(module->db_list, (gpointer)(&uid));
		break;
	default:
		_E("Unknown db_type");
		pthread_mutex_unlock(&logging_update_mutex);
		return RESOURCED_ERROR_FAIL;
	}

	if (!db) {
		_E("Failed to get DB");
		pthread_mutex_unlock(&logging_update_mutex);
		return RESOURCED_ERROR_NO_DATA;
	}

	snprintf(query, LOGGING_BUF_MAX, UPDATE_APPID_QUERY,
			module_name, new_appid, old_appid);

	if (sqlite3_exec(db->file, query, NULL, NULL, NULL) != SQLITE_OK) {
		_E("Failed to apply query");
		pthread_mutex_unlock(&logging_update_mutex);
		return RESOURCED_ERROR_DB_FAILED;
	}

	if (pthread_mutex_unlock(&logging_update_mutex)) {
		_E("Failed to unlock mutex");
		return RESOURCED_ERROR_FAIL;
	}

	return RESOURCED_ERROR_NONE;
}

void logging_update(int force)
{
	int i, ret;
	sqlite3_stmt *stmt = NULL;
	struct logging_table_form table;
	struct logging_module *module;
	char buf[LOGGING_BUF_MAX] = {0, };
	GHashTableIter iter;
	gpointer key, value;
	sqlite3 *db;

	if (!logging_modules)
		return;

	for (i = 0; i < logging_modules->len; i++) {
		module = g_array_index(logging_modules,
				struct logging_module *, i);

		/* no update when module update func is not registered */
		if (module->func == NULL)
			continue;

		if (!force && module->updated_interval > 0) {
			module->updated_interval -= ONE_MINUTE;
			continue;
		}

		module->updated_interval = module->update_interval;

		snprintf(buf, LOGGING_BUF_MAX, SELECT_QUERY, module->name,
				(int)module->latest_update_time, (int)time(NULL));

		g_hash_table_iter_init(&iter, module->db_list);
		while (g_hash_table_iter_next(&iter, &key, &value)) {
			db = ((struct logging_db*)value)->file;

			ret = sqlite3_prepare_v2(db, buf, read_until_null, &stmt, NULL);
			if (ret != SQLITE_OK) {
				_E("select failed");
				sqlite3_finalize(stmt);
				return;
			}

			switch (module->db_type) {
			case SYSTEM_DEFAULT:
			case SYSTEM_OWN:
				table.uid = GPOINTER_TO_INT(key);
				break;
			case USER_DEFAULT:
			case USER_OWN:
				table.uid = *((int*)key);
				break;
			default:
				_E("Unknown db_type");
				return;
			}

			do {
				ret = sqlite3_step(stmt);
				switch (ret) {
				case SQLITE_ROW:
					snprintf(table.appid, MAX_APPID_LENGTH, "%s", (char *)sqlite3_column_text(stmt, 0));
					snprintf(table.pkgid, MAX_PKGNAME_LENGTH, "%s", (char *)sqlite3_column_text(stmt, 1));
					table.time = sqlite3_column_int(stmt, 2);
					if (module->latest_update_time < table.time)
						module->latest_update_time = table.time;
					if (asprintf(&(table.data), "%s", (char *)sqlite3_column_text(stmt, 3)) < 0) {
						_E("asprintf failed");
						sqlite3_finalize(stmt);
						return;
					}
					module->func(&table, NULL);
					free(table.data);
					break;
				case SQLITE_DONE:
					break;
				case SQLITE_ERROR:
					_E("select %s table failed %s", module->name,
							sqlite3_errmsg(db));
					/* FALLTHROUGH */
				default:
					sqlite3_finalize(stmt);
					return;
				}
			} while (ret == SQLITE_ROW);

			sqlite3_finalize(stmt);
		}
	}

}

void logging_save_to_storage(int force)
{
	/* Save cache to storage */
	static int index = 0;
	int i, j, len, ret = 0;
	GHashTableIter iter;
	gpointer key, value;
	char buf[LOGGING_BUF_MAX] = {0, };
	struct logging_module *module;
	struct logging_table_form *table;
	struct logging_db *db_elem;

	if (!logging_modules)
		return;

	for (i = 0; i < logging_modules->len; i++) {
		module = g_array_index(logging_modules, struct logging_module *, i);
		if (!module)
			continue;
		if (!force && module->saved_interval > 0) {
			module->saved_interval -= ONE_MINUTE;
			continue;
		}

		module->saved_interval = module->save_interval;

		/* Find q and pop */
		pthread_mutex_lock(&(module->cache_mutex));
		len = g_queue_get_length(module->cache);
		if (!len) {
			_D("%s cache is empty", module->name);
			pthread_mutex_unlock(&(module->cache_mutex));
			continue;
		}

		g_hash_table_iter_init(&iter, module->db_list);
		while (g_hash_table_iter_next(&iter, &key, &value)) {
			db_elem = (struct logging_db*)value;

			sqlite3_exec(db_elem->file, "BEGIN;", NULL, NULL, NULL);

			snprintf(buf, LOGGING_BUF_MAX, INSERT_QUERY, module->name);
			ret = sqlite3_prepare_v2(db_elem->file, buf, read_until_null,
					&(db_elem->insert_stmt), NULL);
			FINALIZE_AND_RETURN_IF(ret != SQLITE_OK, _E,
					"insert %s table failed %s", module->name,
					sqlite3_errmsg(db_elem->file));
		}

		for (j = 0; j < len; j++) {
			table = g_queue_peek_head(module->cache);
			if (!table)
				continue;

			/* Check destination */
			switch (module->db_type) {
			case SYSTEM_DEFAULT:
			case SYSTEM_OWN:
				db_elem = (struct logging_db*)g_hash_table_lookup(
						module->db_list, GINT_TO_POINTER(UID_FOR_ROOT));
				break;
			case USER_DEFAULT:
			case USER_OWN:
				db_elem = (struct logging_db*)g_hash_table_lookup(
						module->db_list, (gpointer)(&(table->uid)));
				break;
			default:
				_E("HEART-%s uses leveldb or unknown db_type", module->name);
				return;
			}

			if (!db_elem) {
				_E("There is no uid %d in the DB list", table->uid);
				continue;
			}

			switch (table->operation) {
			case INSERT:
				sqlite3_reset(db_elem->insert_stmt);

				ret = sqlite3_bind_text(db_elem->insert_stmt, 1, table->appid, -1, SQLITE_STATIC);
				FINALIZE_AND_RETURN_IF(ret != SQLITE_OK, _SE, "Can not bind appid : %s for preparing statement", table->appid);

				ret = sqlite3_bind_text(db_elem->insert_stmt, 2, table->pkgid, -1, SQLITE_STATIC);
				FINALIZE_AND_RETURN_IF(ret != SQLITE_OK, _SE, "Can not bind pkgid : %s for preparing statement", table->pkgid);

				ret = sqlite3_bind_int(db_elem->insert_stmt, 3, table->time);
				FINALIZE_AND_RETURN_IF(ret != SQLITE_OK, _SE, "Can not bind time : %ld for preparing statement", table->time);

				ret = sqlite3_bind_text(db_elem->insert_stmt, 4, table->data, -1, SQLITE_STATIC);
				FINALIZE_AND_RETURN_IF(ret != SQLITE_OK, _SE, "Can not bind data : %s for preparing statement", table->data);

				ret = sqlite3_bind_int(db_elem->insert_stmt, 5, index++);
				FINALIZE_AND_RETURN_IF(ret != SQLITE_OK, _SE, "Can not bind index : %d for preparing statement", index);

				ret = sqlite3_step(db_elem->insert_stmt);
				FINALIZE_AND_RETURN_IF(ret != SQLITE_DONE, _E, "insert %s table failed %s", module->name, sqlite3_errmsg(db_elem->file));
				break;
			default:
				_E("Unknown SQLite operation %d", table->operation);
				continue;
			}

			table = g_queue_pop_head(module->cache);
			free(table->data);
			free(table);
		}

		pthread_mutex_unlock(&(module->cache_mutex));
		g_hash_table_iter_init(&iter, module->db_list);
		while (g_hash_table_iter_next(&iter, &key, &value)) {
			db_elem = (struct logging_db*)value;
			sqlite3_exec(db_elem->file, "COMMIT;", NULL, NULL, NULL);
			sqlite3_finalize(db_elem->insert_stmt);
			db_elem->insert_stmt = NULL;
		}
	}

	for (i = 0; i < logging_modules->len; i++) {
		module = g_array_index(logging_modules, struct logging_module *, i);

		g_hash_table_iter_init(&iter, module->db_list);
		while (g_hash_table_iter_next(&iter, &key, &value)) {
			db_elem = (struct logging_db*)value;

			/* Check storage limitation by maximum period and storage size (50MiB) */
			if (logging_check_storage_size(db_elem->path) == RESOURCED_ERROR_FAIL) {
				logging_reset(module->name);
				sqlite3_exec(db_elem->file, "VACUUM;", NULL, NULL, NULL);
			}
		}
	}
	sqlite3_release_memory(SQLITE_FLUSH_MAX);
}

static void *logging_sync_thread_main(void *arg)
{
	int ret = 0;

	ret = pthread_mutex_lock(&logging_update_mutex);
	if (ret) {
		_E("logging sync thread::pthread_mutex_lock() failed (%d)", ret);
		ret = -1;
		goto reply;
	}

	logging_save_to_storage(true);

	logging_update(true);

	ret = pthread_mutex_unlock(&logging_update_mutex);
	if (ret) {
		_E("logging sync thread::pthread_mutex_unlock() failed (%d)", ret);
		ret = -1;
	}

reply:
	g_dbus_method_invocation_return_value(sync_reply_invocation, g_variant_new("(i)", ret));

	logging_sync_thread = 0;

	pthread_exit(NULL);
}

int logging_sync(GDBusMethodInvocation *reply_invocation)
{
	if (logging_sync_thread) {
		_I("logging sync thread %u is already running", (unsigned)logging_sync_thread);
		return logging_sync_thread;
	}

	sync_reply_invocation = reply_invocation;

	if (!pthread_create(&logging_sync_thread, NULL, (void *)logging_sync_thread_main, NULL)) {
		pthread_detach(logging_sync_thread);
		return 0;
	}

	_E("Failed to create logging sync thread");
	logging_sync_thread = 0;
	return -1;
}

static int logging_start(void *data)
{
	return RESOURCED_ERROR_NONE;
}

static void *logging_update_thread_main(void *arg)
{
	int ret = 0;

	setpriority(PRIO_PROCESS, 0, LOGGING_PTIORITY);

	while (1) {
		/*
		 * it starts fuction of registered module.
		 */
		ret = pthread_mutex_lock(&logging_update_mutex);
		if (ret) {
			_E("logging update thread::pthread_mutex_lock() failed, %d", ret);
			break;
		}

		ret = pthread_cond_wait(&logging_update_cond, &logging_update_mutex);
		if (ret) {
			_E("logging update thread::pthread_cond_wait() failed, %d", ret);
			ret = pthread_mutex_unlock(&logging_update_mutex);
			if (ret)
				_E("logging update thread::pthread_mutex_lock() failed, %d", ret);
			break;
		}

		logging_update(false);

		logging_save_to_storage(false);

		ret = pthread_mutex_unlock(&logging_update_mutex);
		if (ret) {
			_E("logging update thread::pthread_mutex_unlock() failed, %d", ret);
			break;
		}
	}

	/* Now our thread finishes - cleanup TID */
	logging_update_thread = 0;

	return NULL;
}

static gboolean logging_send_signal_to_update(gpointer data)
{
	int ret;
	DIR *dir_info;

	const GArray *user_list = NULL;
	char path[128];

	dir_info = opendir(LOGGING_FILE_PATH);

	if (dir_info)
		closedir(dir_info);
	else {
		_D("There is no %s. Create new one.", LOGGING_FILE_PATH);
		ret = mkdir(LOGGING_FILE_PATH, S_IRWXU | S_IRWXG | S_IROTH);
		if (ret) {
			_E("mkdir failed %s", LOGGING_FILE_PATH);
			return true;
		}
	}

	user_list = userinfo_get_list();
	if (!user_list) {
		_E("Fail to get user table");
		return true;
	}

	userinfo_for_each(elem, user_list) {
		snprintf(path, 128, HEART_USER_FILE_PATH, elem->home_dir);
		dir_info = opendir(path);

		if (dir_info)
			closedir(dir_info);
		else {
			_D("Make directory %s", path);
			ret = mkdir(path, S_IRWXU | S_IRWXG | S_IROTH);

			if (ret) {
				_E("mkdir failed %s", path);
				return true;
			}
		}
	}


#ifdef LOGGING_DEBUG
	_D("logging timer callback function start");
#endif

	/* signal to logging update thread for start update */
	ret = pthread_mutex_trylock(&logging_update_mutex);

	if (ret)
		_E("pthread_mutex_trylock fail : %d, errno : %d", ret, errno);
	else {
		pthread_cond_signal(&logging_update_cond);
#ifdef LOGGING_DEBUG
		_I("send signal to logging update thread");
#endif
		pthread_mutex_unlock(&logging_update_mutex);
	}

	return true;
}

static int logging_thread_create(void)
{
	int ret = RESOURCED_ERROR_NONE;

	/* initialize logging_update_thread */
	if (logging_update_thread) {
		_I("logging update thread %u already created", (unsigned)logging_update_thread);
	} else {
		ret = pthread_create(&logging_update_thread, NULL,
							(void *)logging_update_thread_main, (void *)NULL);
		if (ret) {
			_E("pthread creation for logging_update_thread_main failed, %d\n", ret);
			logging_update_thread = 0;
		} else {
			_D("pthread creation for logging update success");
			pthread_detach(logging_update_thread);
		}
	}
	return ret;
}

static void logging_thread_stop(void)
{
	/* Delete timer invoking update function */
	g_source_destroy(logging_update_timer);
	logging_update_timer = NULL;

	/* Wait thread for working
	 *
	 * We can wait thread for finishing work by requesting lock
	 * because thread locks before working and unlocks after
	 *
	 * Meanwhile, in this time, timer is deleted
	 * so thread no more works though it is unlocked
	 */
	if (pthread_mutex_lock(&logging_update_mutex))
		_E("Failed to pthread_mutex_lock");
	if (pthread_mutex_unlock(&logging_update_mutex))
		_E("Failed to pthread_mutex_unlock");
}

static int logging_poweroff(void *data)
{
	/* flush module cache */
	logging_save_to_storage(true);
	return RESOURCED_ERROR_NONE;
}

int logging_init(void *data)
{
	int ret = RESOURCED_ERROR_NONE;
	char *err = NULL;
	sqlite3 *db_file = NULL;
	struct logging_db *db_elem;

	_D("logging_init start");

	ret = logging_thread_create();
	if (ret) {
		_E("logging thread create failed");
		return RESOURCED_ERROR_FAIL;
	}

	/* module array create */
	logging_modules = g_array_new(false, false, sizeof(struct logging_module *));

	if (logging_modules == NULL) {
		_E("logging_modules_array create failed");
		return RESOURCED_ERROR_OUT_OF_MEMORY;
	}

	/* Create system default DB */
	system_default_db = g_hash_table_new_full(g_direct_hash, g_direct_equal,
			NULL, logging_free_db_value);
	if (!system_default_db) {
		_E("Fail to create system default DB table");
		return RESOURCED_ERROR_FAIL;
	}
	if (logging_create_default_db(SYSTEM_DEFAULT_DB_NAME, &db_file)
			!= RESOURCED_ERROR_NONE) {
		_E("Fail to create system default DB");
		return RESOURCED_ERROR_DB_FAILED;
	}
	db_elem = (struct logging_db*)malloc(sizeof(struct logging_db));
	if (!db_elem) {
		_E("Out of memory!");
		return RESOURCED_ERROR_OUT_OF_MEMORY;
	}

	db_elem->path = strndup(SYSTEM_DEFAULT_DB_NAME, sizeof(SYSTEM_DEFAULT_DB_NAME));
	if (!(db_elem->path)) {
		_E("Out of memory!");
		free(db_elem);
		return RESOURCED_ERROR_OUT_OF_MEMORY;
	}

	db_elem->file = db_file;
	db_elem->insert_stmt = NULL;
	g_hash_table_insert(system_default_db, GINT_TO_POINTER(UID_FOR_ROOT), (gpointer)db_elem);

	/* Create user default DB */
	if (logging_update_user_db_list(NULL, USER_DEFAULT, &user_default_db)) {
		_E("Fail to create user default DB");
		return RESOURCED_ERROR_FAIL;
	}

	/* Create leveldb */
	options = leveldb_options_create();
	leveldb_options_set_create_if_missing(options, 1);
	logging_leveldb = leveldb_open(options, SYSTEM_LEVEL_DB_DIR, &err);
	if (err != NULL) {
		_E("Failed to open leveldb");
		free(err);
		return RESOURCED_ERROR_DB_FAILED;
	}
	roptions = leveldb_readoptions_create();
	woptions = leveldb_writeoptions_create();
	leveldb_writeoptions_set_sync(woptions, 1);

	/* Register notifier and timer */
	register_notifier(RESOURCED_NOTIFIER_LOGGING_START, logging_start);
	register_notifier(RESOURCED_NOTIFIER_POWER_OFF, logging_poweroff);

	if (logging_update_timer == NULL) {
		_D("logging data timer start");
		logging_update_timer = g_timeout_source_new_seconds(ONE_MINUTE);
		g_source_set_callback(logging_update_timer, logging_send_signal_to_update, NULL, NULL);
		g_source_attach(logging_update_timer, NULL);
	}

	return RESOURCED_ERROR_NONE;
}

int logging_exit(void *data)
{
	int i;
	struct logging_module *module;

	logging_thread_stop();

	unregister_notifier(RESOURCED_NOTIFIER_LOGGING_START, logging_start);
	unregister_notifier(RESOURCED_NOTIFIER_POWER_OFF, logging_poweroff);

	/* flush module cache */
	logging_save_to_storage(true);

	/* logging_modules array deinitialize */
	for (i = 0; i < logging_modules->len; i++) {
		module = g_array_index(logging_modules, struct logging_module *, i);

		if (module->name)
			free(module->name);

		/* Own DB close */
		if (module->db_type == SYSTEM_OWN || module->db_type == USER_OWN)
			g_hash_table_unref(module->db_list);

		if (module->cache)
			g_queue_free_full(module->cache, logging_free_queue);

		if (module->listener)
			g_slist_free_full(module->listener, free);

		free(module);
	}
	g_array_free(logging_modules, true);

	/* Default DB close */
	if (system_default_db)
		g_hash_table_unref(system_default_db);
	if (user_default_db)
		g_hash_table_unref(user_default_db);
	if (logging_leveldb)
		leveldb_close(logging_leveldb);

	_D("logging_exit");
	return RESOURCED_ERROR_NONE;
}
