/*
 * resourced
 *
 * Copyright (c) 2015 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
*/
#include <sys/types.h>
#include <signal.h>
#include <unistd.h>

#include "trace.h"
#include "macro.h"
#include "module.h"
#include "module-data.h"
#include "lowmem-handler.h"
#include "notifier.h"
#include "procfs.h"
#include "cgroup.h"
#include "const.h"
#include "proc-common.h"
#include "memory-common.h"
#include "util.h"
#include "smaps.h"
#include "config-parser.h"
#include "resourced-helper-worker.h"
#include "fd-handler.h"
#include "dbus-handler.h"

#define MEM_CONF_FILE                   RD_CONFIG_FILE(memory)
#define MEMLIMIT_LOG_FILE			MEMPS_LOG_PATH"/memps_memlimit"
#define MEMLIMIT_CONFIG_SECTION	"MemLimit"
#define MEMLIMIT_CONFIG_TRIGGER	"MemLimitTrigger"
#define MEMLIMIT_CONFIG_SERVICE	"MemLimitService"
#define MIN_LIMIT_VALUE		1 /* MB */

enum mem_limit_type {
	/* check only swap usage also */
	MEM_LIMIT_NONE,
	/* register OOM event and control application in kernel side */
	MEM_LIMIT_OOM,
	/* register threshold event and control application in resourced*/
	MEM_LIMIT_TRHESHOLD,
};

static enum mem_limit_type mem_limit;


static GHashTable *memory_limit_hash;
static char *registerpath;
static int mem_service_limit;

struct memory_limit_event {
	fd_handler_h fdh;
	int fd;
	unsigned int threshold;
	char *path;
};

struct memory_info {
	pid_t pid;
	unsigned int size;
};

struct memory_limit_log {
	pid_t pid;
	char *appname;
	char *cgdir;
};

static int get_pid_use_max_memory(GArray *pids_array, unsigned int *max_mem)
{
	int i, ret;
	struct memory_info mi = {0, 0};

	if (!pids_array->len)
		return RESOURCED_ERROR_NO_DATA;

	for (i = 0; i < pids_array->len; i++) {
		pid_t mpid = 0;
		unsigned int size;

		mpid = g_array_index(pids_array, pid_t, i);
		if (!mpid)
			continue;

		ret = proc_get_mem_usage(mpid, &size);
		if (ret != RESOURCED_ERROR_NONE)
			continue;

		_D("%d used %d memory", mpid, size);

		if (size > mi.size) {
			mi.pid = mpid;
			mi.size = size;
		}
	}
	*max_mem = mi.size;
	_D("%d used max %d memory", mi.pid, mi.size);
	return mi.pid;
}

static pid_t get_main_pid(const char *dir, unsigned int *max_mem)
{
	GArray *pids_array = NULL;
	_cleanup_free_ char *path = NULL;
	int ret;
	pid_t main_pid;

	ret = asprintf(&path, "%s/", dir);
	if (ret < 0)
		return -ENOMEM;

	ret = cgroup_get_pids(path, &pids_array);
	if (ret < 0 || !pids_array->len) {
		g_array_free(pids_array, true);
		return RESOURCED_ERROR_NO_DATA;
	}

	main_pid = get_pid_use_max_memory(pids_array, max_mem);
	g_array_free(pids_array, true);
	return main_pid;
}

static void memory_limit_hash_destroy(gpointer data)
{
	struct memory_limit_event *mle = (struct memory_limit_event *)data;
	close(mle->fd);
	free(mle->path);

	/* NB: `mle->fdh` is NOT cleaned up here. This is because the removal
	 * of `memory_limit_event` only happens in two cases:
	 *
	 *  a) at the end of `lowmem_limit_cb`, which returns `false`
	 *     on removal, which results in `fd_handler` getting cleaned
	 *     up independently (see `channel_changed` in `fd-handler.c`)
	 *
	 *  b) at exit, by which time GIO channels should've already been
	 *     cleaned up (due to `g_main_loop_quit` et al.). */

	free(mle);
}

static int make_memlimit_logs(void *data)
{
	struct memory_limit_log *mlog = (struct memory_limit_log *)data;

	if (!mlog)
		return RESOURCED_ERROR_NO_DATA;

	if (access("/opt/etc/.debugmode", F_OK) == 0) {
		const char *argv[7] = {"/usr/bin/log_dump", "-d", "-v", "-t", "-o", NULL, NULL};
		_cleanup_free_ char *path = NULL;

		if (asprintf(&path, "log_dump_memlimit_%s", mlog->appname) < 0)
			return RESOURCED_ERROR_OUT_OF_MEMORY;

		argv[5] = path;
		exec_cmd(ARRAY_SIZE(argv), argv);
	}

	make_memps_log(MEMLIMIT_LOG_FILE, mlog->pid, mlog->appname);
	return RESOURCED_ERROR_NONE;
}

static void memlimit_finish_cb(void *data, int ret)
{
	struct memory_limit_log *mlog = (struct memory_limit_log *)data;

	if (!mlog || !mlog->cgdir)
		return;

	/*
	 * send sigabt signal
	 * If debug is enabled, it makes crash popup with log files.
	 * otherwise, kill this application silently.
	 */
	if (mem_limit == MEM_LIMIT_TRHESHOLD)
		kill(mlog->pid, SIGABRT);
	/*
	 * If oom control is disabled, kernel context of this process waits oom_waitq.
	 * It is why resource enable oom control again
	 * since kernel can find and kill victim on cgroup oom.
	 */
	else if (mem_limit == MEM_LIMIT_OOM)
		cgroup_write_node_uint32(mlog->cgdir, MEMCG_OOM_CONTROL_PATH, 0);

	if (mlog->appname)
		free(mlog->appname);
	if (mlog->cgdir)
		free(mlog->cgdir);
	free(mlog);
}

int lowmem_limit_broadcast(pid_t pid)
{
	char *pa[2];
	char pidbuf[MAX_DEC_SIZE(int)];
	int ret;
	char appname[PROC_NAME_MAX];
	struct proc_app_info *pai = NULL;

	snprintf(pidbuf, sizeof(pidbuf), "%d", pid);
	pa[0] = pidbuf;

	pai = find_app_info(pid);
	if (pai) {
		pa[1] = pai->appid;
	} else {
		ret = proc_get_cmdline(pid, appname);
		if (ret < 0) {
			_E("Failed to get cmdline basename of pid(%d)", pid);
			return ret;
		}
		pa[1] = appname;
	}

	ret = d_bus_broadcast_signal(RESOURCED_PATH_OOM,
		    RESOURCED_INTERFACE_OOM, SIGNAL_OOM_MEMLIMIT_EVENT, "is", pa);
	if (ret < 0)
		_E("Fail to broadcast dbus signal with pid(%d), appname(%s)", pid, appname);
	return ret;
}

static bool lowmem_limit_cb(int fd, void *data)
{
	static char *prev_dir;
	char *cg_dir = (char *)data;
	int ret;
	unsigned int usage = 0, max_mem = 0, anon_usage = 0;
	uint64_t dummy_efd;
	pid_t main_pid;
	char appname[PROC_NAME_MAX];
	struct memory_limit_event *mle;
	struct memory_limit_log *mlog = NULL;
	_cleanup_free_ struct cgroup_memory_stat *mem_stat = NULL;

	_I("receive lowmem limit about %s", cg_dir);
	mle = g_hash_table_lookup(memory_limit_hash, cg_dir);
	if (!mle) {
		_E("invalid event\n");
		return false;
	}

	ret = read(fd, &dummy_efd, sizeof(dummy_efd));
	if (ret < 0) {
		_E("wrong eventfd %s\n", cg_dir);
		goto remove_mle;
	}

	/*
	 * The eventfd about threshold can be triggered in three cases.
	 * Firstly, if there is no cgroup any longer, eventfd will be sent.
	 * And in both entering threshold and escaping threshold,
	 * it will be also triggered.
	 * Therefore, user space should check all cases.
	 */
	if (prev_dir == cg_dir) {
		_D("this event means %s cgroup escaped low memory status.\n", cg_dir);
		return true;
	}
	prev_dir = cg_dir;

	if (access(cg_dir, F_OK) == -1) {
		_D("there is no (%s) cgroup any longer, removing it", cg_dir);
		goto remove_mle;
	}

	ret = cgroup_read_node_uint32(cg_dir, MEMCG_SWAP_USAGE, &usage);
	if (ret)
		ret = cgroup_read_node_uint32(cg_dir, MEMCG_USAGE, &usage);

	if (ret) {
		_D("there is no (%s) cgroup any longer, removed it", cg_dir);
		goto remove_mle;
	}

	if (usage < mle->threshold) {
		_D("(%s) cgroup escaped low memory status. usage(%d), threshold(%d)",
			cg_dir, usage, mle->threshold);
		return true;
	}
	ret = cgroup_get_memory_stat(cg_dir, &mem_stat);
	if (ret) {
		_D("fail to get memory status : %s", cg_dir);
		goto remove_mle;
	}

	mlog = (struct memory_limit_log *)malloc(sizeof(struct memory_limit_log));
	if (!mlog) {
		_D("out of memory");
		return true;
	}

	if (mem_limit == MEM_LIMIT_TRHESHOLD) {
		anon_usage = mem_stat->value[CGROUP_MEMORY_STAT_RSS] +
				    mem_stat->value[CGROUP_MEMORY_STAT_SWAP];
		if (anon_usage < mle->threshold) {
			_D("(%s) cgroup escaped low memory status. usage(%d), anon usage (%d), threshold(%d)",
			    cg_dir, usage, anon_usage, mle->threshold);
			free(mlog);
			return true;
		}

		main_pid = get_main_pid(cg_dir, &max_mem);
		if (main_pid <= 0) {
			_D("there is no victim, removed cgroup : %s", cg_dir);
			goto remove_mle;
		}

		ret = proc_get_cmdline(main_pid, appname);
		if (ret)
			_D("fail to get appname");

		_E("kill %d (name %s, memory %d) by exceeding the threshold (%d)",
		    main_pid, ret ? NULL : appname, max_mem, mle->threshold);
	} else {
		char *filename;

		main_pid = get_main_pid(cg_dir, &max_mem);
		if (main_pid <= 0) {
			_D("there is no victim, removed cgroup : %s", cg_dir);
			goto remove_mle;
		}

		filename = strrchr(cg_dir, '/');
		snprintf(appname, sizeof(appname), "kerneloom_%s", filename ? filename+1 : cg_dir);
		_E("some application was crashed by kernel when it exceeded the threshold (%d)",
		    mle->threshold);
	}
	mlog->pid = main_pid;
	mlog->appname = strdup(appname);
	mlog->cgdir = strdup(cg_dir);

	/*
	 * If there are both appname and cgdir in mlog sturcture,
	 * it can register to request helper worker.
	 * But, otherwise due to low memory,
	 * it has to call callback function directly in order to kill application or clean up memory.
	 */
	if (mlog->appname && mlog->cgdir) {
		lowmem_limit_broadcast(main_pid);
		request_helper_worker(MAKE_LOGS, mlog, make_memlimit_logs, memlimit_finish_cb);
	} else {
		memlimit_finish_cb(mlog, RESOURCED_ERROR_OUT_OF_MEMORY);
	}
	return true;

remove_mle:
	g_hash_table_remove(memory_limit_hash, cg_dir);
	if (mlog)
		free(mlog);
	return false;
}

void lowmem_reassign_limit(const char *dir, unsigned int limit)
{
	int fd, ret;
	gpointer hash_entry;
	fd_handler_h fdh = NULL;
	struct memory_limit_event *mle = NULL;
	char buf[MAX_DEC_SIZE(int)] = {0};
	unsigned int prev;

	if (memory_limit_hash) {
		hash_entry = g_hash_table_lookup(memory_limit_hash, dir);
		if (hash_entry) {
			mle = (struct memory_limit_event *)hash_entry;
			if (mle->threshold == limit) {
				_D("%s has already registered", dir);
				return;
			}
		}
	} else {
		memory_limit_hash = g_hash_table_new_full(g_str_hash, g_str_equal,
				NULL, memory_limit_hash_destroy);
		if (!memory_limit_hash) {
			_E("Failed to create hash table");
			return;
		}
	}

	ret = cgroup_read_node_uint32(dir, MEMCG_LIMIT_PATH, &prev);
	if (ret) {
		_E("Failed to get %s from %s", MEMCG_LIMIT_PATH, dir);
		return;
	}

	/*
	 * Kernel always compares new value with memswlimit.
	* So, if new value is smaller than memswlimit, memlimit should be set at first.
	* Otherwise, memswlimit should be set before setting memlimit.
	*/
	switch (mem_limit) {
	case MEM_LIMIT_NONE:
		/*
		 * only set memory limitation to memory.memsw.limit_in_bytes
		 * because almost memcg does't have a limitation about swap usage
		 */
		if (prev > limit) {
			cgroup_write_node_uint32(dir, MEMCG_LIMIT_PATH, limit);
			cgroup_write_node_uint32(dir, MEMCG_SWAP_LIMIT_PATH, limit);
		} else if (prev < limit) {
			cgroup_write_node_uint32(dir, MEMCG_SWAP_LIMIT_PATH, limit);
			cgroup_write_node_uint32(dir, MEMCG_LIMIT_PATH, limit);
		} else {
			cgroup_write_node_uint32(dir, MEMCG_SWAP_LIMIT_PATH, limit);
		}
		return;
	case MEM_LIMIT_OOM:
		/*
		 * Previous kernel didn't support to disable oom control in user space.
		 * But recent kernel over 4.0 version allows to disable oom.
		 * If kernel supports it, resourced can receive oom event before killing application in kernel side
		 * and collect log files about problem situation.
		 * Otherwise, MEM_LIMIT_NONE and MEM_LIMIT_OOM will be same.
		 */
		cgroup_write_node_uint32(dir, MEMCG_OOM_CONTROL_PATH, 1);
		if (prev > limit) {
			cgroup_write_node_uint32(dir, MEMCG_LIMIT_PATH, limit);
			cgroup_write_node_uint32(dir, MEMCG_SWAP_LIMIT_PATH, limit);
		} else if (prev < limit) {
			cgroup_write_node_uint32(dir, MEMCG_SWAP_LIMIT_PATH, limit);
			cgroup_write_node_uint32(dir, MEMCG_LIMIT_PATH, limit);
		} else {
			cgroup_write_node_uint32(dir, MEMCG_SWAP_LIMIT_PATH, limit);
		}
		break;
	case MEM_LIMIT_TRHESHOLD:
		snprintf(buf, sizeof(buf), "%d", limit);
		/*
		 * need to set limitation value again for preventing kernel OOM
		 * it is larger than original limitation value
		 */
		cgroup_write_node_uint32(dir, MEMCG_LIMIT_PATH, (limit * 1.2));
		break;
	default:
		return;
	}

	if (mle) {
		mle->threshold = limit;
		return;
	}

	fd = memcg_set_eventfd(dir, registerpath, buf);
	if (fd) {
		mle = calloc(1, sizeof(struct memory_limit_event));
		if (!mle) {
			_E("out of memory");
			return;
		}
		mle->fd = fd;
		mle->path = (char *)strdup(dir);
		if (!mle->path) {
			_E("out of memory");
			free(mle);
			return;
		}
		mle->threshold = limit;
		add_fd_read_handler(fd, lowmem_limit_cb, mle->path, NULL, &fdh);
		mle->fdh = fdh;
		g_hash_table_insert(memory_limit_hash, (gpointer)mle->path,
		    (gpointer)mle);
	}
	return;
}

int lowmem_limit_move_cgroup(struct proc_app_info *pai)
{
	GSList *iter = NULL;
	_cleanup_free_ char *path = NULL;
	int ret;

	if (!pai)
		return RESOURCED_ERROR_NO_DATA;

	if (!pai->memory.use_mem_limit)
		return RESOURCED_ERROR_NO_DATA;

	ret = asprintf(&path, "%s/%s", LOWMEM_MEMLIMIT_CGROUP, pai->appid);
	if (ret < 0) {
		_E("not enough memory");
		return RESOURCED_ERROR_OUT_OF_MEMORY;
	}
	cgroup_write_node_uint32(path, CGROUP_FILE_NAME, pai->main_pid);
	if (pai->childs) {
		gslist_for_each_item(iter, pai->childs)
			cgroup_write_node_uint32(path, CGROUP_FILE_NAME,
			    GPOINTER_TO_PID(iter->data));
	}
	return RESOURCED_ERROR_NONE;
}

static int memlimit_config_parse(struct parse_result *result, void *user_data)
{
	int ret;
	unsigned int usage;

	/* Single config section is expected */
	if (!result->section)
		return RESOURCED_ERROR_NONE;

	if (strncmp(result->section, MEMLIMIT_CONFIG_SECTION,
	    sizeof(MEMLIMIT_CONFIG_SECTION)))
		return RESOURCED_ERROR_NONE;

	if (!result->name || !result->value)
		return RESOURCED_ERROR_NONE;

	if (!strncmp(result->name, MEMLIMIT_CONFIG_TRIGGER,
	    sizeof(MEMLIMIT_CONFIG_TRIGGER))) {
		/*
		 * allowed strings: threshold, oom
		 */
		if (!strncmp(result->value, "thre", 4)) {
			mem_limit = MEM_LIMIT_TRHESHOLD;
		} else if (!strncmp(result->value, "oom", 3)) {
			mem_limit = MEM_LIMIT_OOM;
		} else {
			mem_limit = MEM_LIMIT_NONE;
			return RESOURCED_ERROR_NONE;
		}

		if (mem_limit == MEM_LIMIT_OOM) {
			/*
			 * resourced registers memory.oom_control.
			 * kernel will kill oom application and
			 * resourced will receive event after application is terminated.
			 * But instead, resourced can select the kill signals,
			 * if kernel is supported.
			 * one of SIGKILL, SIGTRAP, SIGTERM
			 */
			 registerpath = MEMCG_OOM_CONTROL_PATH;
		} else {
			/*
			 * check swap status when resourced registers threshold event.
			 * if the device support swap,
			 * resourced registers memory.memsw.usage_in_bytes,
			 * otherwise, register memory.limit_in_bytes.
			 */
			ret = cgroup_read_node_uint32(LOWMEM_ROOT_CGROUP,
			    MEMCG_SWAP_USAGE, &usage);
			if (ret == RESOURCED_ERROR_NONE)
				registerpath = MEMCG_SWAP_USAGE;
			else
				registerpath = MEMCG_USAGE;
		}
		_I("Control %s event with %s node about memcg", result->value, registerpath);
	} else if (!strncmp(result->name, MEMLIMIT_CONFIG_SERVICE,
	    sizeof(MEMLIMIT_CONFIG_SERVICE))) {
		if (mem_limit != MEM_LIMIT_NONE)
			mem_service_limit = atoi(result->value);
	}
	return RESOURCED_ERROR_NONE;
}

void lowmem_limit_set(pid_t pid, unsigned int limit)
{
	char *cgpath, appname[PROC_NAME_MAX];
	struct proc_app_info *pai = NULL;
	_cleanup_free_ char *path = NULL;
	GSList *iter = NULL;
	int ret;
	unsigned int totalramMB = KBYTE_TO_MBYTE(lowmem_get_ktotalram());

	if (limit < MIN_LIMIT_VALUE || limit > totalramMB) {
		_E("It's meaningless to set memory limit with size (%d)", limit);
		return;
	}

	pai = find_app_info(pid);
	if (pai) {
		cgpath = pai->appid;
	} else {
		ret = proc_get_cmdline(pid, appname);
		if (ret < 0) {
			_E("Failed to get cmdline basename of pid(%d)", pid);
			return;
		}
		cgpath = appname;
	}

	ret = asprintf(&path, "%s/%s", LOWMEM_MEMLIMIT_CGROUP, cgpath);
	if (ret < 0) {
		_E("not enough memory");
		return;
	}

	ret = cgroup_make_subdir(LOWMEM_MEMLIMIT_CGROUP, cgpath, NULL);
	if (ret < 0) {
		_E("Failed to create cgroup subdir '%s/%s'",
		   LOWMEM_MEMLIMIT_CGROUP, cgpath);
		return;
	}

	lowmem_reassign_limit(path, MBYTE_TO_BYTE(limit));

	ret = cgroup_write_node_uint32(path, CGROUP_FILE_NAME, pid);
	if (ret < 0) {
		_E("Failed to write pid(%d) to '%s/%s'", pid, path, CGROUP_FILE_NAME);
		return;
	}

	if (!pai)
		return;

	pai->memory.use_mem_limit = true;

	if (!pai->childs)
		return;

	gslist_for_each_item(iter, pai->childs) {
		pid_t child = GPOINTER_TO_PID(iter->data);

		ret = cgroup_write_node_uint32(path, CGROUP_FILE_NAME, child);
		if (ret < 0)
			_E("Failed to write pid(%d) to '%s/%s', ignoring for other children",
			   child, path, CGROUP_FILE_NAME);
	}
}

static int lowmem_limit_service(void *data)
{
	struct proc_status *ps = (struct proc_status *)data;

	if (mem_service_limit)
		lowmem_limit_set(ps->pid, mem_service_limit);
	return RESOURCED_ERROR_NONE;
}

void lowmem_limit_init(void)
{
	/* Load configuration */
	config_parse(MEM_CONF_FILE, memlimit_config_parse,
			NULL);
	if (mem_service_limit)
		register_notifier(RESOURCED_NOTIFIER_SERVICE_LAUNCH, lowmem_limit_service);
}

void lowmem_limit_exit(void)
{
	if (memory_limit_hash)
		g_hash_table_destroy(memory_limit_hash);

	if (mem_service_limit)
		unregister_notifier(RESOURCED_NOTIFIER_SERVICE_LAUNCH, lowmem_limit_service);
}
