/*
 * resourced
 *
 * Copyright (c) 2015-2019 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
*/
#include <sys/types.h>
#include <signal.h>
#include <unistd.h>

#include "trace.h"
#include "macro.h"
#include "module.h"
#include "module-data.h"
#include "lowmem-handler.h"
#include "notifier.h"
#include "procfs.h"
#include "cgroup.h"
#include "const.h"
#include "proc-common.h"
#include "memory-common.h"
#include "util.h"
#include "smaps.h"
#include "config-parser.h"

static GSource *lowmem_checksystemd_timer;

#define CHECKLIMIT_FIRST_INTERVAL	  30
#define CHECKLIMIT_INTERVAL	  	  24*3600
#define MEMORY_SYSTEMD_SYSTEM	"/sys/fs/cgroup/memory/system.slice"
#define MEMORY_SYSTEMD_USER	"/sys/fs/cgroup/memory/user.slice"

/*
 * Resourced couldn't control any systemd services
 * even though some systemd services set either the memory limit or the oom score value.
 * If resourced controls all processes including systemd services,
 * it has to check system.slice of the memcg.
 */
static int search_systemd_cgroup(const char *dir)
{
	_cleanup_closedir_ DIR *d = NULL;
	struct dirent *de;
	int ret;
	int rootswappiness, memcgswappiness;
	int changeswappiness = -1;

	d = opendir(dir);
	if (!d) {
		_D("Failed to open dir: %s", dir);
		return -errno;
	}

	ret = cgroup_read_node_int32(LOWMEM_ROOT_CGROUP, MEMCG_SWAPNESS_PATH,
			&rootswappiness);
	if (ret != RESOURCED_ERROR_NONE)
		return -errno;

	ret = cgroup_read_node_int32(dir, MEMCG_SWAPNESS_PATH,
			&memcgswappiness);
	if (ret != RESOURCED_ERROR_NONE)
		return -errno;

	if ((rootswappiness >=0) && (rootswappiness != memcgswappiness)) {
		changeswappiness = rootswappiness;
		ret = cgroup_write_node_uint32(dir,
			MEMCG_SWAPNESS_PATH, changeswappiness);
		if (ret != RESOURCED_ERROR_NONE) {
			_I("failed to write %s %d to %s the",
				MEMCG_SWAPNESS_PATH, changeswappiness, dir);
			return -errno;
		}
	}

	/*
	 * set memory limit
	 * The purpose of this operation is to make debug information
	 * when systemd services spent much memory more than limit.
	 * And it can even track the amount of swapped memory.
	 */
	FOREACH_DIRENT(de, d, return -errno) {
		_cleanup_free_ char *path = NULL;

		unsigned int limit;

		if (de->d_type != DT_DIR)
			continue;

		ret = asprintf(&path, "%s/%s", dir, de->d_name);
		if (ret < 0)
			return -ENOMEM;

		ret = cgroup_read_node_uint32(path, MEMCG_LIMIT_PATH, &limit);
		if (ret != RESOURCED_ERROR_NONE ||limit <= 0)
			continue;

		if (changeswappiness >= 0) {
			ret = cgroup_write_node_uint32(path,
					MEMCG_SWAPNESS_PATH, changeswappiness);
			if (ret != RESOURCED_ERROR_NONE)
				_I("failed to write %s %d to %s the",
					MEMCG_SWAPNESS_PATH, changeswappiness, path);
		}

		lowmem_reassign_limit(path, limit);
	}
	return RESOURCED_ERROR_NONE;
}

static gboolean checksystemd_cb(gpointer data)
{
	search_systemd_cgroup(MEMORY_SYSTEMD_SYSTEM);
	search_systemd_cgroup(MEMORY_SYSTEMD_USER);

	static bool first = true;
	if (first) {
		first = false; // pop this cherry
		g_source_destroy(lowmem_checksystemd_timer);
		lowmem_checksystemd_timer = g_timeout_source_new_seconds(CHECKLIMIT_INTERVAL);
		g_source_set_callback(lowmem_checksystemd_timer, checksystemd_cb, NULL, NULL);
		g_source_attach(lowmem_checksystemd_timer, NULL);
		return false;
	}
	return true;
}

static int booting_done(void *data)
{
	lowmem_checksystemd_timer =
		g_timeout_source_new_seconds(CHECKLIMIT_FIRST_INTERVAL);
	g_source_set_callback(lowmem_checksystemd_timer, checksystemd_cb, NULL, NULL);
	g_source_attach(lowmem_checksystemd_timer, NULL);

	return RESOURCED_ERROR_NONE;
}

void lowmem_system_init()
{
	register_notifier(RESOURCED_NOTIFIER_BOOTING_DONE, booting_done);
}

void lowmem_system_exit(void)
{
	if (lowmem_checksystemd_timer) {
		g_source_destroy(lowmem_checksystemd_timer);
		lowmem_checksystemd_timer = NULL;
	}
	unregister_notifier(RESOURCED_NOTIFIER_BOOTING_DONE, booting_done);
}
