/*
 * resourced
 *
 * Copyright (c) 2014 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * @file memcontrol.c
 *
 * @desc structure and operation for memory cgroups
 *
 * Copyright (c) 2014 Samsung Electronics Co., Ltd. All rights reserved.
 *
 */

#include <string.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/eventfd.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "resourced.h"
#include "trace.h"
#include "macro.h"
#include "memory-common.h"
#include "cgroup.h"
#include "module.h"
#include "util.h"

#include <stdlib.h>

#define BUF_MAX				1023

static void memcg_info_show(struct memcg_info *mi)
{
	int i;
	_D("======================================");
	_D("memcg_info->name = %s", mi->name);
	_D("memcg_info->limit_ratio = %.f", mi->limit_ratio);
	_D("memcg_info->limit = %u", mi->limit);
	_D("memcg_info->oomleave = %u", mi->oomleave);
	for (i = 0; i < LOWMEM_MAX_LEVEL; i++)
		_D("memcg_info->threshold = %u", mi->threshold[i]);
	_D("memcg_info->threshold_leave = %u", mi->threshold_leave);
	_D("memcg_info->evfd = %d", mi->evfd);
}

void memcg_show(struct memcg *memcg)
{
	GSList *iter = NULL;
	memcg_info_show(memcg->info);
	gslist_for_each_item(iter, memcg->cgroups) {
		struct memcg_info *mi =
			(struct memcg_info *)(iter->data);
		memcg_info_show(mi);
	}
}

void memcg_init(struct memcg *memcg)
{
	memcg->use_hierarchy = MEMCG_DEFAULT_USE_HIERARCHY;
	memcg->info = NULL;
	memcg->cgroups = NULL;
}

int memcg_get_anon_usage(char *memcg, unsigned int *anon_usage)
{
	int r;
	_cleanup_free_ struct cgroup_memory_stat *mem_stat = NULL;

	r = cgroup_get_memory_stat(memcg, &mem_stat);
	if (r) {
		_D("fail to get memory status : %s", memcg);
		return r;
	}

	*anon_usage = mem_stat->value[CGROUP_MEMORY_STAT_INACTIVE_ANON] +
				mem_stat->value[CGROUP_MEMORY_STAT_ACTIVE_ANON];
	return 0;
}

int memcg_get_swap_usage(char *memcg, unsigned int *usage)
{
	int r;
	_cleanup_free_ struct cgroup_memory_stat *mem_stat = NULL;

	r = cgroup_get_memory_stat(memcg, &mem_stat);
	if (r) {
		_D("fail to get memory status : %s", memcg);
		return r;
	}

	*usage = mem_stat->value[CGROUP_MEMORY_STAT_SWAP];
	return 0;
}

/*
 * Usage example:
 *	int i;
 *	pid_t pid;
 *	GArray *pids_array = g_array_new(false, false, sizeof(pid_t));
 *
 *	memcg_get_pids(mi, pids_array);
 *
 *	for (i=0; i < pids_array->len; i++)
 *		_D("pid_t: %d", g_array_index(pids_array, pid_t, i));
 *	g_array_free(pids_array, TRUE);
 *
 */
int memcg_get_pids(const char *name, GArray *pids)
{
	FILE *f;
	pid_t tpid;
	char buf[MAX_PATH_LENGTH] = {0, };

	if (pids == NULL)
		return RESOURCED_ERROR_FAIL;

	snprintf(buf, sizeof(buf), "%s/cgroup.procs", name);

	f = fopen(buf, "r");
	if (!f) {
		_E("%s open failed, %m", buf);
		return RESOURCED_ERROR_FAIL;
	}

	while (fgets(buf, 32, f) != NULL) {
		tpid = atoi(buf);
		g_array_append_val(pids, tpid);
	}
	fclose(f);

	return RESOURCED_ERROR_NONE;
}

/*
 * From memory.txt kernel document,
 * To register a event for memcg, an application must:
 *  - create an eventfd using eventfd(2);
 * - open a node of memory cgroup
 * - write string like "<event_fd> <opened fd> <value>" to cgroup.event_control
 *
 * Current memory cgroup supports eventfd about only
 * usage_in_byte, oom_control and pressure_level.
 */
int memcg_set_eventfd(const char *memcg, const char *event, char *value)
{
	_cleanup_close_ int mcgfd = -1;
	_cleanup_close_ int cgfd = -1;
	int evfd, res = 0, sz, ret = -1;
	char buf[PATH_MAX] = {0,};

	/* create an eventfd using eventfd(2)*/
	evfd = eventfd(0, 0);
	ret = fcntl(evfd, F_SETFL, O_NONBLOCK);
	if (ret < 0)
		return RESOURCED_ERROR_FAIL;

	/* open a node of memory cgroup */
	snprintf(buf, PATH_MAX, "%s/%s", memcg, MEMCG_EVENTFD_CONTROL);
	cgfd = open(buf, O_WRONLY);
	if (cgfd < 0) {
		const int saved_errno = errno;
		_E("open event_control failed");
		errno = saved_errno;
		return RESOURCED_ERROR_FAIL;
	}

	snprintf(buf, PATH_MAX, "%s/%s", memcg, event);
	mcgfd = open(buf, O_RDONLY);
	if (mcgfd < 0) {
		const int saved_errno = errno;
		_E("open memory control failed");
		errno = saved_errno;
		return RESOURCED_ERROR_FAIL;
	}

	_D("%s %s %s registerd", memcg, event, value);
	/* write string like "<event_fd> <opened fd> <value>" to cgroup.event_control */
	sz = snprintf(buf, PATH_MAX, "%d %d %s", evfd, mcgfd, value);
	sz += 1;
	res = write(cgfd, buf, sz);
	if (res != sz) {
		int saved_errno = errno;
		_E("write cgfd failed : %d", res);
		errno = saved_errno;
		return RESOURCED_ERROR_FAIL;
	}
	return evfd;
}
