/*
 * resourced
 *
 * Copyright (c) 2012 - 2019 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * @file vmpressure-lowmem-handler.c
 *
 * @desc lowmem handler using memcgroup
 *
 * Copyright (c) 2014 Samsung Electronics Co., Ltd. All rights reserved.
 *
 */

#include <stdio.h>
#include <fcntl.h>
#include <assert.h>
#include <limits.h>
#include <vconf.h>
#include <unistd.h>
#include <time.h>
#include <limits.h>
#include <dirent.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/shm.h>
#include <sys/sysinfo.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <ctype.h>
#include <bundle.h>
#include <eventsystem.h>
#include <malloc.h>

#include "trace.h"
#include "cgroup.h"
#include "lowmem-handler.h"
#include "proc-common.h"
#include "procfs.h"
#include "freezer.h"
#include "resourced.h"
#include "macro.h"
#include "notifier.h"
#include "config-parser.h"
#include "module.h"
#include "swap-common.h"
#include "cgroup.h"
#include "memory-common.h"
#include "heart-common.h"
#include "proc-main.h"
#include "dbus-handler.h"
#include "util.h"
#include "fd-handler.h"
#include "resourced-helper-worker.h"

#define LOWMEM_NO_LIMIT			0
#define LOWMEM_THRES_INIT		0

#define MEMCG_MOVE_CHARGE_PATH		"memory.move_charge_at_immigrate"
#define MEMCG_EVENTFD_MEMORY_PRESSURE	"memory.pressure_level"
#define MEM_CONF_FILE                   RD_CONFIG_FILE(memory)
#define MEM_VIP_SECTION			"VIP_PROCESS"
#define MEM_VIP_PREDEFINE		"PREDEFINE"
#define MEM_POPUP_SECTION		"POPUP"
#define MEM_POPUP_STRING		"oom_popup"
#define MEM_BG_RECLAIM_SECTION	"BackgroundReclaim"
#define MEM_BG_RECLAIM_STRING		"AfterScreenDim"

#define BUF_MAX				1024
#define MAX_VICTIMS_BETWEEN_CHECK	3
#define MAX_PROACTIVE_HIGH_VICTIMS	4
#define OOM_TIMER_INTERVAL		2
#define OOM_KILLER_PRIORITY		-20
#define MAX_MEMPS_LOGS                  50
#define NUM_RM_LOGS			5
#define THRESHOLD_MARGIN		10 /* MB */

#define MEM_SIZE_64			64  /* MB */
#define MEM_SIZE_256			256 /* MB */
#define MEM_SIZE_448			448 /* MB */
#define MEM_SIZE_512			512 /* MB */
#define MEM_SIZE_768			768 /* MB */
#define MEM_SIZE_1024			1024 /* MB */

/* thresholds for 64M RAM*/
#define PROACTIVE_64_THRES			10 /* MB */
#define PROACTIVE_64_LEAVE			30 /* MB */
#define MEMCG_MEMORY_64_THRES_SWAP		15 /* MB */
#define MEMCG_MEMORY_64_THRES_LOW		8  /* MB */
#define MEMCG_MEMORY_64_THRES_MEDIUM		5  /* MB */
#define MEMCG_MEMORY_64_THRES_LEAVE		8  /* MB */

/* thresholds for 256M RAM */
#define PROACTIVE_256_THRES			50 /* MB */
#define PROACTIVE_256_LEAVE			80 /* MB */
#define MEMCG_MEMORY_256_THRES_SWAP		40 /* MB */
#define MEMCG_MEMORY_256_THRES_LOW		20 /* MB */
#define MEMCG_MEMORY_256_THRES_MEDIUM		10 /* MB */
#define MEMCG_MEMORY_256_THRES_LEAVE		20 /* MB */

/* threshold for 448M RAM */
#define PROACTIVE_448_THRES			80 /* MB */
#define PROACTIVE_448_LEAVE			100 /* MB */
#define MEMCG_MEMORY_448_THRES_SWAP		100 /* MB */
#define MEMCG_MEMORY_448_THRES_LOW		50  /* MB */
#define MEMCG_MEMORY_448_THRES_MEDIUM		40  /* MB */
#define MEMCG_MEMORY_448_THRES_LEAVE		60  /* MB */

/* threshold for 512M RAM */
#define PROACTIVE_512_THRES			80 /* MB */
#define PROACTIVE_512_LEAVE			100 /* MB */
#define MEMCG_MEMORY_512_THRES_SWAP		100 /* MB */
#define MEMCG_MEMORY_512_THRES_LOW		50  /* MB */
#define MEMCG_MEMORY_512_THRES_MEDIUM		40  /* MB */
#define MEMCG_MEMORY_512_THRES_LEAVE		60  /* MB */

/* threshold for 768 RAM */
#define PROACTIVE_768_THRES			100 /* MB */
#define PROACTIVE_768_LEAVE			120 /* MB */
#define MEMCG_MEMORY_768_THRES_SWAP		150 /* MB */
#define MEMCG_MEMORY_768_THRES_LOW		100  /* MB */
#define MEMCG_MEMORY_768_THRES_MEDIUM		60  /* MB */
#define MEMCG_MEMORY_768_THRES_LEAVE		100  /* MB */

/* threshold for more than 1024M RAM */
#define PROACTIVE_1024_THRES			150 /* MB */
#define PROACTIVE_1024_LEAVE			300 /* MB */
#define MEMCG_MEMORY_1024_THRES_SWAP		300 /* MB */
#define MEMCG_MEMORY_1024_THRES_LOW		200 /* MB */
#define MEMCG_MEMORY_1024_THRES_MEDIUM		100 /* MB */
#define MEMCG_MEMORY_1024_THRES_LEAVE		150 /* MB */

/* threshold for more than 2048M RAM */
#define PROACTIVE_2048_THRES			200 /* MB */
#define PROACTIVE_2048_LEAVE			500 /* MB */
#define MEMCG_MEMORY_2048_THRES_SWAP		300 /* MB */
#define MEMCG_MEMORY_2048_THRES_LOW		200 /* MB */
#define MEMCG_MEMORY_2048_THRES_MEDIUM		160 /* MB */
#define MEMCG_MEMORY_2048_THRES_LEAVE		300 /* MB */

static unsigned proactive_threshold;
static unsigned proactive_leave;
static unsigned lmk_start_threshold;

static char *event_level = MEMCG_DEFAULT_EVENT_LEVEL;

/**
 * Resourced Low Memory Killer
 * NOTE: planned to be moved to a separate file.
 */
/*-------------------------------------------------*/
#define OOM_TIMER_INTERVAL_SEC	2
#define LMW_LOOP_WAIT_TIMEOUT_MSEC	OOM_TIMER_INTERVAL_SEC*(G_USEC_PER_SEC)
#define LMW_RETRY_WAIT_TIMEOUT_MSEC	(G_USEC_PER_SEC)

struct lowmem_control {
	/*
	 * For each queued request the following properties
	 * are required with two exceptions:
	 *  - status is being set by LMK
	 *  - callback is optional
	 */
	/* Processing flags*/
	unsigned int flags;
	/* Indictator for OOM score of targeted processes */
	enum lmk_type type;
	/* Desired size to be restored - level to be reached (MB)*/
	unsigned int size;
	/* Max number of processes to be considered */
	unsigned int count;
	/* Memory reclaim status */
	int status;
	/*
	 * Optional - if set, will be triggered by LMK once the request
	 * is handled.
	 */
	void (*callback) (struct lowmem_control *);
};

struct lowmem_worker {
	pthread_t	worker_thread;
	GAsyncQueue	*queue;
	int		active;
	int		running;
};

static struct lowmem_worker lmw;

#define LOWMEM_WORKER_IS_ACTIVE(_lmw)	g_atomic_int_get(&(_lmw)->active)
#define LOWMEM_WORKER_ACTIVATE(_lmw)	g_atomic_int_set(&(_lmw)->active, 1)
#define LOWMEM_WORKER_DEACTIVATE(_lmw)	g_atomic_int_set(&(_lmw)->active, 0)

#define LOWMEM_WORKER_IS_RUNNING(_lmw)	g_atomic_int_get(&(_lmw)->running)
#define LOWMEM_WORKER_RUN(_lmw)	g_atomic_int_set(&(_lmw)->running, 1)
#define LOWMEM_WORKER_IDLE(_lmw)	g_atomic_int_set(&(_lmw)->running, 0)

#define LOWMEM_NEW_REQUEST() g_slice_new0(struct lowmem_control)

#define LOWMEM_DESTROY_REQUEST(_ctl)		\
	g_slice_free(typeof(*(_ctl)), _ctl);	\

#define LOWMEM_SET_REQUEST(c, __flags, __type, __size, __count, __cb)	\
{									\
	(c)->flags	= __flags; (c)->type	= __type;		\
	(c)->size	= __size;  (c)->count	= __count;		\
	(c)->callback	= __cb;						\
}

static void lowmem_queue_request(struct lowmem_worker *lmw,
				struct lowmem_control *ctl)
{
	if (LOWMEM_WORKER_IS_ACTIVE(lmw))
		g_async_queue_push(lmw->queue, ctl);
}

/* internal */
static void lowmem_drain_queue(struct lowmem_worker *lmw)
{
	struct lowmem_control *ctl;

	g_async_queue_lock(lmw->queue);
	while ((ctl = g_async_queue_try_pop_unlocked(lmw->queue))) {
		if (ctl->callback)
			ctl->callback(ctl);
		LOWMEM_DESTROY_REQUEST(ctl);
	}
	g_async_queue_unlock(lmw->queue);
}

static void lowmem_request_destroy(gpointer data)
{
	struct lowmem_control *ctl = (struct lowmem_control*) data;

	if (ctl->callback)
		ctl->callback(ctl);
	LOWMEM_DESTROY_REQUEST(ctl);
}

/*-------------------------------------------------*/

/* low memory action function for cgroup */
static void memory_cgroup_medium_act(enum lmk_type type, struct memcg_info *mi);
/* low memory action function */
static void normal_act(void);
static void swap_act(void);
static void low_act(void);
static void medium_act(void);


static size_t cur_mem_state = LOWMEM_NORMAL;
static int num_max_victims = MAX_MEMORY_CGROUP_VICTIMS;
static int num_vict_between_check = MAX_VICTIMS_BETWEEN_CHECK;

static unsigned long long totalram;
static unsigned long ktotalram;

static struct module_ops memory_modules_ops;
static const struct module_ops *lowmem_ops;
static bool oom_popup_enable;
static bool oom_popup;
static bool memcg_swap_status;
static bool bg_reclaim;

static struct memcg_info gmi[MEMCG_MAX] = {
	{LOWMEM_ROOT_CGROUP,      "/",              MEMCG_ROOT,},
	{LOWMEM_APPS_CGROUP,      "Apps",           MEMCG_MEMORY,},
	{LOWMEM_BGLOCKED_CGROUP,  "Apps/BgLocked",  MEMCG_APPS,},
	{LOWMEM_MEMLIMIT_CGROUP,  "MemLimit",       MEMCG_MEMORY,},
	{LOWMEM_SWAP_CGROUP,      "Swap",           MEMCG_MEMORY,},
};

enum memory_level {
	MEMORY_LEVEL_NORMAL,
	MEMORY_LEVEL_LOW,
	MEMORY_LEVEL_CRITICAL,
};

/*
 * This structure has full hierarchy of memory cgroups on running system.
 * It is exported through lowmem-handler.h file.
 **/
static struct memcg **memcg_tree;

/*
 * Special node that point's to /sys/fs/cgroup/memory - root of memcg group.
 * This is the same as memcg_tree[MEMCG_MEMORY]->info.
 */
static struct memcg_info *memcg_root;

static GPtrArray *vip_apps;

static const char *convert_memstate_to_str(int mem_state)
{
	static const char *state_table[] = {"mem normal", "mem swap", "mem low",
			"mem medium"};
	if (mem_state >= 0 && mem_state < LOWMEM_MAX_LEVEL)
		return state_table[mem_state];
	return "";
}

static int lowmem_launch_oompopup(void)
{
	GVariantBuilder *const gv_builder = g_variant_builder_new(G_VARIANT_TYPE("a{ss}"));
	g_variant_builder_add(gv_builder, "{ss}", "_SYSPOPUP_CONTENT_", "lowmemory_oom");

	GVariant *const params = g_variant_new("(a{ss})", gv_builder);
	g_variant_builder_unref(gv_builder);

	int ret = d_bus_call_method_sync_gvariant(SYSTEM_POPUP_BUS_NAME,
	    SYSTEM_POPUP_PATH_SYSTEM, SYSTEM_POPUP_IFACE_SYSTEM,
	    "PopupLaunch", params);

	g_variant_unref(params);

	return ret;
}

static inline void get_total_memory(void)
{
	struct sysinfo si;
	unsigned long long total_memory;

	if (totalram)
		return;

	if (!sysinfo(&si)) {
		total_memory = (unsigned long long)si.totalram * si.mem_unit;
		totalram = total_memory;

		/* ktotalram can cover up to 4TB */
		total_memory = BYTE_TO_KBYTE(totalram);
		ktotalram = (total_memory >= ULONG_MAX) ? ULONG_MAX : total_memory;
	}
}

static int lowmem_mem_usage_uss(pid_t pid, unsigned int *usage)
{
	unsigned int uss, zram = 0;
	int ret;

	*usage = 0;

	/*
	 * In lowmem we need to know memory size of processes to
	 * for terminating apps. To get most real value of usage
	 * we should use USS + ZRAM usage for selected process.
	 *
	 * Those values will contain the most approximated amount
	 * of memory that will be freed after process termination.
	 */
	ret = proc_get_uss(pid, &uss);
	if (ret != RESOURCED_ERROR_NONE)
		return ret;

	if (swap_get_state() == SWAP_ON) {
		ret = proc_get_zram_usage(pid, &zram);
		/* If we don't get zram usage, it's not a problem */
		if (ret != RESOURCED_ERROR_NONE)
			zram = 0;
	}
	*usage = uss + zram;
	return RESOURCED_ERROR_NONE;
}

unsigned int lowmem_get_task_mem_usage_rss(const struct task_info *tsk)
{
	unsigned int size = 0, total_size = 0;
	int index, ret;
	pid_t pid;

	/*
	 * If pids are allocated only when there are multiple processes with
	 * the same pgid e.g., browser and web process. Mostly, single process
	 * is used.
	 */
	if (tsk->pids == NULL) {
		ret = proc_get_ram_usage(tsk->pid, &size);

		/* If there is no proc entry for given pid the process
		 * should be abandoned during further processing
		 */
		if (ret < 0)
			_D("failed to get rss memory usage of %d", tsk->pid);

		return size;
	}

	for (index = 0; index < tsk->pids->len; index++) {
		pid = g_array_index(tsk->pids, pid_t, index);
		ret = proc_get_ram_usage(pid, &size);
		if (ret != RESOURCED_ERROR_NONE)
			continue;
		total_size += size;
	}

	return total_size;
}

static int memps_file_select(const struct dirent *entry)
{
	return strstr(entry->d_name, "memps") ? 1 : 0;
}

int clear_logs(void *data)
{
	struct dirent **namelist;
	int n, i, ret;
	char fpath[BUF_MAX];
	char *fname;
	char *dir = (char*)data;
	int len;

	if (!dir) {
		_E("Invalid parameter - dir is NULL");
		return RESOURCED_ERROR_INVALID_PARAMETER;
	}

	len = strlen(dir);
	if (len <= 0 || len >= sizeof fpath - 1) {
		_E("Invalid parameter - Directory path is too short or too long");
		return RESOURCED_ERROR_INVALID_PARAMETER;
	}

	n = scandir(dir, &namelist, memps_file_select, alphasort);
	_D("num of log files %d", n);
	if (n < MAX_MEMPS_LOGS) {
		while (n--)
			free(namelist[n]);
		free(namelist);
		return RESOURCED_ERROR_NONE;
	}

	strncpy(fpath, dir, sizeof fpath);
	fpath[sizeof fpath - 1] = '\0';
	fname = fpath + len;
	*fname++ = '/';

	len = sizeof fpath - len - 1;
	for (i = 0; i < n; i++) {
		if (i < NUM_RM_LOGS) {
			if (strlen(namelist[i]->d_name) > len - 1)
				continue;
			strncpy(fname, namelist[i]->d_name, len - 1);
			fpath[sizeof fpath - 1] = '\0';
			_D("remove log file %s", fpath);
			ret = remove(fpath);
			if (ret < 0)
				_E("%s file cannot removed", fpath);
		}

		free(namelist[i]);
	}
	free(namelist);
	return RESOURCED_ERROR_NONE;
}

void make_memps_log(char *file, pid_t pid, char *victim_name)
{
	_cleanup_fclose_ FILE *f = NULL;
	time_t now;
	struct tm cur_tm;
	char new_log[BUF_MAX];
	static pid_t old_pid;

	if (old_pid == pid)
		return;
	old_pid = pid;

	now = time(NULL);

	if (localtime_r(&now, &cur_tm) == NULL) {
		_E("Fail to get localtime");
		return;
	}

	snprintf(new_log, sizeof(new_log),
			"%s_%s_%d_%.4d%.2d%.2d%.2d%.2d%.2d", file, victim_name,
			pid, (1900 + cur_tm.tm_year), 1 + cur_tm.tm_mon,
			cur_tm.tm_mday, cur_tm.tm_hour, cur_tm.tm_min,
			cur_tm.tm_sec);

	f = fopen(new_log, "w");
	if (!f) {
		_E("fail to create memps log %s", new_log);
		return;
	}

	proc_print_meninfo(f);
}

static int lowmem_kill_victim(const struct task_info *tsk,
		int flags, unsigned int *victim_size)
{
	pid_t pid;
	int ret;
	char appname[PATH_MAX];
	int sigterm = 0;
	struct proc_app_info *pai;

	pid = tsk->pid;

	if (pid <= 0 || pid == getpid())
		return RESOURCED_ERROR_FAIL;

	ret = proc_get_cmdline(pid, appname);
	if (ret == RESOURCED_ERROR_FAIL)
		return RESOURCED_ERROR_FAIL;

	if (!strcmp("crash-worker", appname) ||
	    !strcmp("system-syspopup", appname)) {
		_E("%s(%d) was selected, skip it", appname, pid);
		return RESOURCED_ERROR_FAIL;
	}

	pai = find_app_info(pid);

	if (pai) {
		resourced_proc_status_change(PROC_CGROUP_SET_TERMINATE_REQUEST,
			pid, NULL, NULL, PROC_TYPE_NONE);

		if (tsk->oom_score_adj < OOMADJ_BACKGRD_LOCKED) {
			sigterm = 1;
		} else if (tsk->oom_score_adj == OOMADJ_BACKGRD_LOCKED) {
			int app_flag = pai->flags;
			sigterm = app_flag & PROC_SIGTERM;
		}

		if (pai->memory.oom_killed)
			sigterm = 0;

		pai->memory.oom_killed = true;
	}

	if (sigterm)
		kill(pid, SIGTERM);
	else
		kill(pid, SIGKILL);

	_E("we killed, force(%d), %d (%s) score = %d, size: rss = %u, sigterm = %d\n",
	   flags & OOM_FORCE, pid, appname, tsk->oom_score_adj,
	   tsk->size, sigterm);
	*victim_size = tsk->size;

	if (tsk->oom_score_adj > OOMADJ_FOREGRD_UNLOCKED)
		return RESOURCED_ERROR_NONE;

	if (oom_popup_enable && !oom_popup) {
		lowmem_launch_oompopup();
		oom_popup = true;
	}

	return RESOURCED_ERROR_NONE;
}

/* return LOWMEM_RECLAIM_CONT when killing should be continued */
static int lowmem_check_kill_continued(struct task_info *tsk, int flags)
{
	unsigned int available;

	if (tsk->oom_score_adj > OOMADJ_BACKGRD_PERCEPTIBLE)
		return LOWMEM_RECLAIM_CONT;

	if ((flags & (OOM_FORCE|OOM_SINGLE_SHOT)) || !(flags & OOM_REVISE)) {
		_I("%d is dropped during force kill, flag = %d",
			tsk->pid, flags);
		return LOWMEM_RECLAIM_DROP;
	}
	available = proc_get_mem_available();
	if (available > lmk_start_threshold) {
		_I("available: %d MB, larger than %u MB, do not kill foreground",
			available, lmk_start_threshold);
		return LOWMEM_RECLAIM_WAIT;
	}
	return LOWMEM_RECLAIM_CONT;
}

static int compare_victims(const struct task_info *ta, const struct task_info *tb)
{
	/*
	 * Firstly, sort by oom_score_adj
	 * Secondly, sort by task size
	 */
	assert(ta != NULL);
	assert(tb != NULL);

	if (ta->oom_score_adj != tb->oom_score_adj)
		return tb->oom_score_adj - ta->oom_score_adj;

	/*
	 * Get memory usage for tasks with the same oom score.
	 * Since there could be multiple tasks with the same pgid,
	 * we sum all of the tasks.
	 * Even though we calculate memory usage when we should compare it,
	 * there could be frequent cases to get memory usage of tasks that
	 * would not be selected as victim. Therefore we just use RSS + Swap,
	 * which causes much less overhead.
	 */
	return (int)(tb->size) - (int)(ta->size);
}

static void lowmem_free_task_info_array(GArray *array)
{
	int i;

	for (i = 0; i < array->len; i++) {
		struct task_info *tsk;

		tsk = &g_array_index(array, struct task_info, i);
		if (tsk->pids)
			g_array_free(tsk->pids, true);
	}

	g_array_free(array, true);
}

static unsigned int is_memory_recovered(unsigned int *avail, unsigned int *thres)
{
	unsigned int available = proc_get_mem_available();
	unsigned int leave_threshold = memcg_root->threshold_leave;
	unsigned int should_be_freed = 0;

	if (available < leave_threshold)
		should_be_freed = leave_threshold - available;
	/*
	 * free THRESHOLD_MARGIN more than real should be freed,
	 * because launching app is consuming up the memory.
	 */
	if (should_be_freed > 0)
		should_be_freed += THRESHOLD_MARGIN;

	*avail = available;
	*thres = leave_threshold;

	_I("should_be_freed = %u MB", should_be_freed);
	return should_be_freed;
}

/**
 * @brief Terminate up to max_victims processes after finding them from pai.
	It depends on proc_app_info lists
	and it also reference systemservice cgroup
	because some processes in this group don't have proc_app_info.
 *
 * @max_victims:	    max number of processes to be terminated
 * @start_oom:	    find victims from start oom adj score value
 * @end_oom: find victims to end oom adj score value
 * @should_be_freed: amount of memory to be reclaimed (in MB)
 * @total_size[out]: total size of possibly reclaimed memory (required)
 * @completed:	    final outcome (optional)
 * @threshold:		desired value of memory available
 */
static int lowmem_kill_victims(int max_victims,
	int start_oom, int end_oom, unsigned should_be_freed, int flags,
	unsigned int *total_size, int *completed, int threshold)
{
	GSList *proc_app_list = NULL;
	int i, ret, victim = 0;
	unsigned int victim_size = 0;
	unsigned int total_victim_size = 0;
	unsigned int total_task_size = 0;
	int status = LOWMEM_RECLAIM_NONE;
	GArray *candidates = NULL;
	GSList *iter, *iterchild;
	struct proc_app_info *pai = NULL;
	int oom_score_adj;
	int should_be_freed_kb = MBYTE_TO_KBYTE(should_be_freed);

	candidates = g_array_new(false, false, sizeof(struct task_info));

	proc_app_list = proc_app_list_open();
	gslist_for_each_item(iter, proc_app_list) {
		struct task_info ti;

		pai = (struct proc_app_info *)iter->data;
		if (!pai->main_pid)
			continue;

		oom_score_adj = pai->memory.oom_score_adj;
		if (oom_score_adj > end_oom || oom_score_adj < start_oom)
			continue;

		if ((flags & OOM_REVISE) && pai->memory.oom_killed)
			continue;

		ti.pid = pai->main_pid;
		ti.pgid = getpgid(ti.pid);
		ti.oom_score_adj = oom_score_adj;

		if (pai->childs) {
			ti.pids = g_array_new(false, false, sizeof(pid_t));
			g_array_append_val(ti.pids, ti.pid);
			gslist_for_each_item(iterchild, pai->childs) {
				pid_t child = GPOINTER_TO_PID(iterchild->data);
				g_array_append_val(ti.pids, child);
			}
		} else
			ti.pids = NULL;

		ti.size = lowmem_get_task_mem_usage_rss(&ti);
		g_array_append_val(candidates, ti);
		total_task_size += ti.size;
	}

	proc_app_list_close();

	if (!candidates->len)
		goto leave;

	g_array_sort(candidates, (GCompareFunc)compare_victims);

	for (i = 0; i < candidates->len; i++) {
		struct task_info *tsk;

		if (i >= max_victims) {
			status = LOWMEM_RECLAIM_WAIT;
			break;
		}

		/*
		 * Available memory is checking only every
		 * num_vict_between_check process for reducing burden.
		 */
		if (!(i % num_vict_between_check)) {
			if (proc_get_mem_available() > threshold) {
				status = LOWMEM_RECLAIM_DONE;
				break;
			}
		}

		if (!(flags & OOM_NOMEMORY_CHECK) &&
		    total_victim_size >= should_be_freed_kb) {
			_E("victim = %d, max_victims = %d, total_size = %u",
				victim, max_victims, total_victim_size);
			status = LOWMEM_RECLAIM_DONE;
			break;
		}

		tsk = &g_array_index(candidates, struct task_info, i);

		status = lowmem_check_kill_continued(tsk, flags);
		if (status != LOWMEM_RECLAIM_CONT)
			break;

		_I("select victims from proc_app_list pid(%d)\n", tsk->pid);

		ret = lowmem_kill_victim(tsk, flags, &victim_size);
		if (ret != RESOURCED_ERROR_NONE)
			continue;
		victim++;
		total_victim_size += victim_size;
	}

leave:
	lowmem_free_task_info_array(candidates);
	*total_size = total_victim_size;
	*completed = status;
	return victim;
}

static void calualate_range_of_oom(enum lmk_type lmk, int *min, int *max)
{
	if (lmk == LMK_OLDEST) {
		*max = OOMADJ_APP_MAX;
		*min = OOMADJ_BACKGRD_OLD;
	} else if (lmk == LMK_RECENTLY_USE) {
		*max = OOMADJ_RECENTLY_USED;
		*min = OOMADJ_FAVORITE;
	} else if (lmk == LMK_ACTIVE) {
		*max = OOMADJ_BACKGRD_LOCKED;
		*min = OOMADJ_PREVIOUS_DEFAULT;
	} else if (lmk == LMK_FOREGROUND) {
		*max = OOMADJ_BACKGRD_PERCEPTIBLE;
		*min = OOMADJ_SU;
	} else { /* lmk == LMK_MEMORY */
		*max = OOMADJ_APP_MAX;
		*min = OOMADJ_SU;
	}
}

static void lowmem_handle_request(struct lowmem_control *ctl)
{
	int start_oom, end_oom;
	int count = 0, victim_cnt = 0;
	int status = LOWMEM_RECLAIM_NONE;
	unsigned int available = 0, leave_threshold = 0;
	unsigned int total_size = 0;
	unsigned int current_size = 0;
	unsigned int reclaim_size, shortfall = 0;
	enum lmk_type lmk_type = ctl->type;

	available = proc_get_mem_available();
	reclaim_size = ctl->size  > available
		     ? ctl->size - available : 0;

	if (!reclaim_size) {
		status = LOWMEM_RECLAIM_DONE;
		goto done;
	}

retry:
	/* Prepare LMK to start doing it's job. Check preconditions. */
	calualate_range_of_oom(lmk_type, &start_oom, &end_oom);
	lmk_start_threshold = memcg_root->threshold[LOWMEM_MEDIUM];
	shortfall = is_memory_recovered(&available, &leave_threshold);

	if (!shortfall || !reclaim_size) {
		status = LOWMEM_RECLAIM_DONE;
		goto done;
	}
	_D("reclaim start, available = %u, currently required = %u, possibly reclaimed: %u, type: %d",
		available, reclaim_size, total_size, lmk_type);

	/* precaution */
	current_size = 0;
	victim_cnt = lowmem_kill_victims(ctl->count - count, start_oom, end_oom,
			    reclaim_size, ctl->flags, &current_size, &status, ctl->size);

	current_size = KBYTE_TO_MBYTE(current_size);
	count += victim_cnt;
	total_size += current_size;

	reclaim_size -= reclaim_size > current_size
			    ? current_size : reclaim_size;
	_E("kill %d victims, total_size = %u from %d to %d %d",
		victim_cnt, current_size, start_oom, end_oom, status);

	if ((status == LOWMEM_RECLAIM_DONE) ||
	    (status == LOWMEM_RECLAIM_WAIT) ||
	    (status == LOWMEM_RECLAIM_DROP))
		goto done;

	if (victim_cnt)
		_E("So far: killed %d processes reclaimed: %u remaining: %u shortfall: %u\n",
			count, total_size, reclaim_size, shortfall);

	/*
	 * If it doesn't finish reclaiming memory in first operation,
		- if flags has OOM_IN_DEPTH,
		   try to find victims again in the active cgroup.
		   otherwise, just return because there is no more victims in the desired cgroup.
		- if flags has OOM_REVISE,
		   it means that resourced can't find victims from proc_app_list.
		   So, it should search victims or malicious process from /proc.
		   But searching /proc leads to abnormal behaviour.
		   (Make sluggish or kill same victims continuously)
		   Thus, otherwise, just return in first operation and wait some period.
	 */
	if ((count < ctl->count) && (ctl->flags & OOM_IN_DEPTH)) {
		if (lmk_type == LMK_OLDEST) {
			/* We tried in inactive processes and didn't succeed, try immediately in active */
			_D("Inactive wasn't enough, lets try in Active.");
			lmk_type = LMK_ACTIVE;
			goto retry;
		} else if (lmk_type == LMK_ACTIVE) {
			/*
			 * We tried in INACTIVE and ACTIVE but still didn't succeed
			 * so it's time to try in /proc. Before doing it wait some time.
			 */
			ctl->status = LOWMEM_RECLAIM_WAIT;
			ctl->type = LMK_MEMORY;
			_D("We would need to kill from /proc wait some time and try again");
			return;
		}
	}
done:
	_E("Done: killed %d processes reclaimed: %u remaining: %u shortfall: %u status: %d\n",
		count, total_size, reclaim_size, shortfall, status);

	/* After we finish reclaiming it's worth to remove oldest memps logs */
	if (status == LOWMEM_RECLAIM_DONE)
		request_helper_worker(CLEAR_LOGS, MEMPS_LOG_PATH, clear_logs, NULL);
	ctl->status = status;
}

static void *lowmem_reclaim_worker(void *arg)
{
	struct lowmem_worker *lmw = (struct lowmem_worker *)arg;

	setpriority(PRIO_PROCESS, 0, OOM_KILLER_PRIORITY);

	g_async_queue_ref(lmw->queue);

	while (1) {
		struct lowmem_control *ctl;

		LOWMEM_WORKER_IDLE(lmw);
		/* Wait on any wake-up call */
		ctl = g_async_queue_pop(lmw->queue);

		if (!LOWMEM_WORKER_IS_ACTIVE(lmw) || (ctl->flags & OOM_DROP))
			break;

		LOWMEM_WORKER_RUN(lmw);
process_again:
		lowmem_handle_request(ctl);
		/**
		 * Case the process failed to reclaim requested amount of memory
		 * or still under have memory pressure - try the timeout wait.
		 * There is a chance this will get woken-up in a better reality.
		 */
		if (ctl->status != LOWMEM_RECLAIM_DONE &&
		    !(ctl->flags & OOM_SINGLE_SHOT)) {
			unsigned int available = proc_get_mem_available();

			if (available >= ctl->size) {
				_E("Memory restored: requested: %u available: %u\n",
					ctl->size, available);
				ctl->status = LOWMEM_RECLAIM_DONE;
				if (ctl->callback)
					ctl->callback(ctl);
				LOWMEM_DESTROY_REQUEST(ctl);
				LOWMEM_WORKER_IDLE(lmw);
				continue;
			}

			if (LOWMEM_WORKER_IS_ACTIVE(lmw)) {
				g_usleep(LMW_RETRY_WAIT_TIMEOUT_MSEC);
				ctl->flags |= OOM_REVISE;
				goto process_again;
			}
		}

		/*
		 * The ctl callback would check available size again.
		 * And it is last point in reclaiming worker.
		 * Resourced sent SIGKILL signal to victim processes
		 * so it should wait for a some seconds until each processes returns memory.
		 */
		g_usleep(LMW_LOOP_WAIT_TIMEOUT_MSEC);
		if (ctl->callback)
			ctl->callback(ctl);

		/* The lmk becomes the owner of all queued requests .. */
		LOWMEM_DESTROY_REQUEST(ctl);
		LOWMEM_WORKER_IDLE(lmw);
	}
	g_async_queue_unref(lmw->queue);
	pthread_exit(NULL);
}

static void change_lowmem_state(unsigned int mem_state)
{
	_I("[LOW MEM STATE] %s ==> %s", convert_memstate_to_str(cur_mem_state),
		convert_memstate_to_str(mem_state));
	cur_mem_state = mem_state;
	lmk_start_threshold = memcg_root->threshold[LOWMEM_MEDIUM];

	resourced_notify(RESOURCED_NOTIFIER_MEM_STATE_CHANGED,
		(void *)&cur_mem_state);
}

static void lowmem_swap_memory(enum memcg_type type, struct memcg_info *mi)
{
	unsigned int available;
	struct swap_status_msg msg;

	if (cur_mem_state == LOWMEM_NORMAL)
		return;

	if (swap_get_state() != SWAP_ON)
		return;

	available = proc_get_mem_available();
	if (cur_mem_state != LOWMEM_SWAP &&
	    available <= memcg_root->threshold[LOWMEM_SWAP])
		swap_act();

	msg.type = type;
	msg.info = mi;
	resourced_notify(RESOURCED_NOTIFIER_SWAP_START, &msg);
	memcg_swap_status = true;
}

void lowmem_trigger_swap(pid_t pid, int memcg_idx)
{
	struct memcg_info *mi;
	struct swap_status_msg msg;

	mi = memcg_tree[memcg_idx]->info;
	_D("name : %s, pid : %d", mi->name, pid);
	cgroup_write_node_uint32(mi->name, CGROUP_FILE_NAME, pid);
	msg.type = memcg_idx;
	msg.info = mi;
	resourced_notify(RESOURCED_NOTIFIER_SWAP_START, &msg);
}

static void memory_level_send_system_event(int lv)
{
	bundle *b;
	const char *str;

	switch (lv) {
	case MEMORY_LEVEL_NORMAL:
		str = EVT_VAL_MEMORY_NORMAL;
		break;
	case MEMORY_LEVEL_LOW:
		str = EVT_VAL_MEMORY_SOFT_WARNING;
		break;
	case MEMORY_LEVEL_CRITICAL:
		str = EVT_VAL_MEMORY_HARD_WARNING;
		break;
	default:
		_E("Invalid state");
		return;
	}

	b = bundle_create();
	if (!b) {
		_E("Failed to create bundle");
		return;
	}

	bundle_add_str(b, EVT_KEY_LOW_MEMORY, str);
	eventsystem_send_system_event(SYS_EVENT_LOW_MEMORY, b);
	bundle_free(b);
}

static void normal_act(void)
{
	int ret, status;
	struct swap_status_msg msg;

	ret = vconf_get_int(VCONFKEY_SYSMAN_LOW_MEMORY, &status);
	if (ret)
		_D("vconf_get_int fail %s", VCONFKEY_SYSMAN_LOW_MEMORY);
	if (status != VCONFKEY_SYSMAN_LOW_MEMORY_NORMAL) {
		vconf_set_int(VCONFKEY_SYSMAN_LOW_MEMORY,
			      VCONFKEY_SYSMAN_LOW_MEMORY_NORMAL);
		memory_level_send_system_event(MEMORY_LEVEL_NORMAL);
	}

	change_lowmem_state(LOWMEM_NORMAL);

	if (swap_get_state() == SWAP_ON && memcg_swap_status) {
		msg.type = MEMCG_SWAP;
		msg.info = memcg_tree[msg.type]->info;
		resourced_notify(RESOURCED_NOTIFIER_SWAP_UNSET_LIMIT, &msg);
		memcg_swap_status = false;
	}
	if (proc_get_freezer_status() == CGROUP_FREEZER_PAUSED)
		resourced_notify(RESOURCED_NOTIFIER_FREEZER_CGROUP_STATE,
			(void *)CGROUP_FREEZER_ENABLED);
}

static void swap_act(void)
{
	int ret, status;

	ret = vconf_get_int(VCONFKEY_SYSMAN_LOW_MEMORY, &status);
	if (ret)
		_E("vconf get failed %s", VCONFKEY_SYSMAN_LOW_MEMORY);

	if (status != VCONFKEY_SYSMAN_LOW_MEMORY_NORMAL) {
		vconf_set_int(VCONFKEY_SYSMAN_LOW_MEMORY,
				VCONFKEY_SYSMAN_LOW_MEMORY_NORMAL);
		memory_level_send_system_event(MEMORY_LEVEL_NORMAL);
	}
	change_lowmem_state(LOWMEM_SWAP);
	if (proc_get_freezer_status() == CGROUP_FREEZER_PAUSED)
		resourced_notify(RESOURCED_NOTIFIER_FREEZER_CGROUP_STATE,
			(void *)CGROUP_FREEZER_ENABLED);

	if (swap_get_state() != SWAP_ON)
		resourced_notify(RESOURCED_NOTIFIER_SWAP_ACTIVATE, NULL);
}

static void low_act(void)
{
	change_lowmem_state(LOWMEM_LOW);
	resourced_notify(RESOURCED_NOTIFIER_SWAP_COMPACT, (void *)SWAP_COMPACT_LOWMEM_LOW);
	memory_level_send_system_event(MEMORY_LEVEL_LOW);
}

static void medium_cb(struct lowmem_control *ctl)
{
	if (ctl->status == LOWMEM_RECLAIM_DONE)
		oom_popup = false;
	lowmem_change_memory_state(LOWMEM_NORMAL, 0);
}

static void medium_act(void)
{
	unsigned int available;
	int ret;
	int status = VCONFKEY_SYSMAN_LOW_MEMORY_NORMAL;

	/*
	 * Don't trigger reclaim worker
	 * if it is already running
	 */
	if (LOWMEM_WORKER_IS_RUNNING(&lmw))
		return;

	ret = vconf_get_int(VCONFKEY_SYSMAN_LOW_MEMORY, &status);
	if (ret)
		_D("vconf_get_int fail %s", VCONFKEY_SYSMAN_LOW_MEMORY);

	memory_level_send_system_event(MEMORY_LEVEL_CRITICAL);
	if (status != VCONFKEY_SYSMAN_LOW_MEMORY_HARD_WARNING) {
		if (proc_get_freezer_status() == CGROUP_FREEZER_ENABLED)
			resourced_notify(RESOURCED_NOTIFIER_FREEZER_CGROUP_STATE,
				(void *)CGROUP_FREEZER_PAUSED);
		vconf_set_int(VCONFKEY_SYSMAN_LOW_MEMORY,
			      VCONFKEY_SYSMAN_LOW_MEMORY_HARD_WARNING);
	}
	available = proc_get_mem_available();

	change_lowmem_state(LOWMEM_MEDIUM);

	if (available < memcg_root->threshold_leave) {
		struct lowmem_control *ctl = LOWMEM_NEW_REQUEST();

		if (ctl) {
			LOWMEM_SET_REQUEST(ctl, OOM_IN_DEPTH,
				LMK_OLDEST, memcg_root->threshold_leave,
				num_max_victims, medium_cb);
			lowmem_queue_request(&lmw, ctl);
		}
	}

	resourced_notify(RESOURCED_NOTIFIER_SWAP_COMPACT, (void *)SWAP_COMPACT_LOWMEM_MEDIUM);

	/*
	 * Flush resourced memory such as other processes.
	 * Resourced can use both many fast bins and sqlite3 cache memery.
	 */
	malloc_trim(0);

	return;
}

static void lowmem_trigger_memory_state_action(int mem_state)
{
	/*
	 * Check if the state we want to set is different from current
	 * But it should except this condition if mem_state is already medium.
	 * Otherwise, recalim worker couldn't run any more.
	 */
	if (mem_state != LOWMEM_MEDIUM && cur_mem_state == mem_state)
		return;

	switch (mem_state) {
	case LOWMEM_NORMAL:
		normal_act();
		break;
	case LOWMEM_SWAP:
		swap_act();
		break;
	case LOWMEM_LOW:
		low_act();
		break;
	case LOWMEM_MEDIUM:
		medium_act();
		break;
	default:
		assert(0);
	}
}

static void lowmem_dump_cgroup_procs(struct memcg_info *mi)
{
	int i;
	unsigned int size;
	pid_t pid;
	GArray *pids_array = NULL;

	cgroup_get_pids(mi->name, &pids_array);

	for (i = 0; i < pids_array->len; i++) {
		pid = g_array_index(pids_array, pid_t, i);
		lowmem_mem_usage_uss(pid, &size);
		_I("pid = %d, size = %u KB", pid, size);
	}
	g_array_free(pids_array, true);
}

static void memory_cgroup_medium_act(enum lmk_type type, struct memcg_info *mi)
{
	struct lowmem_control *ctl;

	_I("[LOW MEM STATE] memory cgroup %s oom state",
		mi->name);

	/* To Do: only start to kill fg victim when no pending fg victim */
	lowmem_dump_cgroup_procs(mi);

	ctl = LOWMEM_NEW_REQUEST();
	if (ctl) {
		LOWMEM_SET_REQUEST(ctl, OOM_SINGLE_SHOT | OOM_IN_DEPTH, type,
			mi->oomleave, num_max_victims, NULL);
		lowmem_queue_request(&lmw, ctl);
	}
}

static unsigned int check_mem_state(unsigned int available)
{
	int mem_state;
	for (mem_state = LOWMEM_MAX_LEVEL - 1; mem_state > LOWMEM_NORMAL; mem_state--) {
		if (mem_state != LOWMEM_MEDIUM && available <= memcg_root->threshold[mem_state])
			break;
		else if (mem_state == LOWMEM_MEDIUM && available <= lmk_start_threshold)
			break;
	}

	return mem_state;
}

static int load_vip_config(struct parse_result *result, void *user_data)
{
	char *app_name;

	if (!result || !vip_apps)
		return RESOURCED_ERROR_INVALID_PARAMETER;

	if (strncmp(result->section, MEM_VIP_SECTION, sizeof(MEM_VIP_SECTION)))
		return RESOURCED_ERROR_NONE;

	if (!strcmp(result->name, MEM_VIP_PREDEFINE)) {
		app_name = g_strdup(result->value);
		g_ptr_array_add(vip_apps, (gpointer)app_name);
	}

	return RESOURCED_ERROR_NONE;
}

static int load_bg_reclaim_config(struct parse_result *result, void *user_data)
{
	if (!result)
		return RESOURCED_ERROR_INVALID_PARAMETER;

	if (strncmp(result->section, MEM_BG_RECLAIM_SECTION, strlen(MEM_BG_RECLAIM_SECTION)+1))
		return RESOURCED_ERROR_NONE;

	if (!strncmp(result->name, MEM_BG_RECLAIM_STRING, strlen(MEM_BG_RECLAIM_STRING)+1)) {
		if (!strncmp(result->value, "yes", strlen("yes")+1))
			bg_reclaim = true;
		else if (!strncmp(result->value, "no", strlen("no")+1))
			bg_reclaim = false;
	}

	_I("Background reclaim enabled = %d", bg_reclaim);

	return RESOURCED_ERROR_NONE;
}

static int load_mem_config(struct parse_result *result, void *user_data)
{
	if (!result)
		return RESOURCED_ERROR_INVALID_PARAMETER;

	if (strncmp(result->section, MEM_POPUP_SECTION, strlen(MEM_POPUP_SECTION)+1))
		return RESOURCED_ERROR_NONE;

	if (!strncmp(result->name, MEM_POPUP_STRING, strlen(MEM_POPUP_STRING)+1)) {
		if (!strncmp(result->value, "yes", strlen("yes")+1))
			oom_popup_enable = 1;
		else if (!strncmp(result->value, "no", strlen("no")+1))
			oom_popup_enable = 0;
	}

	_I("oom_popup_enable = %d", oom_popup_enable);

	return RESOURCED_ERROR_NONE;
}

static int set_memory_config(const char *section_name, const struct parse_result *result)
{
	if (!result || !section_name)
		return -EINVAL;

	if (strncmp(result->section, section_name, strlen(section_name)+1))
		return RESOURCED_ERROR_NONE;

	if (!strncmp(result->name, "ThresholdSwap", strlen("ThresholdSwap")+1)) {
		int value = atoi(result->value);
		lowmem_memcg_set_threshold(MEMCG_MEMORY, LOWMEM_SWAP, value);
	} else if (!strncmp(result->name, "ThresholdLow", strlen("ThresholdLow")+1)) {
		int value = atoi(result->value);
		lowmem_memcg_set_threshold(MEMCG_MEMORY, LOWMEM_LOW, value);
	} else if (!strncmp(result->name, "ThresholdMedium", strlen("ThresholdMedium")+1)) {
		int value = atoi(result->value);
		lowmem_memcg_set_threshold(MEMCG_MEMORY, LOWMEM_MEDIUM, value);
	} else if (!strncmp(result->name, "ThresholdLeave", strlen("ThresholdLeave")+1)) {
		int value = atoi(result->value);
		lowmem_memcg_set_leave_threshold(MEMCG_MEMORY, value);
	} else if (!strncmp(result->name, "NumMaxVictims", strlen("NumMaxVictims")+1)) {
		int value = atoi(result->value);
		num_max_victims = value;
		/*
		 * calculate number of kill victims between check memory available
		 * depends on number of max kill victims
		 */
		num_vict_between_check = value > MAX_MEMORY_CGROUP_VICTIMS/2
						? 3 : value > MAX_MEMORY_CGROUP_VICTIMS/4
								? 2 : 1;
	} else if (!strncmp(result->name, "ProactiveThreshold", strlen("ProactiveThreshold")+1)) {
		int value = atoi(result->value);
		proactive_threshold = value;
	} else if (!strncmp(result->name, "ProactiveLeave", strlen("ProactiveLeave")+1)) {
		int value = atoi(result->value);
		proactive_leave = value;
	} else if (!strncmp(result->name, "EventLevel", strlen("EventLevel")+1)) {
		if (strncmp(event_level, result->value, strlen(event_level)))
			event_level = strdup(result->value);
		if (!event_level)
			return RESOURCED_ERROR_OUT_OF_MEMORY;
	}

	return RESOURCED_ERROR_NONE;
}

static int memory_load_64_config(struct parse_result *result, void *user_data)
{
	return set_memory_config("Memory64", result);
}

static int memory_load_256_config(struct parse_result *result, void *user_data)
{
	return set_memory_config("Memory256", result);
}

static int memory_load_448_config(struct parse_result *result, void *user_data)
{
	return set_memory_config("Memory448", result);
}

static int memory_load_512_config(struct parse_result *result, void *user_data)
{
	return set_memory_config("Memory512", result);
}

static int memory_load_768_config(struct parse_result *result, void *user_data)
{
	return set_memory_config("Memory768", result);
}

static int memory_load_1024_config(struct parse_result *result, void *user_data)
{
	return set_memory_config("Memory1024", result);
}

static int memory_load_2048_config(struct parse_result *result, void *user_data)
{
	return set_memory_config("Memory2048", result);
}

/* setup memcg parameters depending on total ram size. */
static void setup_memcg_params(void)
{
	int i;
	unsigned long long total_ramsize = BYTE_TO_MBYTE(totalram);
	_D("Total: %llu MB", total_ramsize);
	if (total_ramsize <= MEM_SIZE_64) {
		/* set thresholds for ram size 64M */
		proactive_threshold = PROACTIVE_64_THRES;
		proactive_leave = PROACTIVE_64_LEAVE;
		lowmem_memcg_set_threshold(MEMCG_MEMORY, LOWMEM_SWAP, MEMCG_MEMORY_64_THRES_SWAP);
		lowmem_memcg_set_threshold(MEMCG_MEMORY, LOWMEM_LOW, MEMCG_MEMORY_64_THRES_LOW);
		lowmem_memcg_set_threshold(MEMCG_MEMORY, LOWMEM_MEDIUM, MEMCG_MEMORY_64_THRES_MEDIUM);
		lowmem_memcg_set_leave_threshold(MEMCG_MEMORY, MEMCG_MEMORY_64_THRES_LEAVE);
		config_parse(MEM_CONF_FILE, memory_load_64_config, NULL);
	} else if (total_ramsize <= MEM_SIZE_256) {
		/* set thresholds for ram size 256M */
		proactive_threshold = PROACTIVE_256_THRES;
		proactive_leave = PROACTIVE_256_LEAVE;
		lowmem_memcg_set_threshold(MEMCG_MEMORY, LOWMEM_SWAP, MEMCG_MEMORY_256_THRES_SWAP);
		lowmem_memcg_set_threshold(MEMCG_MEMORY, LOWMEM_LOW, MEMCG_MEMORY_256_THRES_LOW);
		lowmem_memcg_set_threshold(MEMCG_MEMORY, LOWMEM_MEDIUM, MEMCG_MEMORY_256_THRES_MEDIUM);
		lowmem_memcg_set_leave_threshold(MEMCG_MEMORY, MEMCG_MEMORY_256_THRES_LEAVE);
		config_parse(MEM_CONF_FILE, memory_load_256_config, NULL);
	} else if (total_ramsize <= MEM_SIZE_448) {
		/* set thresholds for ram size 448M */
		proactive_threshold = PROACTIVE_448_THRES;
		proactive_leave = PROACTIVE_448_LEAVE;
		lowmem_memcg_set_threshold(MEMCG_MEMORY, LOWMEM_SWAP, MEMCG_MEMORY_448_THRES_SWAP);
		lowmem_memcg_set_threshold(MEMCG_MEMORY, LOWMEM_LOW, MEMCG_MEMORY_448_THRES_LOW);
		lowmem_memcg_set_threshold(MEMCG_MEMORY, LOWMEM_MEDIUM, MEMCG_MEMORY_448_THRES_MEDIUM);
		lowmem_memcg_set_leave_threshold(MEMCG_MEMORY, MEMCG_MEMORY_448_THRES_LEAVE);
		config_parse(MEM_CONF_FILE, memory_load_448_config, NULL);
	} else if (total_ramsize <= MEM_SIZE_512) {
		/* set thresholds for ram size 512M */
		proactive_threshold = PROACTIVE_512_THRES;
		proactive_leave = PROACTIVE_512_LEAVE;
		lowmem_memcg_set_threshold(MEMCG_MEMORY, LOWMEM_SWAP, MEMCG_MEMORY_512_THRES_SWAP);
		lowmem_memcg_set_threshold(MEMCG_MEMORY, LOWMEM_LOW, MEMCG_MEMORY_512_THRES_LOW);
		lowmem_memcg_set_threshold(MEMCG_MEMORY, LOWMEM_MEDIUM, MEMCG_MEMORY_512_THRES_MEDIUM);
		lowmem_memcg_set_leave_threshold(MEMCG_MEMORY, MEMCG_MEMORY_512_THRES_LEAVE);
		config_parse(MEM_CONF_FILE, memory_load_512_config, NULL);
	}  else if (total_ramsize <= MEM_SIZE_768) {
		/* set thresholds for ram size 512M */
		proactive_threshold = PROACTIVE_768_THRES;
		proactive_leave = PROACTIVE_768_LEAVE;
		lowmem_memcg_set_threshold(MEMCG_MEMORY, LOWMEM_SWAP, MEMCG_MEMORY_768_THRES_SWAP);
		lowmem_memcg_set_threshold(MEMCG_MEMORY, LOWMEM_LOW, MEMCG_MEMORY_768_THRES_LOW);
		lowmem_memcg_set_threshold(MEMCG_MEMORY, LOWMEM_MEDIUM, MEMCG_MEMORY_768_THRES_MEDIUM);
		lowmem_memcg_set_leave_threshold(MEMCG_MEMORY, MEMCG_MEMORY_768_THRES_LEAVE);
		config_parse(MEM_CONF_FILE, memory_load_768_config, NULL);
	} else if (total_ramsize <= MEM_SIZE_1024) {
		/* set thresholds for ram size more than 1G */
		proactive_threshold = PROACTIVE_1024_THRES;
		proactive_leave = PROACTIVE_1024_LEAVE;
		lowmem_memcg_set_threshold(MEMCG_MEMORY, LOWMEM_SWAP, MEMCG_MEMORY_1024_THRES_SWAP);
		lowmem_memcg_set_threshold(MEMCG_MEMORY, LOWMEM_LOW, MEMCG_MEMORY_1024_THRES_LOW);
		lowmem_memcg_set_threshold(MEMCG_MEMORY, LOWMEM_MEDIUM, MEMCG_MEMORY_1024_THRES_MEDIUM);
		lowmem_memcg_set_leave_threshold(MEMCG_MEMORY, MEMCG_MEMORY_1024_THRES_LEAVE);
		config_parse(MEM_CONF_FILE, memory_load_1024_config, NULL);
	} else {
		proactive_threshold = PROACTIVE_2048_THRES;
		proactive_leave = PROACTIVE_2048_LEAVE;
		lowmem_memcg_set_threshold(MEMCG_MEMORY, LOWMEM_SWAP, MEMCG_MEMORY_2048_THRES_SWAP);
		lowmem_memcg_set_threshold(MEMCG_MEMORY, LOWMEM_LOW, MEMCG_MEMORY_2048_THRES_LOW);
		lowmem_memcg_set_threshold(MEMCG_MEMORY, LOWMEM_MEDIUM, MEMCG_MEMORY_2048_THRES_MEDIUM);
		lowmem_memcg_set_leave_threshold(MEMCG_MEMORY, MEMCG_MEMORY_2048_THRES_LEAVE);
		config_parse(MEM_CONF_FILE, memory_load_2048_config, NULL);
	}

	for (i = 0; i < LOWMEM_MAX_LEVEL; i++)
		_I("set threshold for state '%s' to %u MB", convert_memstate_to_str(i), memcg_root->threshold[i]);

	_I("set number of max victims as %d", num_max_victims);
	_I("set threshold leave to %u MB", memcg_root->threshold_leave);
	_I("set proactive threshold to %u MB", proactive_threshold);
	_I("set proactive low memory killer leave to %u MB", proactive_leave);
}

static void init_memcg_params(void)
{
	int idx = 0;
	GSList *cgroups;

	memcg_tree = (struct memcg **)malloc(sizeof(struct memcg *) *
		MEMCG_MAX);
	assert(memcg_tree);

	for (idx = 0; idx < MEMCG_MAX; idx++) {
		struct memcg_info *mi = &gmi[idx];
		memcg_tree[idx] = (struct memcg *)malloc(sizeof(struct memcg));
		assert(memcg_tree[idx]);

		memcg_init(memcg_tree[idx]);
		memcg_tree[idx]->info = mi;
		_I("init memory cgroup for %s", mi->name);

		if (mi->parent_memcg == MEMCG_ROOT) {
			memcg_root = memcg_tree[idx]->info;
		} else {
			cgroups = memcg_tree[mi->parent_memcg]->cgroups;
			cgroups = g_slist_prepend(cgroups, mi);
			memcg_tree[mi->parent_memcg]->use_hierarchy = true;
		}
	}
}

static int write_params_memcg_info(struct memcg_info *mi)
{
	unsigned int limit = mi->limit;
	const char *name = mi->name;
	int ret = RESOURCED_ERROR_NONE;
	_I("write memcg param for %s", name);
	/* enable cgroup move */
	ret = cgroup_write_node_uint32(name,
			MEMCG_MOVE_CHARGE_PATH, 3);
	if (ret)
		return ret;

	if (mi->limit_ratio == LOWMEM_NO_LIMIT)
		return ret;

	/* write limit_in_bytes */
	ret = cgroup_write_node_uint32(name,
			MEMCG_LIMIT_PATH, limit);
	_I("set %s's limit to %u", name, limit);
	return ret;
}

static int write_memcg_params(void)
{
	unsigned int i;

	for (i = 0; i < MEMCG_MAX; i++) {
		struct memcg_info *mi = memcg_tree[i]->info;
		write_params_memcg_info(mi);
	}

	return RESOURCED_ERROR_NONE;
}

static void lowmem_move_memcgroup(int pid, int oom_score_adj)
{
	struct proc_app_info *pai = find_app_info(pid);
	struct memcg_info *mi;
	int ret, memcg_idx, should_swap = 0;

	if (!pai)
		return;

	if (oom_score_adj >= OOMADJ_BACKGRD_PERCEPTIBLE &&
		       oom_score_adj < OOMADJ_BACKGRD_UNLOCKED) {
		/* We'd like lowmem a chance to do a bit more with some
		 * long-lived processes such as background-locked or
		 * favourite apps. This is ideally done with a separate
		 * memcg because:
		 * - Page reclamation in resourced is by design done
		 *   on a per-cgroup basis, not per-process
		 * - The only way to control per-process memory is to
		 *   abort the individual process, which means LMK.
		 *   The favourites and most background-locked processes
		 *   are the final candidates of LMK (ACTIVE or MEMORY)
		 *   and their memory consumption is most of the time
		 *   unhandled by resourced.
		 * - Bg-locked and fav processes are not even candidates
		 *   for swap since swap is only available for bg-unlocked
		 *   processes (the swap module doesn't even care about
		 *   clean pages as it sets move_charge_at_immigrate=1).
		 * - The only remaining practice left for us would be to
		 *   control the Apps memcg, but that is undesirable since
		 *   active processes can share it.
		 * Therefore, we distinguish between active and background
		 * non-unlocked processes to enable optional advanced page
		 * management in userspace. A separate memcg won't affect
		 * the other aspects of lowmem and swap if there is no set
		 * action for it. */
		if (pai->memory.memcg_idx == MEMCG_BGLOCKED)
			return;
		memcg_idx = MEMCG_BGLOCKED;
		mi = memcg_tree[memcg_idx]->info;
	} else if (oom_score_adj > OOMADJ_BACKGRD_UNLOCKED + OOMADJ_APP_INCREASE) {
		if (oom_score_adj != pai->memory.oom_score_adj)
			proc_set_process_memory_state(pai, pai->memory.memcg_idx,
				pai->memory.memcg_info, oom_score_adj);
		return;
	} else if (oom_score_adj >= OOMADJ_INIT) {
		memcg_idx = MEMCG_APPS;
		mi = memcg_tree[memcg_idx]->info;

		if (oom_score_adj >= OOMADJ_FAVORITE)
			should_swap = 1;
	} else
		return;

	/*
	 * Check whether this application has memory limit cgroup or not.
	 * If it is moved to memory limit cgroup,
	 * it is no necessary to control memcg any longer.
	 * Stop other operation and return.
	*/
	if (memcg_idx == MEMCG_APPS) {
		ret = lowmem_limit_move_cgroup(pai);
		if (!ret) {
			memcg_idx = MEMCG_LIMIT;
			mi = memcg_tree[memcg_idx]->info;
			proc_set_process_memory_state(pai, memcg_idx, mi, oom_score_adj);
			return;
		}
	}
	cgroup_write_node_uint32(mi->name, CGROUP_FILE_NAME, pid);
	proc_set_process_memory_state(pai, memcg_idx, mi, oom_score_adj);

	/*
	 * We should first move process to cgroup and then start reclaim on that
	 * cgroup.
	 */
	if (should_swap)
		lowmem_swap_memory(memcg_idx, memcg_tree[memcg_idx]->info);

}

static int lowmem_activate_worker(void)
{
	int ret = RESOURCED_ERROR_NONE;

	if (LOWMEM_WORKER_IS_ACTIVE(&lmw)) {
		_I("LMK worker thread [%d] has already been created\n",
			(unsigned)lmw.worker_thread);
		return ret;
	}

	lmw.queue = g_async_queue_new_full(lowmem_request_destroy);
	if (!lmw.queue) {
		_E("Failed to create request queue\n");
		return RESOURCED_ERROR_FAIL;
	}
	LOWMEM_WORKER_ACTIVATE(&lmw);
	ret = pthread_create(&lmw.worker_thread, NULL,
		(void *)lowmem_reclaim_worker, (void *)&lmw);
	if (ret) {
		LOWMEM_WORKER_DEACTIVATE(&lmw);
		_E("Failed to create LMK thread: %d\n", ret);
	} else {
		pthread_detach(lmw.worker_thread);
		ret = RESOURCED_ERROR_NONE;
	}
	return ret;
}

static void lowmem_deactivate_worker(void)
{
	struct lowmem_control *ctl;

	if (!LOWMEM_WORKER_IS_ACTIVE(&lmw))
		return;

	LOWMEM_WORKER_DEACTIVATE(&lmw);
	lowmem_drain_queue(&lmw);

	ctl = LOWMEM_NEW_REQUEST();
	if (!ctl) {
		_E("Critical - g_slice alloc failed - Lowmem cannot be deactivated");
		return;
	}
	ctl->flags = OOM_DROP;
	g_async_queue_push(lmw.queue, ctl);
	g_async_queue_unref(lmw.queue);
}

static int create_memcgs(void)
{
	int i = 0;
	int ret = RESOURCED_ERROR_NONE;
	struct memcg_info *mi;
	char *name;

	/* skip for memory cgroup */
	for (i = 0; i < MEMCG_MAX; i++) {
		if (memcg_root == memcg_tree[i]->info)
			continue;
		mi = memcg_tree[i]->info;
		name = mi->hashname;
		ret = cgroup_make_subdir(LOWMEM_ROOT_CGROUP, name, NULL);
		_D("create memory cgroup for %s, ret = %d", name, ret);
	}

	return ret;
}

static unsigned int lowmem_press_eventfd_read(int fd)
{
	unsigned int ret;
	uint64_t dummy_state;

	ret = read(fd, &dummy_state, sizeof(dummy_state));
	return ret;
}

static void lowmem_press_root_cgroup_handler(void)
{
	static unsigned int prev_available;
	unsigned int available;
	int mem_state;

	available = proc_get_mem_available();
	if (prev_available == available)
		return;

	mem_state = check_mem_state(available);
	lowmem_trigger_memory_state_action(mem_state);

	prev_available = available;
}

static void lowmem_press_cgroup_handler(enum lmk_type type, struct memcg_info *mi)
{
	unsigned int usage, threshold;
	int ret;

	ret = memcg_get_anon_usage(mi->name, &usage);
	if (ret) {
		_D("getting anonymous memory usage fails");
		return;
	}

	threshold = mi->threshold[LOWMEM_MEDIUM];
	if (usage >= threshold)
		memory_cgroup_medium_act(type, mi);
	else
		_I("anon page %u MB < medium threshold %u MB", BYTE_TO_MBYTE(usage),
				BYTE_TO_MBYTE(threshold));
}

static bool lowmem_press_eventfd_handler(int fd, void *data)
{
	int i;
	struct memcg_info *mi;
	GSList *iter = NULL;
	enum lmk_type lmk_type = LMK_MEMORY;

	// FIXME: probably shouldn't get ignored
	(void)lowmem_press_eventfd_read(fd);

	for (i = 0; i < MEMCG_MAX; i++) {
		if (!memcg_tree[i] || !memcg_tree[i]->info)
			continue;
		mi = memcg_tree[i]->info;
		if (fd == mi->evfd) {
			/* call low memory handler for this memcg */
			if (i == MEMCG_MEMORY)
				lowmem_press_root_cgroup_handler();
			else {
				if (i == MEMCG_APPS)
					lmk_type = LMK_ACTIVE;
				else if (i == MEMCG_SWAP)
					lmk_type = LMK_OLDEST;
				lowmem_press_cgroup_handler(lmk_type, mi);
			}
			return true;
		}
		/* ToDo: iterate child memcgs */
		gslist_for_each_item(iter, memcg_tree[i]->cgroups)
		{
			mi = (struct memcg_info *)(iter->data);
			if (fd == mi->evfd) {
				if (i == MEMCG_APPS)
					lmk_type = LMK_ACTIVE;
				else if (i == MEMCG_SWAP)
					lmk_type = LMK_OLDEST;
				lowmem_press_cgroup_handler(lmk_type, mi);
				_D("lowmem cgroup handler is called for %s",
						mi->name);
				return true;
			}
		}
	}

	return true;
}

static int lowmem_press_register_eventfd(struct memcg_info *mi)
{
	int evfd;
	const char *name = mi->name;
	static fd_handler_h handler;

	if (mi->threshold[LOWMEM_MEDIUM] == LOWMEM_THRES_INIT)
		return 0;

	evfd = memcg_set_eventfd(name, MEMCG_EVENTFD_MEMORY_PRESSURE,
			event_level);

	if (evfd == RESOURCED_ERROR_FAIL) {
		int saved_errno = errno;
		_I("fail to register event press fd %s cgroup", name);
		return -saved_errno;
	}

	mi->evfd = evfd;

	_I("register event fd success for %s cgroup", name);
	add_fd_read_handler(evfd, lowmem_press_eventfd_handler, NULL, NULL, &handler);
	return 0;
}

static int lowmem_press_setup_eventfd(void)
{
	unsigned int i;

	for (i = 0; i < MEMCG_MAX; i++) {
		if (!memcg_tree[i]->use_hierarchy)
			continue;

		const int r = lowmem_press_register_eventfd(memcg_tree[i]->info);
		if (r < 0)
			return RESOURCED_ERROR_FAIL;
	}
	return RESOURCED_ERROR_NONE;
}

static void lowmem_force_reclaim_cb(struct lowmem_control *ctl)
{
	lowmem_change_memory_state(LOWMEM_NORMAL, 0);
}

int lowmem_trigger_reclaim(int flags, int victims, enum lmk_type type, int threshold)
{
	struct lowmem_control *ctl = LOWMEM_NEW_REQUEST();

	if (!ctl)
		return -ENOMEM;

	flags |= OOM_FORCE | OOM_IN_DEPTH | OOM_SINGLE_SHOT;
	victims = victims > 0 ? victims : MAX_MEMORY_CGROUP_VICTIMS;
	type = type > 0 ? type : LMK_OLDEST;
	threshold = threshold > 0 ? threshold : memcg_root->threshold_leave;

	lowmem_change_memory_state(LOWMEM_LOW, 1);
	LOWMEM_SET_REQUEST(ctl, flags,
		type, threshold, victims,
		lowmem_force_reclaim_cb);
	lowmem_queue_request(&lmw, ctl);

	return 0;
}

void lowmem_trigger_swap_reclaim(enum lmk_type type, int swap_size)
{
	int size, victims;

	victims = num_max_victims  > MAX_PROACTIVE_HIGH_VICTIMS
				 ? MAX_PROACTIVE_HIGH_VICTIMS : num_max_victims;

	size = memcg_root->threshold_leave + swap_size;
	_I("reclaim from swap module, type : %d, size : %d, victims: %d", type, size, victims);
	lowmem_trigger_reclaim(0, victims, type, size);
}

bool lowmem_fragmentated(void)
{
	struct buddyinfo bi;
	int ret;

	ret = proc_get_buddyinfo("Normal", &bi);
	if (ret < 0)
		return false;

	if (bi.page[PAGE_32K] + bi.page[PAGE_64K] + bi.page[PAGE_128K] == 0) {
		_I("fragmentation detected, need to execute proactive oom killer");
		return true;
	}
	return false;
}

static void lowmem_proactive_oom_killer(int flags, char *appid)
{
	unsigned int before;
	int victims;

	before = proc_get_mem_available();

	/* If low memory state, just return and kill in oom killer */
	if (before < memcg_root->threshold[LOWMEM_MEDIUM])
		return;

	victims = num_max_victims  > MAX_PROACTIVE_HIGH_VICTIMS
				 ? MAX_PROACTIVE_HIGH_VICTIMS : num_max_victims;
#ifdef HEART_SUPPORT
	struct heart_memory_data *md = NULL;

	/* Get HEART-memory data only if this module is enabled */
	if (find_module("MEMORY") != NULL)
		md = heart_memory_get_data(appid, DATA_LATEST);

	if (md) {
		unsigned int rss, after, size;

		rss = KBYTE_TO_MBYTE(md->avg_rss);

		free(md);

		after = before - rss;
		/*
		 * after launching app, ensure that available memory is
		 * above threshold_leave
		 */
		if (after >= memcg_root->threshold[LOWMEM_MEDIUM])
			return;

		if (proactive_threshold - rss >= memcg_root->threshold[LOWMEM_MEDIUM])
			size = proactive_threshold;
		else
			size = rss + memcg_root->threshold[LOWMEM_MEDIUM] + THRESHOLD_MARGIN;

		_D("history based proactive LMK : avg rss %u, available %u required = %u MB",
			rss, before, size);
		lowmem_trigger_reclaim(0, victims, LMK_OLDEST, size);

		return;
	}
#endif

	/*
	 * run proactive oom killer only when available is larger than
	 * proactive threshold
	 */
	if (!proactive_threshold || before >= proactive_threshold)
		return;

	/*
	 * When there is no history data for the launching app,
	 * it is necessary to check current fragmentation state or application manifest file.
	 * So, resourced feels proactive LMK is required, run oom killer based on dynamic
	 * threshold.
	 */
	if (!lowmem_fragmentated() && !(flags & PROC_LARGEMEMORY))
		return;

	/*
	 * free THRESHOLD_MARGIN more than real should be freed,
	 * because launching app is consuming up the memory.
	 */
	_D("Run threshold based proactive LMK: memory level to reach: %u\n",
		proactive_leave + THRESHOLD_MARGIN);
	lowmem_trigger_reclaim(0, victims, LMK_OLDEST, proactive_leave + THRESHOLD_MARGIN);
	return;
}

unsigned int lowmem_get_proactive_thres(void)
{
	return proactive_threshold;
}

static int lowmem_prelaunch_handler(void *data)
{
	struct proc_status *ps = (struct proc_status *)data;

	if (!ps->pai)
		return RESOURCED_ERROR_NO_DATA;

	lowmem_proactive_oom_killer(ps->pai->flags, ps->pai->appid);
	return RESOURCED_ERROR_NONE;
}

static int lowmem_control_handler(void *data)
{
	struct lowmem_control_data *lowmem_data;

	lowmem_data = (struct lowmem_control_data *)data;
	switch (lowmem_data->control_type) {
	case LOWMEM_MOVE_CGROUP:
		lowmem_move_memcgroup((pid_t)lowmem_data->args[0],
					(int)lowmem_data->args[1]);
		break;
	default:
		break;
	}
	return RESOURCED_ERROR_NONE;
}

static int lowmem_bg_reclaim_handler(void *data)
{
	struct swap_status_msg msg;

	if (swap_get_state() != SWAP_ON)
		return RESOURCED_ERROR_NONE;

	if (!bg_reclaim)
		return RESOURCED_ERROR_NONE;

	/*
	 * Proactively reclaiming memory used by long-lived background processes
	 * (such as widget instances) may be efficient on devices with limited
	 * memory constraints. The pages used by such processes could be reclaimed
	 * (if swap is enabled) earlier than they used to while minimizing the
	 * impact on the user experience.
	 */
	msg.type = MEMCG_BGLOCKED;
	msg.info = memcg_tree[msg.type]->info;
	resourced_notify(RESOURCED_NOTIFIER_SWAP_START, &msg);

	return RESOURCED_ERROR_NONE;
}

static int allocate_vip_app_list(void)
{
	vip_apps = g_ptr_array_new();
	if (!vip_apps) {
		_E("g_ptr_array_new : out of memory");
		return RESOURCED_ERROR_OUT_OF_MEMORY;
	}

	return RESOURCED_ERROR_NONE;
}

static int set_vip_list(void)
{
	pid_t pid = -1;
	DIR *dp;
	struct dirent *dentry;
	char proc_name[PROC_NAME_MAX];
	int vip_index;
	char *vip_name;

	dp = opendir("/proc");
	if (!dp) {
		_E("fail to open /proc");
		return RESOURCED_ERROR_FAIL;
	}

	while ((dentry = readdir(dp))) {
		if (!isdigit(dentry->d_name[0]))
			continue;

		pid = atoi(dentry->d_name);
		if (!pid)
			continue;

		if (proc_get_cmdline(pid, proc_name) != RESOURCED_ERROR_NONE)
			continue;

		for (vip_index = 0; vip_index < vip_apps->len; vip_index++) {
			vip_name = g_ptr_array_index(vip_apps, vip_index);
			if (strncmp(vip_name, proc_name, strlen(proc_name)))
				continue;

			if (pid > 0) {
				proc_set_oom_score_adj(pid, OOMADJ_SERVICE_MIN);
				break;
			}
		}
	}
	closedir(dp);

	return RESOURCED_ERROR_NONE;
}

static void free_vip_app_list(void)
{
	if (vip_apps) {
		g_ptr_array_foreach(vip_apps, (GFunc)g_free, NULL);
		g_ptr_array_free(vip_apps, true);
		vip_apps = NULL;
	}
}

/* To Do: should we need lowmem_fd_start, lowmem_fd_stop ?? */
static int lowmem_init(void)
{
	int ret = RESOURCED_ERROR_NONE;

	get_total_memory();

	init_memcg_params();
	setup_memcg_params();
	if (allocate_vip_app_list() != RESOURCED_ERROR_NONE)
		_E("allocate_vip_app_list FAIL");

	if (config_parse(MEM_CONF_FILE, load_vip_config, NULL))
		_E("(%s) parse Fail", MEM_CONF_FILE);

	if (set_vip_list() != RESOURCED_ERROR_NONE)
		_E("set_vip_list FAIL");

	/* vip_list is only needed at the set_vip_list */
	free_vip_app_list();
	config_parse(MEM_CONF_FILE, load_mem_config, NULL);
	config_parse(MEM_CONF_FILE, load_bg_reclaim_config, NULL);

	create_memcgs();
	write_memcg_params();

	ret = lowmem_activate_worker();
	if (ret) {
		_E("oom thread create failed\n");
		return ret;
	}

	/* register threshold and event fd */
	ret = lowmem_press_setup_eventfd();
	if (ret) {
		_E("eventfd setup failed");
		return ret;
	}

	lowmem_dbus_init();
	lowmem_limit_init();
	lowmem_system_init();

	register_notifier(RESOURCED_NOTIFIER_APP_PRELAUNCH, lowmem_prelaunch_handler);
	register_notifier(RESOURCED_NOTIFIER_MEM_CONTROL, lowmem_control_handler);
	register_notifier(RESOURCED_NOTIFIER_LCD_OFF, lowmem_bg_reclaim_handler);

	return ret;
}

static int lowmem_exit(void)
{
	int i;
	for (i = 0; i < MEMCG_MAX; i++) {
		g_slist_free_full(memcg_tree[i]->cgroups, free);
		free(memcg_tree[i]);
	}
	if (strncmp(event_level, MEMCG_DEFAULT_EVENT_LEVEL, sizeof(MEMCG_DEFAULT_EVENT_LEVEL)))
		free(event_level);

	lowmem_deactivate_worker();
	lowmem_limit_exit();
	lowmem_system_exit();

	unregister_notifier(RESOURCED_NOTIFIER_APP_PRELAUNCH, lowmem_prelaunch_handler);
	unregister_notifier(RESOURCED_NOTIFIER_MEM_CONTROL, lowmem_control_handler);
	unregister_notifier(RESOURCED_NOTIFIER_LCD_OFF, lowmem_bg_reclaim_handler);

	return RESOURCED_ERROR_NONE;
}

static int resourced_memory_init(void *data)
{
	lowmem_ops = &memory_modules_ops;
	return lowmem_init();
}

static int resourced_memory_finalize(void *data)
{
	return lowmem_exit();
}

void lowmem_change_memory_state(int state, int force)
{
	int mem_state;

	if (force) {
		mem_state = state;
	} else {
		unsigned int available = proc_get_mem_available();
		mem_state = check_mem_state(available);
	}

	lowmem_trigger_memory_state_action(mem_state);
}

void lowmem_memcg_set_threshold(int type, int level, int value)
{
	memcg_tree[type]->info->threshold[level] = value;
}

void lowmem_memcg_set_leave_threshold(int type, int value)
{
	memcg_tree[type]->info->threshold_leave = value;
}

unsigned long lowmem_get_ktotalram(void)
{
	return ktotalram;
}

int lowmem_get_memcg(enum memcg_type type, struct memcg **memcg_ptr)
{

	if (memcg_ptr == NULL || memcg_tree == NULL || type >= MEMCG_MAX)
		return RESOURCED_ERROR_FAIL;

	*memcg_ptr = memcg_tree[type];

	return RESOURCED_ERROR_NONE;
}

void lowmem_restore_memcg(struct proc_app_info *pai)
{
	char *cgpath;
	int index, ret;
	struct memcg_info *mi;
	pid_t pid = pai->main_pid;

	ret = cgroup_pid_get_path("memory", pid, &cgpath);
	if (ret < 0)
		return;

	for (index = MEMCG_MAX-1; index >= MEMCG_MEMORY; index--) {
		mi = &gmi[index];
		if (strstr(cgpath, mi->hashname))
			break;
	}
	pai->memory.memcg_idx = index;
	pai->memory.memcg_info = mi;
	if (index == MEMCG_LIMIT)
		pai->memory.use_mem_limit = true;

	free(cgpath);
}

static struct module_ops memory_modules_ops = {
	.priority	= MODULE_PRIORITY_HIGH,
	.name		= "lowmem",
	.init		= resourced_memory_init,
	.exit		= resourced_memory_finalize,
};

MODULE_REGISTER(&memory_modules_ops)
