/*
 * resourced
 *
 * Copyright (c) 2000 - 2013 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

/*
 * @file proc-main.c
 *
 * Copyright (c) 2013 Samsung Electronics Co., Ltd. All rights reserved.
 *
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <glib-object.h>
#include <gio/gio.h>
#include <pthread.h>

#include "freezer.h"
#include "notifier.h"
#include "proc-process.h"
#include "proc-main.h"
#include "cgroup.h"
#include "trace.h"
#include "proc-handler.h"
#include "proc-monitor.h"
#include "module.h"
#include "macro.h"
#include "appid-helper.h"
#include "lowmem-handler.h"
#include "procfs.h"
#include "appinfo-list.h"
#include "util.h"
#include "file-helper.h"
#include "config-parser.h"

#ifndef gettid
#include <sys/syscall.h>

#ifdef SYS_gettid
#define gettid() (pid_t) syscall(SYS_gettid)
#else
#error "SYS_gettid unavailable on this system"
#endif
#endif

#define WATCHDOG_EXCLUDE_CONF_FILE		RD_CONFIG_FILE(proc)
#define WATCHDOG_EXCLUDE_CONF_SECTION	"WATCHDOG_EXCLUDED_PROCESSES"

static GHashTable *watchdog_exclude_list;

static const struct module_ops *freezer;
static GSList *proc_module;  /* proc sub-module list */

#define LOG_PREFIX "resourced"
#define TIZEN_SYSTEM_APPID "org.tizen.system"

struct proc_app_list app_list = { .lock = PTHREAD_MUTEX_INITIALIZER };
static GSList *proc_program_list;
static struct proc_app_info *prelaunch;

static void proc_remove_app_info(struct proc_app_info *pai);

GSList *proc_app_list_open(void)
{
	if (app_list.tid != gettid()) {
		pthread_mutex_lock(&app_list.lock);
		app_list.tid = gettid();
	}

	app_list.ref++;

	return app_list.list;
}

void proc_app_list_close(void)
{
	int err;

	app_list.ref--;

	if (app_list.ref)
		return;

	app_list.tid = 0;

	err = pthread_mutex_unlock(&app_list.lock);
	if (err)
		_E("Failed to unlock app list mutex: %m");
}

static bool is_ui_app(enum application_type type)
{
	if (type == PROC_TYPE_GUI || type == PROC_TYPE_WIDGET ||
	    type == PROC_TYPE_WATCH)
		return true;
	return false;
}

static char *proc_get_runtime_app_info_path(const struct proc_app_info *pai)
{
	char *p = NULL;

	assert(pai);
	assert(pai->appid);

	if (asprintf(&p, RUNTIME_APP_INFO_DIR "/%s", pai->appid) < 0)
		return NULL;

	return p;
}

#define DEFINE_PROC_RUNTIME_APP_INFO_WRITE_STR(node)			\
	int proc_runtime_app_info_write_##node(const char *path,	\
					       const char *data)	\
	{								\
		char app_info_node[PATH_MAX] = "";			\
									\
		assert(path);						\
		assert(data);						\
									\
		snprintf(app_info_node, PATH_MAX, "%s/"#node"", path);	\
									\
		return fwrite_str(app_info_node, data);			\
	}

#define DEFINE_PROC_RUNTIME_APP_INFO_READ_STR(node)			\
	int proc_runtime_app_info_read_##node(const char *path,		\
					      char **data)		\
	{								\
		char app_info_node[PATH_MAX] = "";			\
									\
		assert(path);						\
		assert(data);						\
									\
		snprintf(app_info_node, PATH_MAX, "%s/"#node"", path);	\
		*data = (char*)calloc(1, PATH_MAX);			\
		if (!(*data))	\
			return -ENOMEM;	\
		return fread_str(app_info_node, data);	\
	}

#define DEFINE_PROC_RUNTIME_APP_INFO_WRITE_INT32(node)			\
	int proc_runtime_app_info_write_##node(const char *path,	\
					       int32_t data)		\
	{								\
		char app_info_node[PATH_MAX] = "";			\
									\
		assert(path);						\
									\
		snprintf(app_info_node, PATH_MAX, "%s/"#node"", path);	\
									\
		return fwrite_int(app_info_node, data);			\
	}

#define DEFINE_PROC_RUNTIME_APP_INFO_READ_INT32(node)			\
	int proc_runtime_app_info_read_##node(const char *path,		\
					      int32_t *data)		\
	{								\
		char app_info_node[PATH_MAX] = "";			\
									\
		assert(path);						\
		assert(data);						\
									\
		snprintf(app_info_node, PATH_MAX, "%s/"#node"", path);	\
									\
		return fread_int(app_info_node, data);	\
	}

#define DEFINE_PROC_RUNTIME_APP_INFO_WRITE_UINT32(node)			\
	int proc_runtime_app_info_write_##node(const char *path,	\
					       uint32_t data)		\
	{								\
		char app_info_node[PATH_MAX] = "";			\
									\
		assert(path);						\
									\
		snprintf(app_info_node, PATH_MAX, "%s/"#node"", path);	\
									\
		return fwrite_uint(app_info_node, data);		\
	}

#define DEFINE_PROC_RUNTIME_APP_INFO_READ_UINT32(node)			\
	int proc_runtime_app_info_read_##node(const char *path,		\
					      uint32_t *data)		\
	{								\
		char app_info_node[PATH_MAX] = "";			\
									\
		assert(path);						\
		assert(data);						\
									\
		snprintf(app_info_node, PATH_MAX, "%s/"#node"", path);	\
									\
		return fread_uint(app_info_node, data);	\
	}

#define DEFINE_PROC_RUNTIME_APP_INFO_WRITE_ULONG(node)			\
	int proc_runtime_app_info_write_##node(const char *path,	\
					       unsigned long data)		\
	{								\
		char app_info_node[PATH_MAX] = "";			\
									\
		assert(path);						\
									\
		snprintf(app_info_node, PATH_MAX, "%s/"#node"", path);	\
									\
		return fwrite_ulong(app_info_node, data);		\
	}

#define DEFINE_PROC_RUNTIME_APP_INFO_READ_ULONG(node)			\
	int proc_runtime_app_info_read_##node(const char *path,		\
					      unsigned long *data)		\
	{								\
		char app_info_node[PATH_MAX] = "";			\
									\
		assert(path);						\
		assert(data);						\
									\
		snprintf(app_info_node, PATH_MAX, "%s/"#node"", path);	\
									\
		return fread_ulong(app_info_node, data);	\
	}

#define DEFINE_PROC_RUNTIME_APP_INFO_REMOVE(node)			\
	int proc_runtime_app_info_remove_##node(const char *path)	\
	{								\
		char app_info_node[PATH_MAX] = "";			\
		int ret;						\
									\
		assert(path);						\
									\
		snprintf(app_info_node, PATH_MAX, "%s/"#node"", path);	\
									\
		ret = unlink(app_info_node);				\
		if (ret < 0)						\
			return -errno;					\
									\
		return ret;						\
	}

DEFINE_PROC_RUNTIME_APP_INFO_WRITE_STR(appid);
DEFINE_PROC_RUNTIME_APP_INFO_READ_STR(appid);
DEFINE_PROC_RUNTIME_APP_INFO_WRITE_UINT32(main_pid);
DEFINE_PROC_RUNTIME_APP_INFO_READ_UINT32(main_pid);
DEFINE_PROC_RUNTIME_APP_INFO_WRITE_INT32(runtime_exclude);
DEFINE_PROC_RUNTIME_APP_INFO_READ_INT32(runtime_exclude);
DEFINE_PROC_RUNTIME_APP_INFO_WRITE_INT32(flags);
DEFINE_PROC_RUNTIME_APP_INFO_READ_INT32(flags);
DEFINE_PROC_RUNTIME_APP_INFO_WRITE_INT32(lru_state);
DEFINE_PROC_RUNTIME_APP_INFO_READ_INT32(lru_state);
DEFINE_PROC_RUNTIME_APP_INFO_WRITE_INT32(categories);
DEFINE_PROC_RUNTIME_APP_INFO_READ_INT32(categories);
DEFINE_PROC_RUNTIME_APP_INFO_WRITE_INT32(state);
DEFINE_PROC_RUNTIME_APP_INFO_READ_INT32(state);
DEFINE_PROC_RUNTIME_APP_INFO_WRITE_INT32(type);
DEFINE_PROC_RUNTIME_APP_INFO_READ_INT32(type);
DEFINE_PROC_RUNTIME_APP_INFO_WRITE_STR(pkgname);
DEFINE_PROC_RUNTIME_APP_INFO_READ_STR(pkgname);
DEFINE_PROC_RUNTIME_APP_INFO_WRITE_ULONG(starttime);
DEFINE_PROC_RUNTIME_APP_INFO_READ_ULONG(starttime);

DEFINE_PROC_RUNTIME_APP_INFO_REMOVE(appid);
DEFINE_PROC_RUNTIME_APP_INFO_REMOVE(procs);
DEFINE_PROC_RUNTIME_APP_INFO_REMOVE(main_pid);
DEFINE_PROC_RUNTIME_APP_INFO_REMOVE(runtime_exclude);
DEFINE_PROC_RUNTIME_APP_INFO_REMOVE(flags);
DEFINE_PROC_RUNTIME_APP_INFO_REMOVE(lru_state);
DEFINE_PROC_RUNTIME_APP_INFO_REMOVE(categories);
DEFINE_PROC_RUNTIME_APP_INFO_REMOVE(state);
DEFINE_PROC_RUNTIME_APP_INFO_REMOVE(type);
DEFINE_PROC_RUNTIME_APP_INFO_REMOVE(pkgname);
DEFINE_PROC_RUNTIME_APP_INFO_REMOVE(starttime);


static int proc_runtime_app_info_write_procs(const char *path, pid_list child_list)
{
	char app_info_node[PATH_MAX] = "";
	_cleanup_fclose_ FILE *f = NULL;
	pid_list list = child_list;

	assert(path);

	if (!child_list)
		return 0;

	snprintf(app_info_node, PATH_MAX, "%s/procs", path);

	f = fopen(app_info_node, "w");
	if (!f)
		return -errno;

	(void) fprintf(f, "%d", GPOINTER_TO_PID(child_list->data));

	for (list = g_slist_next(list); list; list = g_slist_next(list))
		(void) fprintf(f, " %d", GPOINTER_TO_PID(list->data));

	return 0;
}

static int proc_runtime_app_info_read_procs(const char *path, pid_list *child_list)
{
	char app_info_node[PATH_MAX] = "";
	_cleanup_free_ char *buf = NULL;
	pid_list list = NULL;
	char *word, *state;
	size_t len;
	int ret;

	assert(path);
	assert(child_list);

	snprintf(app_info_node, PATH_MAX, "%s/procs", path);

	buf = calloc(1, PATH_MAX);
	if (!buf)
		return RESOURCED_ERROR_OUT_OF_MEMORY;

	ret = fread_str(app_info_node, &buf);
	if (ret < 0)
		return ret;

	FOREACH_WORD(word, len, buf, state) {
		_cleanup_free_ char *pid_s = NULL;
		pid_t pid;

		pid_s = strndup(word, len);
		if (!pid_s)
			goto on_error;

		pid = atoi(pid_s);

		list = g_slist_prepend(list, PID_TO_GPOINTER(pid));
	}

	*child_list = list;

	return 0;

on_error:
	g_slist_free(list);

	return -ENOMEM;
}

static void proc_app_info_add_child_pid(struct proc_app_info *pai, pid_t pid)
{
	_cleanup_free_ char *runtime_app_info_path = NULL;
	int ret;

	assert(pai);
	assert(pid > 0);

	if (pai->childs &&
			g_slist_find((GSList *)pai->childs, PID_TO_GPOINTER(pid)))
		return;

	pai->childs = g_slist_prepend(pai->childs, PID_TO_GPOINTER(pid));

	runtime_app_info_path = proc_get_runtime_app_info_path(pai);
	if (!runtime_app_info_path) {
		_E("Failed to get runtime appinfo path: not enough memory");
		return;
	}

	ret = proc_runtime_app_info_write_procs(runtime_app_info_path, pai->childs);
	if (ret < 0)
		_E("Failed to write appinfo children '%s': %m", runtime_app_info_path);
}

void proc_set_process_memory_state(struct proc_app_info *pai,
	int memcg_idx, struct memcg_info *memcg_info, int oom_score_adj)
{
	if (!pai)
		return;

	pai->memory.memcg_idx = memcg_idx;
	pai->memory.memcg_info = memcg_info;
	pai->memory.oom_score_adj = oom_score_adj;
}

/*
 * There can be many processes with same appid at same time.
 * This function returns the most recently used app of all app list.
 */
struct proc_app_info *find_app_info_by_appid(const char *appid)
{
	_cleanup_app_list_close_ GSList *proc_app_list = PAL_INIT_VALUE;
	GSList *iter = NULL;
	struct proc_app_info *pai;

	proc_app_list = proc_app_list_open();

	if (!appid || !proc_app_list)
		return NULL;

	gslist_for_each_item(iter, proc_app_list) {
		pai = (struct proc_app_info *)iter->data;
		if (equal_name_info(pai->appid, appid))
			return pai;
	}
	return NULL;
}

struct proc_app_info *find_app_info(const pid_t pid)
{
	_cleanup_app_list_close_ GSList *proc_app_list = PAL_INIT_VALUE;
	GSList *iter = NULL;
	struct proc_app_info *pai = NULL;

	proc_app_list = proc_app_list_open();

	if (!proc_app_list)
		return NULL;

	gslist_for_each_item(iter, proc_app_list) {
		pai = (struct proc_app_info *)iter->data;
		if (pai->main_pid == pid)
			return pai;
		if (pai->childs) {
			if (g_slist_find(pai->childs, PID_TO_GPOINTER(pid)))
				return pai;
		}
	}
	return NULL;
}

static struct proc_program_info *find_program_info(const char *pkgname)
{
	GSList *iter = NULL;
	struct proc_program_info *ppi;

	if (!pkgname)
		return NULL;

	gslist_for_each_item(iter, proc_program_list) {
		ppi = (struct proc_program_info *)iter->data;
		if (equal_name_info(ppi->pkgname, pkgname))
			return ppi;
	}
	return NULL;
}

resourced_ret_c proc_set_runtime_exclude_list(const int pid, int type)
{
	_cleanup_free_ char *runtime_app_info_path = NULL;
	struct proc_app_info *pai = NULL;
	struct proc_status ps = {0};
	int ret;

	pai = find_app_info(pid);
	if (!pai)
		return RESOURCED_ERROR_NO_DATA;

	ps.pid = pid;
	ps.pai = pai;
	if (type == PROC_EXCLUDE)
		resourced_notify(RESOURCED_NOTIFIER_APP_WAKEUP, &ps);

	if (pai->runtime_exclude) {
		if (type == PROC_EXCLUDE)
			pai->runtime_exclude++;
		else
			pai->runtime_exclude--;
	} else {
		pai->runtime_exclude = type;
	}
	_D("pid %d set proc exclude list, type = %d, exclude = %d",
		    pid, type, pai->runtime_exclude);

	runtime_app_info_path = proc_get_runtime_app_info_path(pai);
	if (!runtime_app_info_path)
		return RESOURCED_ERROR_OUT_OF_MEMORY;

	ret = proc_runtime_app_info_write_runtime_exclude(runtime_app_info_path, pai->runtime_exclude);
	if (ret < 0)
		return RESOURCED_ERROR_FAIL;

	return RESOURCED_ERROR_NONE;
}

/*
  * find main oom score value from latest launched UI application
  * And set oom score of service app
  */
static void proc_set_default_svc_oomscore
	    (struct proc_program_info *ppi, struct proc_app_info *svc)
{
	struct proc_app_info *pai =
		    (struct proc_app_info *)g_slist_nth_data(ppi->app_list, 0);
	int oom_score_adj = 0, ret ;
	if (pai) {
		if (CHECK_BIT(pai->flags, PROC_VIP_ATTRIBUTE))
			oom_score_adj = OOMADJ_SU;
		else {
			ret = proc_get_oom_score_adj(pai->main_pid, &oom_score_adj);
			if (ret)
				oom_score_adj = 0;
		}
	}
	proc_set_service_oomscore(svc->main_pid, oom_score_adj);
}

static struct proc_program_info *proc_add_program_list(const int type,
	    struct proc_app_info *pai, const char *pkgname, bool restore)
{
	struct proc_program_info *ppi;

	if (!pai || !pkgname || pai->main_pid <= 0)
		return NULL;

	ppi = find_program_info(pkgname);
	if (!ppi) {
		_D("Can't find %s in the program list. Start to add it", pkgname);
		ppi = calloc(sizeof(struct proc_program_info), 1);
		if (!ppi) {
			_E("Fail to add %s : not enough memory", pkgname);
			return NULL;
		}

		if (pai->ai)
			ppi->pkgname = strdup(pai->ai->pkgname);
		else
			ppi->pkgname = strndup(pkgname, strlen(pkgname)+1);

		if (!ppi->pkgname) {
			_E("Fail to add %s : not enough memory", pkgname);
			free(ppi);
			return NULL;
		}
		proc_program_list = g_slist_prepend(proc_program_list, ppi);
		_I("%s is added in the program info", pkgname);
	}
	if (is_ui_app(type))
		ppi->app_list = g_slist_prepend(ppi->app_list, pai);
	else {
		ppi->svc_list = g_slist_prepend(ppi->svc_list, pai);
		if (!restore)
			proc_set_default_svc_oomscore(ppi, pai);
	}
	return ppi;
}

static int proc_runtime_write_app_info(const struct proc_app_info *pai)
{
	_cleanup_free_ char *runtime_app_info_path = NULL;
	int ret;

	assert(pai);
	assert(pai->appid);

	runtime_app_info_path = proc_get_runtime_app_info_path(pai);
	if (!runtime_app_info_path)
		return -ENOMEM;

	ret = mkdir(runtime_app_info_path, S_IRWXU | S_IRGRP | S_IXGRP);
	if (ret && errno != EEXIST)
		return ret;

	ret = proc_runtime_app_info_write_appid(runtime_app_info_path, pai->appid);
	if (ret < 0)
		return ret;

	ret = proc_runtime_app_info_write_main_pid(runtime_app_info_path, pai->main_pid);
	if (ret < 0)
		return ret;

	ret = proc_runtime_app_info_write_procs(runtime_app_info_path, pai->childs);
	if (ret < 0)
		return ret;

	ret = proc_runtime_app_info_write_runtime_exclude(runtime_app_info_path, pai->runtime_exclude);
	if (ret < 0)
		return ret;

	ret = proc_runtime_app_info_write_flags(runtime_app_info_path, pai->flags);
	if (ret < 0)
		return ret;

	ret = proc_runtime_app_info_write_lru_state(runtime_app_info_path, pai->lru_state);
	if (ret < 0)
		return ret;

	ret = proc_runtime_app_info_write_categories(runtime_app_info_path, pai->categories);
	if (ret < 0)
		return ret;

	ret = proc_runtime_app_info_write_state(runtime_app_info_path, pai->state);
	if (ret < 0)
		return ret;

	ret = proc_runtime_app_info_write_type(runtime_app_info_path, pai->type);
	if (ret < 0)
		return ret;

	ret = proc_runtime_app_info_write_starttime(runtime_app_info_path, pai->starttime);
	if (ret < 0)
		return ret;

	if (!pai->program)
		return 0;

	return proc_runtime_app_info_write_pkgname(runtime_app_info_path, pai->program->pkgname);
}

static int proc_runtime_remove_app_info(const struct proc_app_info *pai)
{
	_cleanup_free_ char *runtime_app_info_path = NULL;
	int ret;
	DIR *dir;

	assert(pai);
	assert(pai->appid);

	runtime_app_info_path = proc_get_runtime_app_info_path(pai);
	if (!runtime_app_info_path)
		return -ENOMEM;

	dir = opendir(runtime_app_info_path);
	if (!dir)
		return -ENOTDIR;
	closedir(dir);

	ret = proc_runtime_app_info_remove_appid(runtime_app_info_path);
	if (ret < 0 && ret != -ENOENT)
		goto finish;

	ret = proc_runtime_app_info_remove_main_pid(runtime_app_info_path);
	if (ret < 0 && ret != -ENOENT)
		goto finish;

	ret = proc_runtime_app_info_remove_procs(runtime_app_info_path);
	if (ret < 0 && ret != -ENOENT)
		goto finish;

	ret = proc_runtime_app_info_remove_runtime_exclude(runtime_app_info_path);
	if (ret < 0 && ret != -ENOENT)
		goto finish;

	ret = proc_runtime_app_info_remove_flags(runtime_app_info_path);
	if (ret < 0 && ret != -ENOENT)
		goto finish;

	ret = proc_runtime_app_info_remove_lru_state(runtime_app_info_path);
	if (ret < 0 && ret != -ENOENT)
		goto finish;

	ret = proc_runtime_app_info_remove_categories(runtime_app_info_path);
	if (ret < 0 && ret != -ENOENT)
		goto finish;

	ret = proc_runtime_app_info_remove_state(runtime_app_info_path);
	if (ret < 0 && ret != -ENOENT)
		goto finish;

	ret = proc_runtime_app_info_remove_type(runtime_app_info_path);
	if (ret < 0 && ret != -ENOENT)
		goto finish;

	ret = proc_runtime_app_info_remove_pkgname(runtime_app_info_path);
	if (ret < 0 && ret != -ENOENT)
		goto finish;

	ret = proc_runtime_app_info_remove_starttime(runtime_app_info_path);
	if (ret < 0 && ret != -ENOENT)
		goto finish;

finish:
	if (rmdir(runtime_app_info_path) < 0)
		return -errno;

	return ret == -ENOENT ? 0 : ret;
}

static void proc_app_list_add_app_info(struct proc_app_info *pai)
{
	GSList *proc_app_list = PAL_INIT_VALUE;
	int ret;

	if (pai->main_pid <= 0)
		return;

	ret = proc_runtime_write_app_info(pai);
	if (ret < 0)
		_E("Failed to add runtime app info: %m");

	proc_app_list = proc_app_list_open();
	app_list.list = g_slist_prepend(proc_app_list, pai);
	proc_app_list_close();
}

static GSList *proc_app_list_remove_app_info(struct proc_app_info *pai)
{
	_cleanup_app_list_close_ GSList *proc_app_list = PAL_INIT_VALUE;
	int ret;

	ret = proc_runtime_remove_app_info(pai);
	if (ret < 0)
		_E("Failed to remove appinfo '%s': %m", pai->appid);

	proc_app_list = proc_app_list_open();
	app_list.list = g_slist_remove(proc_app_list, pai);

	return app_list.list;
}

static struct proc_app_info *proc_create_app_info(const char *appid,
					   const char *pkgid,
					   pid_t pid,
					   int flags,
					   int categories,
					   enum application_type type,
					   enum proc_state state)
{
	struct proc_app_info *pai;
	unsigned long uptime = 0;
	int ret;

	if (!appid || !pkgid) {
		_E("Each app should have appid and pkgid");
		return NULL;
	}

	/*
	 * app info creation is able to be performed with pid 0 by
	 * prelaunch request. In this case, resourced will create an
	 * app info to prelaunch app info not app info list. The
	 * created prelaunch app info will be used next app info
	 * creation.
	 */
	if (pid > 0) {
		/* launch */
		if (prelaunch) {
			pai = prelaunch;
			prelaunch = NULL;

			if (!streq(pai->appid, appid)) {
				_E("prelaunched app(%s) is not matched with new app info(%s)",
					pai->appid, appid);
				proc_remove_app_info(pai);

				pai = calloc(sizeof(struct proc_app_info), 1);
				if (!pai)
					return NULL;
			}
		} else {
			pai = calloc(sizeof(struct proc_app_info), 1);
			if (!pai)
				return NULL;

			_I("app_info(%s) will be added without prelaunch", appid);
		}
	} else {
		/* prelaunch */
		if (prelaunch) {
			_E("prelaunched app was not gone to launch state, appid(%s)", prelaunch->appid);
			proc_remove_app_info(prelaunch);
			prelaunch = NULL;
		}

		pai = calloc(sizeof(struct proc_app_info), 1);
		if (!pai)
			return NULL;

		prelaunch = pai;
	}

	if (!pai->ai) {
		pai->ai = resourced_appinfo_get(pai->ai, appid, pkgid);
		if (!pai->ai) {
			_E("Failed to get resourced_appinfo, appid(%s), pkgid(%s)", appid, pkgid);
			free(pai);
			return NULL;
		}
	}

	pai->appid = pai->ai->appid;

	pai->watchdog_exclude = resourced_watchdog_excluded(appid);

	pai->type = type;
	pai->state = state;
	pai->main_pid = pid;

	if (flags)
		pai->flags = flags;
	if (categories)
		pai->categories = categories;

	if (pid > 0) {
		pai->program = proc_add_program_list(type, pai, pkgid, false);
		if (!pai->starttime) {
			ret = proc_get_uptime(&uptime);
			if (ret)
				_E("Failed to get uptime");
			else
				pai->starttime = uptime;
		}
	}
	return pai;
}
struct proc_app_info *proc_add_app_info(const char *appid,
					const char *pkgid,
					pid_t pid,
					int flags,
					int categories,
					enum application_type type,
					enum proc_state state)
{
	struct proc_app_info *pai;

	pai = proc_create_app_info(appid, pkgid, pid, flags, categories, type, state);
	if (!pai)
		return NULL;

	(void) proc_app_list_add_app_info(pai);

	return pai;
}

static void proc_app_info_remove_child_pid(struct proc_app_info *pai, pid_t pid)
{
	_cleanup_free_ char *runtime_app_info_path = NULL;
	int ret;

	assert(pai);
	assert(pid >= 0);

	if (!pai->childs)
		return;

	runtime_app_info_path = proc_get_runtime_app_info_path(pai);
	if (!runtime_app_info_path) {
		_E("Failed to get runtime appinfo path: not enough memory");
		return;
	}

	if (pid == 0)
		goto remove_all;

	if (!g_slist_find(pai->childs, PID_TO_GPOINTER(pid)))
		return;

	pai->childs = g_slist_remove(pai->childs, PID_TO_GPOINTER(pid));
	if (!g_slist_length(pai->childs))
		goto remove_all;

	ret = proc_runtime_app_info_write_procs(runtime_app_info_path, pai->childs);
	if (ret < 0)
		_E("Failed to write appinfo children '%s': %m", runtime_app_info_path);

	return;

remove_all:
	g_slist_free(pai->childs);
	pai->childs = NULL;

	ret = proc_runtime_app_info_remove_procs(runtime_app_info_path);
	if (ret < 0)
		_E("Failed to remove appinfo children '%s': %m", runtime_app_info_path);

	return;
}

static void proc_remove_program_info(struct proc_program_info *ppi,
				     struct proc_app_info *pai)
{
	if (!ppi || !pai)
		return;

	if (is_ui_app(pai->type))
		ppi->app_list = g_slist_remove(ppi->app_list, pai);
	else if (pai->type == PROC_TYPE_SERVICE)
		ppi->svc_list = g_slist_remove(ppi->svc_list, pai);

	if (!ppi->app_list && !ppi->svc_list) {
		proc_program_list = g_slist_remove(proc_program_list, ppi);
		free(ppi);
	}
}

static void proc_remove_app_info(struct proc_app_info *pai)
{
	if (!pai)
		return;

	proc_app_info_remove_child_pid(pai, 0);
	proc_remove_program_info(pai->program, pai);

	(void) proc_app_list_remove_app_info(pai);
	resourced_appinfo_put(pai->ai);
	free(pai);
}

static void proc_app_info_remove_pid(struct proc_app_info *pai, pid_t pid)
{
	g_assert(pid > 0);

	if (!pai)
		return;

	if (pai->main_pid == pid)
		proc_remove_app_info(pai);
	else
		proc_app_info_remove_child_pid(pai, pid);
}

static int proc_delete_all_lists(void)
{
	_cleanup_app_list_close_ GSList *proc_app_list = PAL_INIT_VALUE;
	GSList *iter, *next;
	struct proc_app_info *pai = NULL;
	struct proc_program_info *ppi = NULL;

	proc_app_list = proc_app_list_open();

	gslist_for_each_safe(proc_app_list, iter, next, pai)
		proc_remove_app_info(pai);

	gslist_for_each_safe(proc_program_list, iter, next, ppi) {
		proc_program_list = g_slist_remove(proc_program_list, ppi);
		free(ppi);
	}
	return 0;
}

int proc_get_svc_state(struct proc_program_info *ppi)
{
	GSList *iter = NULL;
	int state = PROC_STATE_DEFAULT;

	if (!ppi->app_list)
		return PROC_STATE_DEFAULT;

	gslist_for_each_item(iter, ppi->app_list) {
		struct proc_app_info *pai = (struct proc_app_info *)(iter->data);

		if (pai->lru_state == PROC_FOREGROUND)
			return PROC_STATE_FOREGROUND;

		if (pai->lru_state >= PROC_BACKGROUND)
			state = PROC_STATE_BACKGROUND;
	}
	return state;
}

static void proc_dump_process_list(FILE *fp)
{
	_cleanup_app_list_close_ GSList *proc_app_list = PAL_INIT_VALUE;
	GSList *iter, *iter_child;
	struct proc_app_info *pai = NULL;
	int index = 0, ret, oom_score_adj;

	LOG_DUMP(fp, "[APPLICATION LISTS]\n");
	proc_app_list = proc_app_list_open();
	gslist_for_each_item(iter, proc_app_list) {
		unsigned int size;
		unsigned long utime, stime, starttime;

		pai = (struct proc_app_info *)iter->data;
		ret = proc_get_oom_score_adj(pai->main_pid, &oom_score_adj);
		if (ret < 0)
			continue;

		if (!pai->ai)
			continue;

		LOG_DUMP(fp, "index : %d, type : %d, pkgname : %s, appid : %s\n"
		    "\t lru : %d, watchdog_exclude : %d, runtime_exclude : %d, flags : %X, "
		    "state : %d\n", index, pai->type, pai->ai->pkgname, pai->ai->appid,
		    pai->lru_state, pai->watchdog_exclude, pai->runtime_exclude,
		    pai->flags, pai->state);

		if (proc_get_mem_usage(pai->main_pid, &size))
			size = 0;
		proc_get_cpu_time(pai->main_pid, &utime, &stime, &starttime);
		LOG_DUMP(fp, "\t main pid : %d, oom score : %d, memory rss : %d KB,"
		    "utime : %lu, stime : %lu, starttime : %lu\n", pai->main_pid, oom_score_adj, size,
		    utime, stime, starttime);
		if (pai->childs) {
			gslist_for_each_item(iter_child, pai->childs) {
				if (proc_get_mem_usage(GPOINTER_TO_PID(iter_child->data), &size))
					size = 0;
				proc_get_cpu_time(GPOINTER_TO_PID(iter_child->data),
								&utime, &stime, &starttime);
				LOG_DUMP(fp, "\t main pid : %d, oom score : %d, "
					"memory rss : %dKB, utime : %lu, stime : %lu, starttime : %lu\n",
					pai->main_pid, oom_score_adj, size, utime, stime, starttime);
			}
		}
		index++;
	}
}

static void proc_free_watchdog_exclude_key(gpointer data)
{
	if (data)
		free(data);
}

int proc_get_id_info(struct proc_status *ps, char **app_name, char **pkg_name)
{
	if (!ps)
		return RESOURCED_ERROR_INVALID_PARAMETER;

	if (!ps->pai || !ps->pai->ai) {
		*app_name = TIZEN_SYSTEM_APPID;
		*pkg_name = TIZEN_SYSTEM_APPID;
	} else {
		*app_name = ps->pai->ai->appid;
		*pkg_name = ps->pai->ai->pkgname;
	}
	return RESOURCED_ERROR_NONE;
}

char *proc_get_appid_from_pid(const pid_t pid)
{
	struct proc_app_info *pai = find_app_info(pid);
	if (!pai)
		return NULL;
	return pai->appid;
}

int resourced_watchdog_excluded(const char *app_name)
{
	gboolean ret = 0;
	if (watchdog_exclude_list)
		ret = g_hash_table_contains(watchdog_exclude_list, (gpointer)app_name);

	return ret ? RESOURCED_ERROR_NONMONITOR : RESOURCED_ERROR_NONE;
}

static int load_watchdog_exclude_config(struct parse_result *result, void *user_data)
{
	if (!result)
		return RESOURCED_ERROR_INVALID_PARAMETER;

	if (!strncmp(result->section, WATCHDOG_EXCLUDE_CONF_SECTION, strlen(WATCHDOG_EXCLUDE_CONF_SECTION) + 1) &&
		!strncmp(result->name, "PREDEFINE", 10)) {
			g_hash_table_insert(watchdog_exclude_list,
					g_strndup(result->value, strlen(result->value)),
					GINT_TO_POINTER(1));
	}

	return RESOURCED_ERROR_NONE;
}

static void watchdog_exclude_init(void)
{
	watchdog_exclude_list = g_hash_table_new_full(
		g_str_hash,
		g_str_equal,
		proc_free_watchdog_exclude_key,
		NULL);

	if (watchdog_exclude_list == NULL) {
		_E("Can't initialize exclude_list!");
		return;
	}

	config_parse(WATCHDOG_EXCLUDE_CONF_FILE, load_watchdog_exclude_config, NULL);
}

void proc_module_add(const struct proc_module_ops *ops)
{
	proc_module = g_slist_append(proc_module, (gpointer)ops);
}

void proc_module_remove(const struct proc_module_ops *ops)
{
	proc_module = g_slist_remove(proc_module, (gpointer)ops);
}

static void proc_module_init(void *data)
{
	GSList *iter;
	const struct proc_module_ops *module;
	int ret = RESOURCED_ERROR_NONE;

	gslist_for_each_item(iter, proc_module) {
		module = (struct proc_module_ops *)iter->data;
		_D("Initialize [%s] module\n", module->name);
		if (module->init)
			ret = module->init(data);
		if (ret != RESOURCED_ERROR_NONE)
			_E("Fail to initialize [%s] module\n", module->name);
	}
}

static void proc_module_exit(void *data)
{
	GSList *iter;
	const struct proc_module_ops *module;
	int ret = RESOURCED_ERROR_NONE;

	gslist_for_each_item(iter, proc_module) {
		module = (struct proc_module_ops *)iter->data;
		_D("Deinitialize [%s] module\n", module->name);
		if (module->exit)
			ret = module->exit(data);
		if (ret != RESOURCED_ERROR_NONE)
			_E("Fail to deinitialize [%s] module\n", module->name);
	}
}

static int proc_restore_runtime_app_info(const char *path)
{
	_cleanup_free_ char *appid = NULL, *pkgname = NULL;
	enum application_type type;
	struct proc_app_info *pai;
	int oom_score_adj = 0, ret;
	int flags, categories, state;
	pid_t main_pid;

	ret = proc_runtime_app_info_read_appid(path, &appid);
	if (ret < 0)
		return ret;

	ret = proc_runtime_app_info_read_pkgname(path, &pkgname);
	if (ret < 0)
		return ret;

	ret = proc_runtime_app_info_read_main_pid(path, (uint32_t *) &main_pid);
	if (ret < 0)
		return ret;

	ret = proc_runtime_app_info_read_flags(path, &flags);
	if (ret < 0)
		return ret;

	ret = proc_runtime_app_info_read_categories(path, &categories);
	if (ret < 0)
		return ret;

	ret = proc_runtime_app_info_read_type(path, (int32_t *) &type);
	if (ret < 0)
		return ret;

	ret = proc_runtime_app_info_read_state(path, (int32_t *) &state);
	if (ret < 0)
		return ret;

	ret = proc_get_oom_score_adj(main_pid, &oom_score_adj);
	if (ret < 0) {
		_I("pid %d is already terminated.", main_pid);
		return ret;
	}

	pai = proc_create_app_info(appid, pkgname, main_pid, flags, categories, type, state);
	if (!pai)
		return -ENOMEM;

	pai->memory.oom_score_adj = oom_score_adj;

	ret = proc_runtime_app_info_read_runtime_exclude(path, &pai->runtime_exclude);
	if (ret < 0)
		return ret;

	ret = proc_runtime_app_info_read_lru_state(path, &pai->lru_state);
	if (ret < 0)
		return ret;

	ret = proc_runtime_app_info_read_starttime(path, &pai->starttime);
	if (ret < 0)
		return ret;

	ret = proc_runtime_app_info_read_procs(path, &pai->childs);
	if (ret < 0 && ret != -ENOENT)
		return ret;

	pai->program = proc_add_program_list(pai->type, pai, pkgname, true);
#ifdef MEMORY_MODULE
	lowmem_restore_memcg(pai);
#endif

	(void) proc_app_list_add_app_info(pai);

	_I("runtime appinfo restored: %s, %s", appid, pkgname);

	return 0;
}

static void proc_restore_runtime_app_list(void)
{
	_cleanup_closedir_ DIR *dp = NULL;
	struct dirent *dir;

	dp = opendir(RUNTIME_APP_INFO_DIR);
	if (!dp)
		return;

	while ((dir = readdir(dp))) {
		_cleanup_free_ char *path = NULL;
		int ret;

		if (dir->d_name[0] == '.')
			continue;

		if (dir->d_type != DT_DIR)
			continue;

		ret = asprintf(&path, RUNTIME_APP_INFO_DIR "/%s", dir->d_name);
		if (ret < 0) {
			_E("Failed to allocate memory");
			return;
		}

		ret = proc_restore_runtime_app_info(path);
		if (ret < 0)
			_E("Failed to restore runtime appinfo '%s': %m", dir->d_name);
	}
}

static int resourced_proc_init(void* data)
{
	int ret = mkdir(RUNTIME_APP_INFO_DIR, S_IRWXU | S_IRGRP | S_IXGRP);
	if (ret < 0 && errno != EEXIST) {
		_E("Failed to create directory %s: %m", RUNTIME_APP_INFO_DIR);
		return RESOURCED_ERROR_FAIL;
	}

	watchdog_exclude_init();
	proc_module_init(data);
	return RESOURCED_ERROR_NONE;
}

static int resourced_proc_exit(void* data)
{
	if (prelaunch)
		free(prelaunch);
	proc_delete_all_lists();
	g_hash_table_destroy(watchdog_exclude_list);
	proc_module_exit(data);
	return RESOURCED_ERROR_NONE;
}

static int resourced_proc_restore(void *data)
{
	proc_restore_runtime_app_list();
	return RESOURCED_ERROR_NONE;
}

int proc_get_freezer_status()
{
	int ret = CGROUP_FREEZER_DISABLED;
	struct freezer_status_data f_data;
	if (!freezer) {
		freezer = find_module("freezer");
		if (!freezer)
			return ret;
	}

	f_data.type = GET_STATUS;
	if (freezer->status)
		ret = freezer->status(&f_data);
	return ret;
}

int proc_get_appflag(const pid_t pid)
{
	struct proc_app_info *pai =
		find_app_info(pid);

	if (pai) {
		_D("get apptype = %d", pai->flags);
		return pai->flags;
	} else
		_D("there is no process info for pid = %d", pid);
	return PROC_NONE;
}

void proc_set_group(pid_t ownerpid, pid_t childpid, char *pkgname)
{
	_cleanup_app_list_close_ GSList *proc_app_list = PAL_INIT_VALUE;
	int owner_oom = 0, child_oom = 0, ret;
	int child_type = 0, child_state = 0;
	struct proc_app_info *pai, *owner;
	struct proc_status ps = {0};
	char group_info[512];

	if (ownerpid <= 0 || childpid <= 0)
		return;

	/*
	 * perceptible application for instance home, indicator, quickpanel and so on
	 * can request to make the transient window with other processes.
	 * However it doesn't mean that these are one application group.
	 */
	ret = proc_get_oom_score_adj(ownerpid, &owner_oom);
	if (ret < 0  || owner_oom == OOMADJ_BACKGRD_PERCEPTIBLE) {
		_D("owner pid(%d) was already terminated", ownerpid);
		return;
	}

	owner = find_app_info(ownerpid);
	pai = find_app_info(childpid);
	if (!owner)
		return;

	if (pkgname && pai) {
		/*
		 * After grouping, resourced considers child pai with parent pai
		 * Therefore, child appid in HEART-CPU DB should be changed
		 * to parent appid
		 */
		snprintf(group_info, sizeof(group_info), "%d %s %s",
				childpid, pai->appid, owner->appid);
		resourced_notify(RESOURCED_NOTIFIER_APP_GROUP, group_info);

		/*
		 * when child application with appid has owner pai and ppi
		 * check and remove them
		 */
		if (pai->main_pid == childpid) {
			ret = proc_get_oom_score_adj(childpid, &child_oom);
			if (ret < 0) {
				_D("can't get oom score for pid (%d)", childpid);
				child_oom = 0;
			}
			child_type = pai->type;
			child_state = pai->state;
			/*
			 * migrate application categories and activities
			 * from child pai to parent pai
			 */
			if (pai->categories)
				owner->categories |= pai->categories;
			if (pai->runtime_exclude)
				owner->runtime_exclude += pai->runtime_exclude;
			proc_remove_app_info(pai);
		} else {
			_D("main pid(%d) was different from childpid(%d)",
			    pai->main_pid, childpid);
			proc_app_info_remove_child_pid(pai, childpid);
		}
	}
	/*
	 * when some process like webprocess or
	 * UI application that has transparent with owner application
	 * needs to group in owner app
	 * and adds to child lists in the proc app info
	 */
	proc_app_info_add_child_pid(owner, childpid);

	/*
	 * When child process was GUI application and it was foreground state
	 * parent application should go to the foreground state also.
	 * Otherwise, child process follows oom and suspend state of parent application
	 */
	if (child_type == PROC_TYPE_GUI &&
	    child_state ==  PROC_STATE_FOREGROUND) {
		owner->lru_state = PROC_FOREGROUND;
		ps.pid = owner->main_pid;
		proc_set_oom_score_adj(owner->main_pid, child_oom);
	} else {
		if (owner_oom <= OOMADJ_BACKGRD_LOCKED)
			ps.pid = childpid;
		proc_set_oom_score_adj(childpid, owner_oom);
	}

	if (!ps.pid)
		return;

	ps.pai = owner;
	resourced_notify(RESOURCED_NOTIFIER_APP_WAKEUP, &ps);
}

bool proc_check_lru_suspend(int val, int lru)
{
	if (proc_get_freezer_status() == CGROUP_FREEZER_DISABLED)
		return false;

	if ((PROC_BACKGROUND + val) == lru)
		return true;
	return false;
}

enum proc_state proc_check_suspend_state(struct proc_app_info *pai)
{
	if (!pai)
		return PROC_STATE_DEFAULT;

	if (pai->type == PROC_TYPE_GUI) {
		/*
		 * check LRU state about UI application
		 * whether it is active state or not
		 */
		if (pai->lru_state < PROC_BACKGROUND)
			return PROC_STATE_DEFAULT;

		/*
		 * if platform has a suspend policy and application has UI,
		 * waits suspend callback or changing LRU.
		 * Otherwise, application goes to suspend state without waiting.
		 */
		if (!(CHECK_BIT(pai->flags, PROC_BGCTRL_PLATFORM)) ||
		    (pai->state == PROC_STATE_SUSPEND_READY))
			return PROC_STATE_SUSPEND;

		pai->state = PROC_STATE_SUSPEND_READY;
		return PROC_STATE_SUSPEND_READY;

	}
	if (pai->type == PROC_TYPE_SERVICE) {
		/*
		 * standalone service goes to suspend state immediately.
		 * if service is connected with UI application,
		 * checks UI state from program list.
		 * if UI has already went to suspend mode,
		 * service goes to suspend state.
		 * Otherwise, service waits until UI app is suspended.
		 */
		struct proc_program_info *ppi = pai->program;
		struct proc_app_info *ui;

		if (!ppi->app_list)
			return PROC_STATE_SUSPEND;

		ui = (struct proc_app_info *)g_slist_nth_data(ppi->app_list, 0);
		if (ui->state == PROC_STATE_SUSPEND)
			return PROC_STATE_SUSPEND;
		pai->state = PROC_STATE_SUSPEND_READY;
		return PROC_STATE_SUSPEND_READY;
	}
	return PROC_STATE_DEFAULT;
}


int resourced_proc_status_change(int status, pid_t pid, char *app_name, char *pkg_name, int apptype)
{
	int ret = 0, oom_score_adj = 0;
	char pidbuf[MAX_DEC_SIZE(int)];
	struct proc_status ps = {0};
	struct proc_program_info *ppi;

	if (status != PROC_CGROUP_SET_TERMINATED &&
	    status != PROC_CGROUP_SET_NOTI_REQUEST) {
		if (!pid) {
			_E("invalid pid : %d of %s", pid, app_name ? app_name : "noprocess");
			return RESOURCED_ERROR_FAIL;
		}
		ret = proc_get_oom_score_adj(pid, &oom_score_adj);
		if (ret < 0) {
			_E("Empty pid or process not exists. %d", pid);
			return RESOURCED_ERROR_FAIL;
		}
	}

	ps.pid = pid;
	ps.pai = NULL;
	switch (status) {
	case PROC_CGROUP_SET_FOREGRD:
		if (app_name)
			_SD("set foreground: app %s, pid %d", app_name, pid);
		else
			_SD("set foreground: pid %d", pid);

		ps.pai = find_app_info(pid);
		if (ps.pai)
			resourced_notify(RESOURCED_NOTIFIER_APP_FOREGRD, &ps);

		if (apptype == PROC_TYPE_WIDGET || apptype == PROC_TYPE_WATCH) {
			if (!ps.pai)
				ps.pai = proc_add_app_info(app_name, pkg_name, pid, 0, 0, apptype, PROC_STATE_FOREGROUND);
			proc_set_oom_score_adj(pid, OOMADJ_FOREGRD_UNLOCKED);
			resourced_notify(RESOURCED_NOTIFIER_WIDGET_FOREGRD, &ps);
		} else {
			if (proc_get_freezer_status() != CGROUP_FREEZER_DISABLED &&
					apptype == PROC_TYPE_GUI)
				resourced_notify(RESOURCED_NOTIFIER_APP_WAKEUP, &ps);
			snprintf(pidbuf, sizeof(pidbuf), "%d", pid);
			dbus_proc_handler(PREDEF_FOREGRD, pidbuf);
			ret = proc_set_foregrd(pid, oom_score_adj);
			if (ret != 0)
				return RESOURCED_ERROR_NO_DATA;
			if (ps.pai)
				resourced_notify(RESOURCED_NOTIFIER_APP_FOREGRD, &ps);
		}
		break;
	case PROC_CGROUP_SET_LAUNCH_REQUEST:
		if (!app_name) {
			_E("launch request: need app name! pid %d", pid);
			return RESOURCED_ERROR_NO_DATA;
		}

		if (pkg_name)
			_SD("launch request: app %s, pkg %s, pid %d", app_name, pkg_name, pid);
		else
			_SD("launch request: app %s, pid %d", app_name, pid);

		ps.pai = proc_add_app_info(app_name, pkg_name, pid, 0, 0, apptype, PROC_STATE_FOREGROUND);
		if (!ps.pai)
			break;

		if (apptype == PROC_TYPE_WIDGET)
			proc_set_oom_score_adj(pid, OOMADJ_BACKGRD_LOCKED);
		else
			proc_set_oom_score_adj(pid, OOMADJ_INIT);

		if (CHECK_BIT(ps.pai->flags, PROC_VIP_ATTRIBUTE))
			proc_set_oom_score_adj(pid, OOMADJ_SU);

		if (ps.pai->categories)
			proc_set_runtime_exclude_list(pid, PROC_EXCLUDE);
		resourced_notify(RESOURCED_NOTIFIER_APP_LAUNCH, &ps);
		if (proc_get_freezer_status() == CGROUP_FREEZER_DISABLED)
			break;
		ppi = ps.pai->program;
		if (ppi->svc_list)
			resourced_notify(RESOURCED_NOTIFIER_SERVICE_WAKEUP, &ps);
		break;
	case PROC_CGROUP_SET_SERVICE_REQUEST:
		if (!app_name) {
			_E("service launch request: need app name! pid = %d", pid);
			return RESOURCED_ERROR_NO_DATA;
		}
		if (pkg_name)
			_SD("service launch request: app %s, pkg %s, pid %d", app_name, pkg_name, pid);
		else
			_SD("service launch request: app %s, pid %d", app_name, pid);

		ps.pai = proc_add_app_info(app_name, pkg_name, pid, 0, 0, apptype, PROC_STATE_FOREGROUND);
		if (!ps.pai)
			break;
		resourced_notify(RESOURCED_NOTIFIER_SERVICE_LAUNCH, &ps);
		if (!(CHECK_BIT(ps.pai->flags, PROC_BGCTRL_APP)) ||
		    ps.pai->categories)
			proc_set_runtime_exclude_list(pid, PROC_EXCLUDE);
		break;
	case PROC_CGROUP_SET_RESUME_REQUEST:
		/* init oom_score_value */
		if (!app_name) {
			_E("resume request: need app name! pid = %d", pid);
			return RESOURCED_ERROR_NO_DATA;
		}
		_SD("resume request: app %s, pid %d", app_name, pid);

		ps.pai = find_app_info(pid);
		if (!ps.pai)
			ps.pai = proc_add_app_info(app_name, pkg_name, pid, 0, 0, apptype, PROC_STATE_FOREGROUND);

		if (!ps.pai)
			return RESOURCED_ERROR_NO_DATA;

		ps.pai->lru_state = PROC_ACTIVE;
		if (apptype == PROC_TYPE_GUI && oom_score_adj >= OOMADJ_FAVORITE) {
			resourced_notify(RESOURCED_NOTIFIER_APP_RESUME, &ps);
			proc_set_oom_score_adj(pid, OOMADJ_INIT);
		}
		if (proc_get_freezer_status() == CGROUP_FREEZER_DISABLED)
			break;
		if (apptype == PROC_TYPE_GUI)
			resourced_notify(RESOURCED_NOTIFIER_APP_WAKEUP, &ps);
		else if (apptype == PROC_TYPE_SERVICE)
			resourced_notify(RESOURCED_NOTIFIER_SERVICE_WAKEUP, &ps);
		break;
	case PROC_CGROUP_SET_TERMINATE_REQUEST:
		if (app_name)
			_SD("terminate request: app %s, pid %d", app_name, pid);
		else
			_SD("terminate request: pid %d", pid);

		ps.pai = find_app_info(pid);
		ps.pid = pid;
		resourced_notify(RESOURCED_NOTIFIER_APP_TERMINATE_START, &ps);
		resourced_notify(RESOURCED_NOTIFIER_APP_WAKEUP, &ps);
		break;
	case PROC_CGROUP_SET_ACTIVE:
		ret = proc_set_active(pid, oom_score_adj);
		if (ret != RESOURCED_ERROR_OK)
			break;

		resourced_notify(RESOURCED_NOTIFIER_APP_ACTIVE, &ps);
		break;
	case PROC_CGROUP_SET_BACKGRD:
		if (app_name)
			_SD("set background: app %s, pid %d", app_name, pid);
		else
			_SD("set background: pid %d", pid);

		if (apptype == PROC_TYPE_WIDGET  || apptype == PROC_TYPE_WATCH) {
			ps.pai = find_app_info(pid);
			if (!ps.pai)
				ps.pai = proc_add_app_info(app_name, pkg_name, pid, 0, 0, apptype, PROC_STATE_FOREGROUND);
			proc_set_oom_score_adj(pid, OOMADJ_BACKGRD_PERCEPTIBLE);
			if (apptype == PROC_TYPE_WATCH)
				break;
			resourced_notify(RESOURCED_NOTIFIER_WIDGET_BACKGRD, &ps);
		} else {
			snprintf(pidbuf, sizeof(pidbuf), "%d", pid);
			dbus_proc_handler(PREDEF_BACKGRD, pidbuf);
			ret = proc_set_backgrd(pid, oom_score_adj);
			if (ret != 0)
				break;

			if (proc_get_freezer_status() == CGROUP_FREEZER_DISABLED)
				break;
			if (resourced_freezer_proc_late_control())
				break;

			ps.pai = find_app_info(pid);
			if (!ps.pai)
				ps.pai = proc_add_app_info(app_name, pkg_name, pid, 0, 0, apptype, PROC_STATE_BACKGROUND);
			ps.pid = pid;
			resourced_notify(RESOURCED_NOTIFIER_APP_WAKEUP, &ps);
		}
		break;
	case PROC_CGROUP_SET_INACTIVE:
		ret = proc_set_inactive(pid, oom_score_adj);
		if (ret != RESOURCED_ERROR_OK)
			break;
		resourced_notify(RESOURCED_NOTIFIER_APP_INACTIVE, &ps);
		break;
	case PROC_CGROUP_GET_MEMSWEEP:
		ret = proc_sweep_memory(PROC_SWEEP_EXCLUDE_ACTIVE, pid);
		break;
	case PROC_CGROUP_SET_NOTI_REQUEST:
		if (!app_name || proc_get_freezer_status() == CGROUP_FREEZER_DISABLED)
			break;
		ps.pai = find_app_info_by_appid(app_name);
		if (!ps.pai) {
			_E("set noti request: no entry for %s in app list!", app_name);
			break;
		}
		ps.pid = ps.pai->main_pid;
		resourced_notify(RESOURCED_NOTIFIER_APP_WAKEUP, &ps);
		resourced_notify(RESOURCED_NOTIFIER_APP_ACTIVE, &ps);
		break;
	case PROC_CGROUP_SET_PROC_EXCLUDE_REQUEST:
		proc_set_runtime_exclude_list(pid, PROC_EXCLUDE);
		break;
	case PROC_CGROUP_SET_TERMINATED:
		ps.pai = find_app_info(pid);
		if (!ps.pai)
			break;

		resourced_notify(RESOURCED_NOTIFIER_APP_TERMINATED, &ps);
		proc_app_info_remove_pid(ps.pai, pid);
		break;
	case PROC_CGROUP_SET_SYSTEM_SERVICE:
		if (oom_score_adj < OOMADJ_BACKGRD_PERCEPTIBLE)
			proc_set_oom_score_adj(pid, OOMADJ_BACKGRD_PERCEPTIBLE);
		resourced_notify(RESOURCED_NOTIFIER_SYSTEM_SERVICE, &ps);
		break;
	default:
		ret = RESOURCED_ERROR_INVALID_PARAMETER;
	}
	return ret;
}

int resourced_proc_action(int status, int argnum, char **arg)
{
	pid_t pid;
	char *pidbuf = NULL, *cgroup_name = NULL, *pkg_name = NULL;
	if (argnum < 1) {
		_E("Unsupported number of arguments!");
		return RESOURCED_ERROR_INVALID_PARAMETER;
	}

	pidbuf = arg[0];
	pid = (pid_t)atoi(pidbuf);
	if (pid < 0) {
		_E("Invalid pid argument!");
		return RESOURCED_ERROR_INVALID_PARAMETER;
	}

	/* Getting appid */
	if (argnum > 1)
		/* It's possible to get appid from arg */
		cgroup_name = arg[1];
	if (argnum == 3)
		pkg_name = arg[2];
	_SD("appid %s, pid %d, status %d\n", cgroup_name, pid, status);
	return resourced_proc_status_change(status, pid, cgroup_name, pkg_name, PROC_TYPE_GUI);
}

int proc_get_state(int type, pid_t pid, char *buf, int len)
{
	int ret = 0;
	switch (type) {
	case PROC_CGROUP_GET_CMDLINE:
		ret = proc_get_raw_cmdline(pid, buf, len);
		break;
	case PROC_CGROUP_GET_PGID_CMDLINE:
		ret = proc_get_raw_cmdline(getpgid(pid), buf, len);
		break;
	case PROC_CGROUP_GET_EXE:
		ret = proc_get_exepath(pid, buf, len);
		break;
	case PROC_CGROUP_GET_STAT:
		ret = proc_get_stat(pid, buf, len);
		break;
	case PROC_CGROUP_GET_STATUS:
		ret = proc_get_status(pid, buf, len);
		break;
	case PROC_CGROUP_GET_OOMSCORE:
		ret = proc_get_oom_score_adj(pid, (int *)buf);
		break;
	default:
		_E("unsupported command %d, pid(%d)", type, pid);
		ret = RESOURCED_ERROR_FAIL;
		break;
	}
	return ret;
}

void resourced_proc_dump(int mode, const char *dirpath)
{
	char buf[MAX_PATH_LENGTH];
	char filename[MAX_NAME_LENGTH];
	int name_len;
	_cleanup_fclose_ FILE *f = NULL;

	if (dirpath) {
		time_t now;
		struct tm cur_tm;

		now = time(NULL);
		if (localtime_r(&now, &cur_tm) == NULL)
			_E("Fail to get localtime");

		name_len = snprintf(filename, MAX_NAME_LENGTH,
				"/%s_%.4d%.2d%.2d%.2d%.2d%.2d.log",
				LOG_PREFIX, (1900 + cur_tm.tm_year), 1 + cur_tm.tm_mon,
				cur_tm.tm_mday, cur_tm.tm_hour, cur_tm.tm_min, cur_tm.tm_sec);

		if (strlen(dirpath) + name_len + 1 > MAX_PATH_LENGTH) {
			_E("Dump path is too long");
			return;
		}

		snprintf(buf, MAX_PATH_LENGTH, "%s%s", dirpath, filename);
		f = fopen(buf, "w+");
	}
	proc_dump_process_list(f);
	modules_dump((void *)f, mode);
}

static struct module_ops proc_modules_ops = {
	.priority	= MODULE_PRIORITY_HIGH,
	.name		= "PROC",
	.init		= resourced_proc_init,
	.exit		= resourced_proc_exit,
	.restore	= resourced_proc_restore,
};

MODULE_REGISTER(&proc_modules_ops)
