/*
 * resourced
 *
 * Copyright (c) 2015 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

/*
 * @file proc-usage-stats.c
 * Copyright (c) 2015 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * ** Handles the dbus method calls made by the runtime-info APIs
 * ** The supported calls are for collection of memory and cpu usage of input processes
 * ** The fields collected can be found in the process_memory_info_s and
 *	process_cpu_usage_s structs in the proc-usage-stats-helper.h file
 * ** The working is as follows:
 *	* The dbus method is called from runtime-info API
 *	* For each request, a runtime_info_task instance is created
 *	* This instance contains info related to that request
 *	* A pipe is created for each request and an glib fd handler is added for the read end
 *	* A thread is swapned for the request and this thread collects the needed info
 *	* After collection of the usage info, the thread writes the success status to the
 *		write end of the pipe
 *	* This activates the handler, which appends the collected usage info to the reply
 *		dbus message and sends it to the process making the method call
 *	* The reply message is as follows:
 *		* If everything succeeded, it is an array of memory or cpu usage info structs
 *		* If some of the PIDs are invalid, then the entries for those PIDs will be
 *			INVALID_PROCESS_FIELD_VALUE
 *		* If anything went wrong in the above steps, then the reply is an integer
 *			which contains the error value
 *	* In case there was any error writing to the pipe from the swapned thread,
 *		then there is no reply and the method call fails after the timeout
 */

#include <sys/types.h>
#include <unistd.h>
#include <errno.h>

#include "proc-main.h"
#include "proc-usage-stats.h"
#include "proc-usage-stats-helper.h"
#include "resourced.h"
#include "macro.h"
#include "trace.h"
#include "dbus-handler.h"
#include "util.h"
#include "fd-handler.h"

/* g file handler for the read end of the pipe.
 * Receives the error status from the runtime info task thread and collects
 * the usage info calculated and sends it in a dbus message or send an error dbus message back
 * with the error status added to the message
 */
static bool proc_runtime_info_task_cb(int fd, void *data)
{
	int i, ret, rsize;
	int result[7];
	GVariantBuilder builder;
	GVariant *usage;
	struct runtime_info_task *rt_task;

	/* In case of errors in glib file hander, the returned dbus message
	 * contains only a failure value */
	rt_task = (struct runtime_info_task *)data;
	g_assert(rt_task && rt_task->task_invocation);

	rsize = read(fd, &ret, sizeof(int));
	if (rsize != sizeof(int)) {
		_E("task %s: error reading value from read end of pipe", rt_task->task_name);
		goto error;
	}
	_D("task %s: received %d on the read end", rt_task->task_name, ret);
	if (ret != RESOURCED_ERROR_NONE) {
		_E("task %s: error in collection of information", rt_task->task_name);
		goto error;
	}

	if (rt_task->task_type == RUNTIME_INFO_TASK_MEMORY)
		g_variant_builder_init(&builder, G_VARIANT_TYPE("a(iiiiiii)"));
	else
		g_variant_builder_init(&builder, G_VARIANT_TYPE("a(ii)"));

	/* Populate the reply message with the usage info */
	for (i = 0; i < rt_task->task_size; ++i) {
		/* Write the fields of the usage info struct to an int array
		 * (this is so that the code of dbus message append could be more elegant) */
		ret = proc_read_from_usage_struct(rt_task->usage_info_list, i, result, rt_task->task_type);
		if (ret != RESOURCED_ERROR_NONE) {
			_E("task %s: error in reading from usage info struct", rt_task->task_name);
			goto error;
		}

		if (rt_task->task_type == RUNTIME_INFO_TASK_MEMORY)
			usage = g_variant_new("(iiiiiii)", result[0], result[1], result[2],
					result[3], result[4], result[5], result[6]);
		else
			usage = g_variant_new("(ii)", result[0], result[1]);

		g_variant_builder_add_value(&builder, usage);
	}

	goto send_message;

error:
	_D("task %s: error occured in collection of usage info, return error value", rt_task->task_name);

	g_dbus_method_invocation_return_error(rt_task->task_invocation, G_DBUS_ERROR,
			G_DBUS_ERROR_IO_ERROR, "Failed to get usage information");
	proc_free_runtime_info_task(rt_task);

	return false;

send_message:
	if (rt_task->task_type == RUNTIME_INFO_TASK_MEMORY)
		g_dbus_method_invocation_return_value(rt_task->task_invocation,
				g_variant_new("(a(iiiiiii))", &builder));
	else
		g_dbus_method_invocation_return_value(rt_task->task_invocation,
				g_variant_new("(a(ii))", &builder));
	proc_free_runtime_info_task(rt_task);

	return false;
}

static int proc_runtime_info_task(struct runtime_info_task *rt_task)
{
	int i, ret;
	int wsize;

	/* Populate the usage_info_list with the needed info. There can be no failures here. */
	if (rt_task->task_type == RUNTIME_INFO_TASK_MEMORY) {
		struct process_memory_info_s *mem_info;

		mem_info = (struct process_memory_info_s *)rt_task->usage_info_list;
		for (i = 0; i < rt_task->task_size; ++i)
			proc_get_memory_usage(rt_task->pid_list[i], &mem_info[i]);
	} else {
		struct process_cpu_usage_s *cpu_usage;

		cpu_usage = (struct process_cpu_usage_s *)rt_task->usage_info_list;
		for (i = 0; i < rt_task->task_size; ++i)
			proc_get_cpu_usage(rt_task->pid_list[i], &cpu_usage[i]);
	}

	/* Write to the write end of the pipe depending on the success of
	 * the info collection (currently this is always success) */
	ret = RESOURCED_ERROR_NONE;
	wsize = write(rt_task->pipe_fds[1], &ret, sizeof(int));
	if (wsize != sizeof(int)) {
		_E("task %s: error in writing to write end of pipe", rt_task->task_name);
		return RESOURCED_ERROR_FAIL;
	}
	return RESOURCED_ERROR_NONE;
}

/* Task thread start routine. Gathers needed info and writes it to
 * the memory in the runtime_info_task instance.
 */
static void *proc_runtime_info_task_thread(void *arg)
{
	int ret;
	struct runtime_info_task *rt_task = (struct runtime_info_task *)arg;
	char task_name[TASK_NAME_SIZE] = {0, };

	if (!rt_task) {
		_E("invalid arguments!");
		return NULL;
	}

	strncpy(task_name, rt_task->task_name, TASK_NAME_SIZE - 1);
	task_name[TASK_NAME_SIZE - 1] = '\0';

	ret = proc_runtime_info_task(rt_task);
	if (ret != RESOURCED_ERROR_NONE) {
		/* TODO: write code to close fds and release the runtime_task_info memory */
		_E("task %s: request was not completed! no reply to be sent to runtime-info!!!", task_name);
	} else {
		_D("task %s: finished processing, task status %d", task_name, ret);
	}

	return NULL;
}

void proc_runtime_info_request_handler(GDBusMethodInvocation *invocation,
		GVariant *params, runtime_info_task_type type)
{
	int ret;
	pthread_t task_thread;
	struct runtime_info_task *rt_task;
	static fd_handler_h fd_handler;
	GVariantIter *iter = NULL;
	int tmp;
	int i = 0;

	static const char *gtype = "(ai)";
	ret_if_gvariant_type_mismatch(params, gtype);

	rt_task = NULL;

	/* Create runtime_info_task for current task */
	rt_task = (struct runtime_info_task *)calloc(1, (sizeof(struct runtime_info_task)));
	if (!rt_task) {
		_E("out of memory: not able to create runtime_info_task");
		ret = -ENOMEM;
		goto error;
	}

	rt_task->usage_info_list = NULL;
	rt_task->pipe_fds[0] = -1;
	rt_task->pipe_fds[1] = -1;

	/* Populate fields of the runtime_info_task */
	proc_get_task_name(rt_task->task_name, sizeof(rt_task->task_name),
			type);
	rt_task->task_type = type;
	rt_task->task_invocation = invocation;

	_D("Received %s usage request, task name is %s",
			(rt_task->task_type == RUNTIME_INFO_TASK_MEMORY) ? "memory" : "cpu",
			rt_task->task_name);

	/* Parse arguments to the pids and count it */
	g_variant_get(params, gtype, &iter);
	rt_task->task_size = g_variant_iter_n_children(iter);
	rt_task->pid_list = (int *)malloc(rt_task->task_size * sizeof(int));
	g_assert(rt_task->pid_list);

	while (g_variant_iter_loop(iter, "i", &tmp))
		rt_task->pid_list[i++] = tmp;

	if (rt_task->task_type == RUNTIME_INFO_TASK_MEMORY)
		rt_task->usage_info_list = (void *)malloc(sizeof(struct process_memory_info_s) * rt_task->task_size);
	else
		rt_task->usage_info_list = (void *)malloc(sizeof(struct process_cpu_usage_s) * rt_task->task_size);
	if (!rt_task->usage_info_list) {
		_E("task %s: out of memory: not able to create usage_info_list of rt_task", rt_task->task_name);
		ret = -ENOMEM;
		goto error;
	}

	/* Create pipe between main loop and (to-be-created) task thread and add glib file handler for the read end */
	ret = pipe(rt_task->pipe_fds);
	if (ret) {
		_E("task %s: error creating pipe.", rt_task->task_name);
		ret = -EIO;
		goto error;
	}
	ret = add_fd_read_handler(rt_task->pipe_fds[0], proc_runtime_info_task_cb,
			(void *)rt_task, NULL, &fd_handler);
	if (ret) {
		_E("task %s: error creating glib file handler", rt_task->task_name);
		ret = -EREMOTEIO;
		goto error;
	}

	/* Create task thread to complete requested task */
	ret = pthread_create(&task_thread, NULL, (void *)proc_runtime_info_task_thread, (void *)rt_task);
	if (ret) {
		_E("task %s: error creating task thread", rt_task->task_name);
		ret = -EREMOTEIO;
		goto error;
	} else
		pthread_detach(task_thread);
	_D("task %s: created thread for task", rt_task->task_name);
	return;

error:
	/* In case of error, return only a failure value in the reply dbus message */
	if (rt_task)
		_D("task %s: error occured, return error value", rt_task->task_name);

	g_dbus_method_invocation_return_error(invocation, G_DBUS_ERROR,
			G_DBUS_ERROR_FAILED, "Failed to create usage task");

	proc_free_runtime_info_task(rt_task);
}

void dbus_proc_memory_usage(GDBusMethodInvocation *invocation, GVariant *params)
{
	proc_runtime_info_request_handler(invocation, params, RUNTIME_INFO_TASK_MEMORY);
}

void dbus_proc_cpu_usage(GDBusMethodInvocation *invocation, GVariant *params)
{
	proc_runtime_info_request_handler(invocation, params, RUNTIME_INFO_TASK_CPU);
}

/*
 * These D-Bus methods are registered at proc-monitor.c
 * because gdbus allows to register object
 * having same path and interface only one time.
 */

static int proc_usage_stats_init(void *data)
{
	return RESOURCED_ERROR_NONE;
}

static int proc_usage_stats_exit(void *data)
{
	return RESOURCED_ERROR_NONE;
}

static const struct proc_module_ops proc_usage_stats_ops = {
	.name           = "PROC_USAGE_STATS",
	.init           = proc_usage_stats_init,
	.exit           = proc_usage_stats_exit,
};
PROC_MODULE_REGISTER(&proc_usage_stats_ops)
