/*
 * resourced
 *
 * Copyright (c) 2014 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

/**
 * @file init.c
 * @desc Resourced initialization
 *
 **/

#include "const.h"
#include "dbus-handler.h"
#include "cgroup.h"
#include "init.h"
#include "macro.h"
#include "module.h"
#include "module-data.h"
#include "proc-main.h"
#include "proc-monitor.h"
#include "swap-common.h"
#include "trace.h"
#include "version.h"
#include "file-helper.h"

#include <glib.h>
#include <getopt.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/types.h>

static int restarted = false;

static void print_root_usage()
{
	puts("You must be root to start it.");
}

static int assert_root(void)
{
	if (getuid() != 0) {
		print_root_usage();
		return RESOURCED_ERROR_FAIL;
	}
	return RESOURCED_ERROR_NONE;
}

int resourced_restarted(void)
{
	return restarted;
}

static int resourced_write_state(void)
{
	pid_t pid = getpid(), old_pid = 0;
	int ret;

	assert(pid >= 0);

	ret = mkdir(RUNTIME_RESOURCED_DIR, S_IRWXU | S_IRGRP | S_IXGRP);
	if (ret < 0 && errno != EEXIST)
		_E("Failed to create directory %s: %m",
				RUNTIME_RESOURCED_DIR);

	if (access(RUNTIME_RESOURCED_PID_PATH, F_OK) < 0)
		goto write_pid;

	ret = fread_uint(RUNTIME_RESOURCED_PID_PATH, (u_int32_t*)&old_pid);
	if (ret < 0) {
		_E("Failed to read old resourced pid: %m");
		return ret;
	}

	assert(pid != old_pid);
	restarted = true;

	_I("resourced is restarted: new_pid(%d), old_pid(%d)", pid, old_pid);

write_pid:
	return fwrite_uint(RUNTIME_RESOURCED_PID_PATH, (u_int32_t)pid);
}

static void sig_term_handler(int sig)
{
	_E("sigterm or sigint received");
	resourced_quit_mainloop();
}

static void add_signal_handler(void)
{
	signal(SIGTERM, sig_term_handler);
	signal(SIGINT, sig_term_handler);
}

GMainLoop *get_main_loop(void)
{
	static GMainLoop *mainloop;

	if (!mainloop)
		mainloop = g_main_loop_new(NULL, false);

	return mainloop;
}

static gboolean quit_main_loop(gpointer user_data)
{
	GMainLoop *mainloop = get_main_loop();

	if (mainloop)
		g_main_loop_quit(mainloop);

	return FALSE;
}

int resourced_init(struct daemon_arg *darg)
{
	int ret;

	ret_value_msg_if(darg == NULL, RESOURCED_ERROR_INVALID_PARAMETER,
			 "Invalid daemon argument\n");
	ret = assert_root();
	ret_value_if(ret < 0, RESOURCED_ERROR_FAIL);

	ret = resourced_write_state();
	if (ret < 0)
		return ret;

	add_signal_handler();
	d_bus_init();

	ret = modules_add_methods();
	if (ret < 0) {
		_E("Failed to add resourced module DBus method calls");
		return ret;
	}

	/* we couldn't create timer in signal callback,
	 * due g_timeout_source_new_seconds allocates memory */
	darg->mainloop_quit = g_timeout_source_new_seconds(TIME_TO_SAFE_DATA);
	g_source_set_callback(darg->mainloop_quit, quit_main_loop, NULL, NULL);

	return RESOURCED_ERROR_NONE;
}

int resourced_deinit(void)
{
	d_bus_exit();
	return RESOURCED_ERROR_NONE;
}

void resourced_quit_mainloop(void)
{
	static bool resourced_quit;

	if (resourced_quit)
		return;
	resourced_quit = true;

	struct shared_modules_data *shared_data = get_shared_modules_data();

	g_source_attach(shared_data->darg->mainloop_quit, NULL);
}

