/*
 * resourced
 *
 * Copyright (c) 2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "macro.h"
#include "module.h"
#include "module-data.h"
#include "trace.h"
#include "util.h"
#include "file-helper.h"
#include "procfs.h"
#include "swap-common.h"
#include "memory-common.h"
#include "config-parser.h"
#include "lowmem-handler.h"

#define SWAP_ZRAM_DISK_SIZE		SWAP_ZRAM_SYSFILE"disksize"
#define SWAP_ZRAM_MAX_COMP_STREAMS	SWAP_ZRAM_SYSFILE"max_comp_streams"
#define SWAP_ZRAM_COMP_ALGORITHM	SWAP_ZRAM_SYSFILE"comp_algorithm"
#define SWAP_ZRAM_COMPACT		SWAP_ZRAM_SYSFILE"compact"
#define SWAP_ZRAM_MEM_USED_TOTAL	SWAP_ZRAM_SYSFILE"mem_used_total"

#define DEFAULT_ZRAM_RATIO ((double) 0.25)
#define ZRAM_FULLNESS_RATIO		0.8
#define SWAPCG_CHECK_RATIO		5  /* 5 % of total swap */

struct swap_zram_control {
	int max_comp_streams;
	char comp_algorithm[MAX_TYPE_LENGTH];
	float ratio;
	unsigned long zram_reclaim_bytes;
};

static struct swap_zram_control zram_control = {
	.max_comp_streams = -1,
	.comp_algorithm = "lzo",
	.ratio = DEFAULT_ZRAM_RATIO,
	.zram_reclaim_bytes = 0,
};

static int swap_zram_compact(void)
{
	unsigned int total;
	static unsigned int last_total;
	int r;

	_D("call zram compact");
	r = fread_uint(SWAP_ZRAM_MEM_USED_TOTAL, &total);
	if (r < 0) {
		_E("fail to read %s", SWAP_ZRAM_MEM_USED_TOTAL);
		return r;
	}

	/*
	 * Until zram size not increased of at least 1 MB from last compaction
	 * then it not makes any sense to compact it again.
	 */
	if ((total - last_total) < MBYTE_TO_BYTE(1))
		return -ENOMEM;

	last_total = total;
	r = fwrite_int(SWAP_ZRAM_COMPACT, 1);
	if (r < 0) {
		_E("fail to write %s", SWAP_ZRAM_COMPACT);
		return r;
	}

	r = fread_uint(SWAP_ZRAM_MEM_USED_TOTAL, &total);
	if (r < 0) {
		_E("fail to read %s", SWAP_ZRAM_MEM_USED_TOTAL);
		return r;
	}

	return 0;
}

static int swap_zram_activate(void *data)
{
	struct swap_module_ops *swap = (struct swap_module_ops *)data;
	unsigned int swap_size_bytes, read_size_bytes;
	int r;

	swap_size_bytes = KBYTE_TO_BYTE(swap->k_size);
	zram_control.zram_reclaim_bytes =
			swap_size_bytes - (swap_size_bytes * ZRAM_FULLNESS_RATIO);

	if (swap_is_on(swap->path))
		return -EEXIST;

	r = fwrite_int(SWAP_ZRAM_MAX_COMP_STREAMS,
				zram_control.max_comp_streams);
	if (r < 0) {
		_E("fail to write max_comp_streams");
		return r;
	}

	r = fwrite_str(SWAP_ZRAM_COMP_ALGORITHM,
				zram_control.comp_algorithm);
	if (r < 0) {
		_E("fail to write comp_algrithm");
		return r;
	}

	r = fwrite_uint(SWAP_ZRAM_DISK_SIZE, swap_size_bytes);
	if (r < 0) {
		_E("fail to write disk_size");
		return r;
	}

	r = fread_uint(SWAP_ZRAM_DISK_SIZE, &read_size_bytes);
	if (r < 0) {
		_E("fail to read zram disk_size");
		return r;
	}

	/* Check if zram was sucessfully initialized (zcomp rollback case) */
	if (read_size_bytes < swap_size_bytes) {
		_E("swap size (%d) less than expected swap size (%d)",
				read_size_bytes, swap_size_bytes);
		return RESOURCED_ERROR_OUT_OF_MEMORY;
	}

	r = do_mkswap(swap->path);
	if (r > 0) {
		_E("Failed to make swap device(%s), exit code: %d", swap->path, r);
		return -EBUSY;
	} else if (r < 0) {
		_E("Failed to make swap device(%s): %m", swap->path);
		return r;
	}

	return 0;
}

static int swap_zram_reclaim(void *data)
{
	int r, type;
	static unsigned int swap_total = 0;
	static bool zram_compact;
	unsigned long swap_available;
	unsigned int swap_usage;
	float swapcg_usage_ratio;
	unsigned int ram_available;

	swap_available = KBYTE_TO_BYTE(proc_get_swap_free());

	_D("swap available %lu, reclaimg byte %lu", swap_available, zram_control.zram_reclaim_bytes);

	/*
	 * Most kernel doesn't support migration and compaction of zmalloc.
	 * So, if zmalloc uses much kernel memory,
	 * there is no high order buddy due to memory fragmentation.
	 * It causes system sluggish or lockup problems.
	 * In case of no high order buddy, trigger LMK.
	 */
	if (lowmem_fragmentated()) {
		if (zram_compact) {
			lowmem_trigger_swap_reclaim(LMK_MEMORY, zram_control.zram_reclaim_bytes);
			zram_compact = false;
		} else {
			swap_zram_compact();
			zram_compact = true;
		}
		return -ENOSPC;
	}

	if (swap_available >= zram_control.zram_reclaim_bytes)
		return 0;

	if (!zram_compact) {
		r = swap_zram_compact();
		if (r != -ENOMEM) {
			zram_compact = true;
			return r;
		}
	}

	ram_available = proc_get_mem_available();
	if (ram_available > lowmem_get_proactive_thres())
		return 0;

	/*
	 * In this case, swap is almost full but can't comapt zram any longer.
	 * It means that there are many background processes or
	 * some process makes memory leak.
	 * So, it requires to trigger proactive oom killer.
	 * At first, check usage of swap cgroup.
	 * If swap usage of this cgroup is higher, run LMK about background applications.
	 * Otherwise, need to check all processes in order to find mallicious process.
	 */
	if (!swap_total)
		swap_total = proc_get_swap_total();

	r = memcg_get_swap_usage(LOWMEM_SWAP_CGROUP, &swap_usage);
	if (r)
		return r;
	swapcg_usage_ratio = (float)(swap_usage / (swap_total - swap_available) *100);
	if (swapcg_usage_ratio > SWAPCG_CHECK_RATIO)
		type = LMK_OLDEST;
	else
		type = LMK_MEMORY;

	lowmem_trigger_swap_reclaim(type, zram_control.zram_reclaim_bytes);
	zram_compact = false;
	return -ENOSPC;
}

static int swap_zram_parse_config_file(void)
{
	ConfigTableItem items[] = {
		{ "ZRAM",		"COMP_ALGORITHM",	config_parse_string,	0,	NULL	},
		{ "ZRAM",		"RATIO",				config_parse_float,	0,	NULL	},
		{ NULL,		NULL,				NULL,				0,	NULL	}
	};

	int r;
	_cleanup_free_ char *algorithm = NULL;

	items[0].data = &algorithm;
	items[1].data = &zram_control.ratio;

	r = config_parse_new(SWAP_CONF_FILE, (void*) items);
	if (r < 0) {
		_E("Failed to parse configuration file: %m");
		return r;
	}

	if (algorithm)
		strncpy(zram_control.comp_algorithm, algorithm, MAX_TYPE_LENGTH - 1);

	return 0;
}

static int swap_zram_init(void *data)
{
	struct swap_module_ops *swap = (struct swap_module_ops *)data;
	int r;

	if (access(swap->path, R_OK) != 0)
		return -ENOENT;

	r = swap_zram_parse_config_file();
	if (r < 0) {
		_E("Failed to parse SwapFile config: %m");
		return r;
	}

	swap->k_size = lowmem_get_ktotalram() * zram_control.ratio;

	if (zram_control.max_comp_streams < 0) {
		int cpu = proc_get_cpu_number();

		zram_control.max_comp_streams =
			/*
			 * On big.LITTLE we can have 8 cores visible
			 * but there can be used 4. Let's limit it to
			 * 4 if there is no specified value in .conf
			 * file.
			 */
			cpu >= 4 ? 4 :
			cpu > 0 ? cpu : 1;
	}

	return 0;
}

static struct swap_module_ops swap_zram_ops = {
	.name = "ZRAM",
	.type = SWAP_TYPE_ZRAM,
	.path = "/dev/zram0",
	.priority = SWAP_PRI_DEFAULT,
	.k_size = 0,
	.init = swap_zram_init,
	.activate = swap_zram_activate,
	.reclaim = swap_zram_reclaim,
};
SWAP_MODULE_REGISTER(&swap_zram_ops)
