/*
 * resourced
 *
 * Copyright (c) 2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "macro.h"
#include "module.h"
#include "module-data.h"
#include "trace.h"
#include "util.h"
#include "file-helper.h"
#include "procfs.h"
#include "swap-common.h"
#include "memory-common.h"
#include "config-parser.h"
#include "lowmem-handler.h"
#include "losetup.h"

#define DEFAULT_ZSWAP_POOL_RATIO (25)
#define ZSWAP_FULLNESS_RATIO	0.8
#define DEFAULT_ZSWAP_FILE_SIZE MBYTE_TO_BYTE(30)

#define ZSWAP_POOL_PERCENT  "/sys/module/zswap/parameters/max_pool_percent"
#define ZSWAP_POOL_TYPE	    "/sys/module/zswap/parameters/zpool"
#define ZSWAP_WRITTEN_SIZE  "/sys/kernel/debug/zswap/written_back_pages"
#define ZSWAP_ENABLED       "/sys/module/zswap/parameters/enabled"

struct swap_zswap_control {
	char crypt_type[MAX_TYPE_LENGTH];
	char swapfile[64];
	float zpool_ratio;
	unsigned long zswap_file_size;
	unsigned long zswap_reclaim_bytes;
	char zpool_type[MAX_TYPE_LENGTH];
};

static struct swap_zswap_control zswap_control = {
	.crypt_type = "aes",
	.swapfile = SWAP_FILE_NAME,
	.zpool_ratio = DEFAULT_ZSWAP_POOL_RATIO,
	.zswap_file_size = DEFAULT_ZSWAP_FILE_SIZE,
	.zswap_reclaim_bytes = 0,
	.zpool_type = "zbud",
};

static int swap_zswap_activate(void *data)
{
	struct swap_module_ops *swap = (struct swap_module_ops *)data;
	int r;

	zswap_control.zswap_reclaim_bytes =
			zswap_control.zswap_file_size * ZSWAP_FULLNESS_RATIO;

	r = swap_set_file(zswap_control.swapfile, swap, zswap_control.crypt_type);
	if (r < 0)
		return r;

	r = fwrite_int(ZSWAP_POOL_PERCENT, zswap_control.zpool_ratio);
	if (r < 0) {
		_E("fail to write max_pool_percent: %.2f", zswap_control.zpool_ratio);
		return r;
	}

	r = fwrite_str(ZSWAP_POOL_TYPE, zswap_control.zpool_type);
	if (r < 0)
		_E("fail to change zpool : %s use default", ZSWAP_POOL_TYPE);

	r = fwrite_int(ZSWAP_ENABLED, true);
	if (r < 0)
		_E("fail to enable zswap");

	return 0;
}

static int swap_zswap_reclaim(void *data)
{
	int r;
	unsigned int swap_size;

	r = fread_uint(ZSWAP_WRITTEN_SIZE, &swap_size);
	if (r < 0) {
		_E("fail to read written swap size");
		return r;
	}

	swap_size <<= PAGE_SHIFT;
	if (swap_size <= zswap_control.zswap_reclaim_bytes)
		return 0;

	/*
	 * In this case, swap storage is almost full.
	 * It means that there are many background processes or
	 * some process makes memory leak.
	 * So, it requires to trigger proactive oom killer.
	 */

	lowmem_trigger_swap_reclaim(LMK_MEMORY, swap_size);
	return -ENOSPC;
}

static int swap_zswap_parse_config_file(void)
{
	ConfigTableItem items[] = {
		{ "FILE",		"CryptType",      config_parse_string,	0,	NULL	},
		{ "FILE",		"FileSize",       config_parse_bytes,	0,	NULL	},
		{ "ZSWAP",		"PoolRatio",      config_parse_float,	0,	NULL	},
		{ "ZSWAP",		"PoolType",       config_parse_string,	0,	NULL	},
		{ NULL,		NULL,			NULL,				0,	NULL	}
	};

	int r;
	_cleanup_free_ char *crypt_type = NULL;
	_cleanup_free_ char *zpool_type = NULL;

	items[0].data = &crypt_type;
	items[1].data = &zswap_control.zswap_file_size;
	items[2].data = &zswap_control.zpool_ratio;
	items[3].data = &zpool_type;

	r = config_parse_new(SWAP_CONF_FILE, (void*) items);
	if (r < 0) {
		_E("Failed to parse configuration file: %s", strerror(-r));
		return r;
	}

	if (check_valid_compressor(crypt_type) == RESOURCED_ERROR_NONE)
		strncpy(zswap_control.crypt_type, crypt_type, MAX_TYPE_LENGTH-1);
	else
		memset(zswap_control.crypt_type, 0, MAX_TYPE_LENGTH);

	if (zpool_type)
		strncpy(zswap_control.zpool_type, zpool_type, MAX_TYPE_LENGTH-1);
	else
		memset(zswap_control.zpool_type, 0, MAX_TYPE_LENGTH);

	return 0;
}

static int swap_zswap_init(void *data)
{
	struct swap_module_ops *swap = (struct swap_module_ops *)data;
	int r;

	if (access(ZSWAP_POOL_PERCENT, R_OK) != 0)
		return -ENOENT;

	r = swap_zswap_parse_config_file();
	if (r < 0) {
		_E("Failed to parse SwapFile config: %s", strerror(-r));
		return r;
	}

	swap->k_size = BYTE_TO_KBYTE(zswap_control.zswap_file_size);
	return 0;
}

static struct swap_module_ops swap_zswap_ops = {
	.name = "ZSWAP",
	.type = SWAP_TYPE_ZSWAP,
	.path = "",
	.priority = SWAP_PRI_DISABLE,
	.k_size = 0,
	.init = swap_zswap_init,
	.activate = swap_zswap_activate,
	.reclaim = swap_zswap_reclaim,
};
SWAP_MODULE_REGISTER(&swap_zswap_ops)
